package estimation.dataManipulation;

import estimation.ssrules.plantRecord;
import java.io.*;
import java.util.*;
import nervecenter.thesisConstants;

/**
 *
 * @author Stephen P. Ryan
 */
public class stageGameDataPrep {

    Random rng;

    /** Creates a new instance of stageGameDataPrep
     */
    public stageGameDataPrep(long rngSeed) {
        rng = new Random(rngSeed);
    }

    public void prepareData() {
        boolean debug = false;
        ArrayList<plantRecord> list = new ArrayList<plantRecord>();
        TreeSet<String> marketList = new TreeSet<String>();

        for (int year = thesisConstants.START_YEAR; year <= thesisConstants.END_YEAR; year++) {
            HashSet<String> zipset = new HashSet<String>();
            int pcount = 0;
            try {
                BufferedReader in = new BufferedReader(new FileReader(new File("data/pca" + year + ".csv")));
                in.readLine();  // read headers

                in.readLine();  // read subheaders

                String line = "";
                double kilnCapacity = 0;
                int kilnCounter = 0;
                String name = new String();
                String state = new String();
                String id = new String();
                int numKilns = 0;
                double dailyCap = 0;

                while (line != null) {
                    line = in.readLine(); // read next line

                    if (line != null) {
                        int a = 0;
                        int b = line.indexOf(",");
                        double capacity = 0;
                        if (b > a) {
                            name = line.substring(0, b);
                            if (debug) {
                                System.out.print(name + ",");
                            }
                        }
                        a = b + 1;
                        b = line.indexOf(",", a); // find next comma (plant location inside)

                        a = b + 1;
                        b = line.indexOf(",", a); // state inside

                        if (b > a) {
                            state = line.substring(a, b);
                            if (state.equalsIgnoreCase("AK") || state.equalsIgnoreCase("HI") || state.equalsIgnoreCase("OR") || state.equalsIgnoreCase("WA")) {
                                state = "AKHIORWA";
                            }
                            if (state.equalsIgnoreCase("AZ") || state.equalsIgnoreCase("NV") || state.equalsIgnoreCase("NM")) {
                                state = "AZNVNM";
                            }
                            if (state.equalsIgnoreCase("AR") || state.equalsIgnoreCase("OK")) {
                                state = "AROK";
                            }
                            if (state.equalsIgnoreCase("CA")) {
                                // have to add dilineation based on zip code or something here
                                state = "CA";
                            }
                            if (state.equalsIgnoreCase("CO") || state.equalsIgnoreCase("WY")) {
                                state = "COWY";
                            }
                            if (state.equalsIgnoreCase("FL")) {
                                state = "FL";
                            }
                            if (state.equalsIgnoreCase("GA") || state.equalsIgnoreCase("TN")) {
                                state = "GATN";
                            }
                            if (state.equalsIgnoreCase("ID") || state.equalsIgnoreCase("MT") || state.equalsIgnoreCase("UT")) {
                                state = "IDMTUT";
                            }
                            if (state.equalsIgnoreCase("IL")) {
                                state = "IL";
                            }
                            if (state.equalsIgnoreCase("IN")) {
                                state = "IN";
                            }
                            if (state.equalsIgnoreCase("IA") || state.equalsIgnoreCase("NE") || state.equalsIgnoreCase("SD")) {
                                state = "IANBSD";
                            }
                            if (state.equalsIgnoreCase("KS")) {
                                state = "KS";
                            }
                            if (state.equalsIgnoreCase("KY") || state.equalsIgnoreCase("MS") || state.equalsIgnoreCase("NC") || state.equalsIgnoreCase("LA")) {
                                state = "KYMSNCLA";
                            }
                            if (state.equalsIgnoreCase("MD") || state.equalsIgnoreCase("VA") || state.equalsIgnoreCase("WV")) {
                                state = "MDVAWV";
                            }
                            if (state.equalsIgnoreCase("MI") || state.equalsIgnoreCase("WI")) {
                                state = "MIWI";
                            }
                            if (state.equalsIgnoreCase("MO")) {
                                state = "MO";
                            }
                            if (state.equalsIgnoreCase("NY") || state.equalsIgnoreCase("ME")) {
                                state = "NYME";
                            }
                            if (state.equalsIgnoreCase("OH")) {
                                state = "OH";
                            }
                            if (state.equalsIgnoreCase("PA")) {
                                // again differentiate
                                state = "PA";
                            }
                            if (state.equalsIgnoreCase("SC")) {
                                state = "SC";
                            }
                            if (state.equalsIgnoreCase("TX")) {
                                // differentiate three ways
                                state = "TX";
                            }
                            if (debug) {
                                System.out.print(state + ",");
                            }
                            marketList.add(state);
                        }
                        a = b + 1;
                        b = line.indexOf(",", a); // Zip code inside

                        if (b > a) {
                            id = state + line.substring(a, b);
                        }
                        a = b + 1;
                        b = line.indexOf(",", a); // number of kilns

                        if (b > a) {
                            numKilns = Integer.valueOf(line.substring(a, b)).intValue();
                            kilnCounter = numKilns - 1;
                            if (debug) {
                                System.out.print(numKilns + ",");
                            }
                        } else {
                            kilnCounter--;
                            if (kilnCounter < 0) {
                                kilnCounter = 0;
                            }
                        }
                        a = b + 1;
                        b = line.indexOf(",", a); // finish grinding capacity

                        a = b + 1;
                        b = line.indexOf(",", a); // year

                        a = b + 1;
                        b = line.indexOf(",", a); // fuel

                        a = b + 1;
                        b = line.indexOf(",", a); // process

                        a = b + 1;
                        b = line.indexOf(",", a); // clinker capacity tons/day

                        if (b > a) {
                            if (!line.substring(a, b).equalsIgnoreCase("Shutdown")) {
                                dailyCap += Double.valueOf(line.substring(a, b)).doubleValue();
                            }
                        }
                        a = b + 1;
                        b = line.indexOf(",", a); // clinker capacity tons/year

                        if (b > 0 && b > a) {
                            if (kilnCounter >= 0) {
                                kilnCapacity += Double.valueOf(line.substring(a, b)).doubleValue();
                            }
                        } else {
                            if (kilnCounter >= 0 && !line.substring(a).equalsIgnoreCase(",")) {
                                kilnCapacity += Double.valueOf(line.substring(a)).doubleValue();
                            }
                        }
                        if (kilnCounter == 0) {
                            plantRecord plant = new plantRecord(name, state, year);
                            plant.setQ(kilnCapacity);
                            plant.setDailyCapacity(dailyCap);
                            plant.setAnnualCapacity(dailyCap * 365.0 / 1000.0);

                            String iden = id.toString();
                            int counter = 0;
                            while (zipset.contains(id)) {
                                id = iden.concat(Integer.toString(counter));
                                counter++;
                            }
                            plant.setID(id.toString());
                            list.add(plant);
                            zipset.add(id);
                            pcount++;
                            kilnCapacity = 0;
                            dailyCap = 0;
                        } else {
                            // capacity += Double.valueOf(line.substring(a,b)).doubleValue();
                        }
                    }
                }
                in.close();

            } catch (Exception e) {
                e.printStackTrace();
            }
            System.out.print(year + " zips: " + zipset.size() + " plants: " + pcount + "\n");
        }

        System.out.println("Number of distinct markets: " + marketList.size());

        /**
         * Tabulate market-wide capacity for each firm and tally number of firms.
         */
        for (int i = 0; i < list.size(); i++) {
            plantRecord p1 = list.get(i);
            ArrayList<Double> marketCapacities = new ArrayList<Double>();
            int nf = 0;
            for (int j = 0; j < list.size(); j++) {
                plantRecord p2 = list.get(j);
                if (p2.getYear() == p1.getYear()) {
                    if (p2.getState().equalsIgnoreCase(p1.getState())) {
                        marketCapacities.add(new Double(p2.getAnnualCapacity()));
                        nf++;
                    }
                }
            }
            p1.setNumFirms(nf);
            p1.setMarketCapacities(marketCapacities);
        }

        /**
         * Tabulate market quantities.
         */
        for (int i = 0; i < list.size(); i++) {
            plantRecord p1 = list.get(i);
            ArrayList<Double> marketQuantities = new ArrayList<Double>();
            double sumQ = 0;
            for (int j = 0; j < list.size(); j++) {
                plantRecord p2 = (plantRecord) list.get(j);
                if (p2.getYear() == p1.getYear()) {
                    if (p2.getState().equalsIgnoreCase(p1.getState())) {
                        marketQuantities.add(new Double(p2.getQ()));
                        sumQ += p2.getQ();
                    }
                }
            }
            p1.setMarketQuantities(marketQuantities);
            p1.setMarketQ(sumQ);
        }

        Collections.sort(list);

        ArrayList<String> bootMarketNames = new ArrayList<String>(marketList);
        if (thesisConstants.SUBSAMPLE_PARAMETERS) {
            bootMarketNames = new ArrayList<String>();
            ArrayList<plantRecord> subsampleList = new ArrayList<plantRecord>();
            ArrayList<Integer> sampledMarketIndexList = new ArrayList<Integer>();
            ArrayList<String> marketNameList = new ArrayList<String>(marketList);

            /**
             * Add AL so that it always shows up in the baseline.
             */
            if (thesisConstants.TRUE_BOOTSTRAP) {
                // draw with replacement the n-out-n bootstrap for comparison
                // to subsampling
                for (int i = 0; i < marketNameList.size(); i++) {
                    int index = (int) Math.floor(marketNameList.size() * rng.nextDouble());
                    if (i == 0) {
                        index = marketNameList.indexOf("AL");
                    } else {
                        while(index==marketNameList.indexOf("AL")) {
                            index = (int) Math.floor(marketNameList.size() * rng.nextDouble());
                        }
                    }
                    String marketName = marketNameList.get(index);
                    bootMarketNames.add(marketName);
                    System.out.print(marketName + " ");
                    for (plantRecord p : list) {
                        if (p.getState().equals(marketName)) {
                            subsampleList.add(p.deepCopy());
                        }
                    }
                    sampledMarketIndexList.add(index);
                }
                System.out.println(".");
                list = subsampleList;
            } else {
                for (int i = 0; i < thesisConstants.SUBSAMPLE_SIZE; i++) {
                    int index = (int) Math.floor(marketNameList.size() * rng.nextDouble());
                    if (i == 0) {
                        index = marketNameList.indexOf("AL");
                    }
                    while (sampledMarketIndexList.contains(index)) {
                        if (sampledMarketIndexList.contains(index)) {
                            index = (int) Math.floor(marketNameList.size() * rng.nextDouble());
                        }
                    }
                    String marketName = marketNameList.get(index);
                    bootMarketNames.add(marketName);
                    System.out.print(marketName + " ");
                    for (plantRecord p : list) {
                        if (p.getState().equals(marketName)) {
                            subsampleList.add(p.deepCopy());
                        }
                    }
                    sampledMarketIndexList.add(index);
                }
                System.out.println(".");
                list = subsampleList;
            }
        }

        Collections.sort(bootMarketNames);

        System.out.print("Writing stageGame.dat...");
        try {
            String late = "";
            if (thesisConstants.LATE_PERIOD) {
                late = "late";
            }
            String filename = "data/stageGame" + late + ".dat";
            ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(new File(filename)));
            out.writeObject(new Integer(list.size()));
            for (int i = 0; i < list.size(); i++) {
                out.writeObject(list.get(i));
            }
            out.close();

            ObjectOutputStream names = new ObjectOutputStream(new FileOutputStream("data/marketNames" + late + ".dat"));
            names.writeObject(bootMarketNames);
            names.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("done.");

        System.out.print("Writing output, capacity, entry, and exit data to data/productionData.csv...");

        TreeSet<String> tree = new TreeSet<String>();
        for (plantRecord p : list) {
            tree.add(p.getID());
        }
        ArrayList<String> idlist = new ArrayList<String>(tree);

        for (String matchID : idlist) {
            int earliestYear = 747800;
            int latestYear = -1;

            for (plantRecord p : list) {
                if (p.getID().equalsIgnoreCase(matchID)) {
                    int year = p.getYear();
                    if (year > latestYear) {
                        latestYear = year;
                    }
                    if (year < earliestYear) {
                        earliestYear = year;
                    }
                }
            }
            for (plantRecord p : list) {
                if (p.getID().equalsIgnoreCase(matchID)) {
                    if (earliestYear != thesisConstants.START_YEAR) {
                        p.setEntered(true);
                    }
                    if (latestYear != thesisConstants.END_YEAR) {
                        p.setExited(true);
                    }
                }
            }
        }

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter("data/productionData.csv"));
            out.write("ID,Year,State,Capacity,Quantity,Ratio,CapOther,Entered,Exited\n");
            for (plantRecord p : list) {
                out.write(p.getID() + "," + p.getYear() + "," + p.getState() + "," + p.getAnnualCapacity() + "," + p.getQ() + "," + (p.getQ() / p.getAnnualCapacity()) + "," + p.getCapOtherDouble() + "," + p.isEntered() + "," + p.isExited() + "\n");
            }
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("OK");
    }
}
