/*
 * To change this template, choose Tools | Templates and open the template in
 * the editor.
 */
package estimation.demand;

import Jama.Matrix;
import nervecenter.thesisConstants;

/**
 *
 * @author Stephen P. Ryan
 */
public class demandDataPoint {

    private double price;
    private double quantity;
    private double gas96;
    private double elec96;
    private double coal96;
    private double wage96;
    private int stateIndex;
    private String region;

    public demandDataPoint(String region, double price, double quantity, double gas96, double elec96, double coal96, double wage96, int stateIndex) {
        this.region = region;
        this.price = Math.log(price);
        this.quantity = Math.log(quantity);
        this.gas96 = gas96;
        this.elec96 = elec96;
        this.coal96 = coal96;
        this.wage96 = wage96;
        this.stateIndex = stateIndex;
    }

    // what do i want to return here when estimating demand?  the vector z'e?
    public Jama.Matrix getMoment(double[] x) {
        double intercept = x[1];
        if (stateIndex > 0 && thesisConstants.USE_STATE_FIXED_EFFECTS) {
            intercept += x[stateIndex + 2];
        }
        double slope = x[2];
        Jama.Matrix g = new Jama.Matrix(estimateDemand.NUM_MOMENTS, 1);
        double error = quantity - (slope * price + intercept);
        g.set(0, 0, error * 1);
        g.set(1, 0, error * gas96);
        g.set(2, 0, error * elec96);
        g.set(3, 0, error * coal96);
        g.set(4, 0, error * wage96);
        if (stateIndex > 0 && thesisConstants.USE_STATE_FIXED_EFFECTS) {
            g.set(4 + stateIndex, 0, error * 1);
        }
        return g;
    }

    public double getError(double[] x) {
        double intercept = x[1];
        if (stateIndex > 0 && thesisConstants.USE_STATE_FIXED_EFFECTS) {
            intercept += x[stateIndex + 2];
        }
        double slope = x[2];
        double error = quantity - (intercept + slope * price);
        System.out.println("quantity: " + quantity + " price: " + price + " intercept: " + intercept + " slope: " + slope + " qhat: " + (intercept + slope * price) + " error: " + error);
        return error;
    }

    /**
     * This one excludes price.
     * @return
     */
    Matrix getExogenousX() {
        double[] X = new double[thesisConstants.SUBSAMPLE_SIZE - 1 + 1];
        X[0] = 1;
        if (stateIndex > 0) {
            X[stateIndex] = 1;
        }
        if (!thesisConstants.USE_STATE_FIXED_EFFECTS) {
            X = new double[1];
            X[0] = 1;
        }
        return new Jama.Matrix(X, 1);
    }

    /**
     * This one includes price.
     * @return
     */
    Matrix getX() {
        double[] X = null;
        if (thesisConstants.USE_STATE_FIXED_EFFECTS) {
            X = new double[thesisConstants.SUBSAMPLE_SIZE - 1 + 2];
            if (stateIndex > 0) {
                X[stateIndex + 1] = 1;
            }
        } else {
            X = new double[2];
        }
        X[0] = 1;
        X[1] = price;
        return new Jama.Matrix(X, 1);

    }

    Matrix getY() {
        double[] Y = {quantity};
        return new Jama.Matrix(Y, 1);
    }

    Matrix getZ() {
        if (thesisConstants.USE_STATE_FIXED_EFFECTS) {
            double[] Z = new double[5 + thesisConstants.SUBSAMPLE_SIZE - 1];
            Z[0] = 1;
            Z[1] = gas96;
            Z[2] = elec96;
            Z[3] = coal96;
            Z[4] = wage96;
            if (stateIndex > 0) {
                Z[4 + stateIndex] = 1;
            }
            return new Jama.Matrix(Z, 1);
        }
        double[] Z = {1, gas96, elec96, coal96, wage96};
        return new Jama.Matrix(Z, 1);
    }

    Matrix getPrice() {
        double[] p = {price};
        return new Jama.Matrix(p, 1);
    }

    /**
     * @return the region
     */
    public String getRegion() {
        return region;
    }

    /**
     * @param region the region to set
     */
    public void setRegion(String region) {
        this.region = region;
    }
}
