/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package estimation.demand;

import Jama.Matrix;
import nervecenter.thesisConstants;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import optimization.Uncmin_f77;
import optimization.Uncmin_methods;
import utility.demandCurve;
import utility.pmUtility;

/**
 *
 * @author Stephen P. Ryan
 */
public class estimateDemand implements Uncmin_methods, mcmc.mcmcFunction {

    public static int NUM_MOMENTS;
    ArrayList<demandDataPoint> data = new ArrayList<demandDataPoint>();
    ArrayList<String> marketNames = null;
    int numParams;
    Jama.Matrix W;
    private Matrix beta2sls;
    private Matrix betaGMM;
    private Matrix betaLIML;
    private double shockVariance2SLS;
    private double shockVarianceGMM;
    private double shockVarianceLIML;
    private demandCurve priceThesis;

    public estimateDemand(demandCurve priceThesis) {
        this.priceThesis = priceThesis;
        prepareData();
        // numParams = 2;  // try just the intercept and slope to make sure this works
    }

    public void execute() {
        W = Jama.Matrix.identity(NUM_MOMENTS, NUM_MOMENTS);
        double[] guess = new double[numParams + 1];
//        guess[1] = -2.68;
//        guess[2] = 19;
        Uncmin_f77 minimizer = new Uncmin_f77(false);

        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        double[] fscale = {0, 1E-8};
        int[] method = {0, 1};
        int[] iexp = {0, 1};
        int[] msg = {0, 1};
        int[] ndigit = {0, 7};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-8};

        runFirstStageRegression();
        
        for (int i = 0; i < numParams; i++) {
            // guess[i + 1] = betaGMM.get(i, 0);
            guess[i + 1] = beta2sls.get(i, 0);
            // guess[i + 1] = betaLIML.get(i, 0);

            priceThesis.setShockVariance(shockVariance2SLS);
            // priceThesis.setShockVariance(shockVarianceLIML);
        }

        if (guess[2] > 0) {
            // run with constraint on elasticity if this happens
            boolean useOptimalWeightingMatrix = false;
            if (useOptimalWeightingMatrix) {
                System.out.println("With Optimal weighting matrix:");
                computeOptimalWeightingMatrix(guess);
                minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
                pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
            }
        }
        
        double[] stateIntercepts = new double[21];
        double demandConstant = guess[1];

        if (thesisConstants.USE_STATE_FIXED_EFFECTS) {
            boolean alabamaPresent = false;
            int indexAL = marketNames.indexOf("AL");
            if (indexAL >= 0) {
                // World with Alabama in the data set
                alabamaPresent = true;
                System.out.println("Alabama present; index = " + indexAL);
            }
            for (int i = 0; i < marketNames.size(); i++) {
                String state = marketNames.get(i);
                int stateIndexInPriceThesis = 0;
                if (state.equalsIgnoreCase("AKHIORWA")) {
                    stateIndexInPriceThesis = 1;
                }
                if (state.equalsIgnoreCase("AZNVNM")) {
                    stateIndexInPriceThesis = 2;
                }
                if (state.equalsIgnoreCase("AROK")) {
                    stateIndexInPriceThesis = 3;
                }
                if (state.equalsIgnoreCase("CA")) {
                    stateIndexInPriceThesis = 4;
                }
                if (state.equalsIgnoreCase("COWY")) {
                    stateIndexInPriceThesis = 5;
                }
                if (state.equalsIgnoreCase("FL")) {
                    stateIndexInPriceThesis = 6;
                }
                if (state.equalsIgnoreCase("GATN")) {
                    stateIndexInPriceThesis = 7;
                }
                if (state.equalsIgnoreCase("IDMTUT")) {
                    stateIndexInPriceThesis = 8;
                }
                if (state.equalsIgnoreCase("IL")) {
                    stateIndexInPriceThesis = 9;
                }
                if (state.equalsIgnoreCase("IN")) {
                    stateIndexInPriceThesis = 10;
                }
                if (state.equalsIgnoreCase("IANBSD")) {
                    stateIndexInPriceThesis = 11;
                }
                if (state.equalsIgnoreCase("KS")) {
                    stateIndexInPriceThesis = 12;
                }
                if (state.equalsIgnoreCase("KYMSNCLA")) {
                    stateIndexInPriceThesis = 13;
                }
                if (state.equalsIgnoreCase("MDVAWV")) {
                    stateIndexInPriceThesis = 14;
                }
                if (state.equalsIgnoreCase("MIWI")) {
                    stateIndexInPriceThesis = 15;
                }
                if (state.equalsIgnoreCase("MO")) {
                    stateIndexInPriceThesis = 16;
                }
                if (state.equalsIgnoreCase("NYME")) {
                    stateIndexInPriceThesis = 17;
                }
                if (state.equalsIgnoreCase("OH")) {
                    stateIndexInPriceThesis = 18;
                }
                if (state.equalsIgnoreCase("PA")) {
                    stateIndexInPriceThesis = 19;
                }
                if (state.equalsIgnoreCase("SC")) {
                    stateIndexInPriceThesis = 20;
                }
                if (state.equalsIgnoreCase("TX")) {
                    stateIndexInPriceThesis = 21;
                }
                
                if (alabamaPresent) {
                    double alabamaLevel = guess[1] + guess[2 + indexAL];
                    if (indexAL == 0) {
                        // this implies that alabama is the excluded state
                        alabamaLevel = guess[1];
                    }
                    if (stateIndexInPriceThesis == 0) {
                        // alabama present, and is the baseline in the specification
                        demandConstant = alabamaLevel;
                    } else {
                        // not the baseline
                        // so what is this: take the shifter (say +0.5), add to the
                        // baseline, say 15, and that's the final answer.  In pricethesis, we want the
                        // different between this and value for alabama
                        if (i > 0) {
                            stateIntercepts[stateIndexInPriceThesis - 1] = (guess[2 + i] + guess[1]) - alabamaLevel;
                        } else {
                            stateIntercepts[stateIndexInPriceThesis - 1] = -guess[2 + indexAL];
                        }
                    }
                } else {
                    // if alabama is not present, then we set the demand intercept to be zero
                    demandConstant = 0;
                    // and each fixed effect will be the constant plus the shifter
                    if (i > 0) {
                        stateIntercepts[stateIndexInPriceThesis - 1] = guess[1] + guess[2 + i];
                    } else {
                        stateIntercepts[stateIndexInPriceThesis - 1] = guess[1];
                    }
                }

                if (stateIndexInPriceThesis > 0) {
                    System.out.println(marketNames.get(i) + "\t" + stateIntercepts[stateIndexInPriceThesis - 1]);
                } else {
                    System.out.println("AL\t" + demandConstant);
                }
            }
        } else {
            /**
             * Set state intercepts to zero.
             */
            for (int i = 0; i < stateIntercepts.length; i++) {
                stateIntercepts[i] = 0;
            }
        }
        
        System.out.println("estimateDemand.java: demandConstant is " + demandConstant);
        System.out.println("estimateDemand.java: slope is " + guess[2]);
        System.out.println("estimateDemand.java: demand shock variance is "+priceThesis.getShockVariance());

        priceThesis.setConstant(demandConstant);
        priceThesis.setFixedEffects(stateIntercepts);
        priceThesis.setSlope(guess[2]);
    }

    @SuppressWarnings("unchecked")
    private void prepareData() {
        try {
            String late = "";
            if (thesisConstants.LATE_PERIOD) {
                late = "late";
            }
            ObjectInputStream in = new ObjectInputStream(new FileInputStream("data/marketNames" + late + ".dat"));
            marketNames = (ArrayList<String>) in.readObject();
            in.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("Read in markets for demand estimation: ");
        for (String s : marketNames) {
            System.out.print(s + " ");
        }
        System.out.println(".");

        // region	year	shipped	price	plants	wage96	coal96	elec96	population	gas96	logq	logp	logpop	bs1	bs2	bs3	bs4	yearnormal	logyear	scaledpop	totalpermits	unit1permits	unit2permits	unit34permits	unit5permits	totalstructures	_merge	_iregion_2	_iregion_3	_iregion_4	_iregion_5	_iregion_6	_iregion_7	_iregion_8	_iregion_9	_iregion_10	_iregion_11	_iregion_12	_iregion_13	_iregion_14	_iregion_15	_iregion_16	_iregion_17	_iregion_18	_iregion_19	_iregion_21	_iregion_22	_iregion_23	_iregion_24	_iregion_25	_iregion_26	_iregion_27

        try {
            BufferedReader in = new BufferedReader(new FileReader("data/cementDec2009.csv"));
            String line = in.readLine(); // headers
            line = in.readLine(); // first line
            while (line != null) {
                if (line != null) {
                    int a = 0;
                    int b = line.indexOf(",", a);
                    String region = line.substring(a, b);
                    int stateIndex = marketNames.indexOf(region);
                    a = b + 1;
                    b = line.indexOf(",", a);
                    // year
                    a = b + 1;
                    b = line.indexOf(",", a);
                    // quantity
                    Double quantity = new Double(line.substring(a, b));
                    a = b + 1;
                    b = line.indexOf(",", a);
                    // price
                    Double price = new Double(line.substring(a, b));
                    a = b + 1;
                    b = line.indexOf(",", a);
                    // plants
                    a = b + 1;
                    b = line.indexOf(",", a);
                    Double wage96 = new Double(line.substring(a, b));
                    a = b + 1;
                    b = line.indexOf(",", a);
                    Double coal96 = new Double(line.substring(a, b));
                    a = b + 1;
                    b = line.indexOf(",", a);
                    Double elec96 = new Double(line.substring(a, b));
                    a = b + 1;
                    b = line.indexOf(",", a);
                    // population
                    a = b + 1;
                    b = line.indexOf(",", a);
                    Double gas96 = new Double(line.substring(a, b));
                    if (stateIndex >= 0) {
                        data.add(new demandDataPoint(region, price, quantity, gas96, elec96, coal96, wage96, stateIndex));
                    }
                    line = in.readLine();
                }
            }
            in.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
        /**
         * state-specific level constants (minus Alabama), plus intercept and
         * the slope
         */
        numParams = marketNames.size() - 1 + 2;
        // number of moments: instruments + market-specific intercepts - 1 + constant
        NUM_MOMENTS = 4 + marketNames.size();
        if (!thesisConstants.USE_STATE_FIXED_EFFECTS) {
            NUM_MOMENTS = 5;
            numParams = 2;
        }
        System.out.println("NUM_MOMENTS = " + NUM_MOMENTS + "; NUM_MARKETS: " + marketNames.size());
    }

    public void runFirstStageRegression() {
        Jama.Matrix X = null;
        Jama.Matrix Y = null;
        Jama.Matrix Z = null;
        Jama.Matrix exoX = null;
        Jama.Matrix price = null;
        for (demandDataPoint p : data) {
            if (X == null) {
                exoX = p.getExogenousX();
                X = p.getX();
                Y = p.getY();
                Z = p.getZ();
                price = p.getPrice();
            } else {
                X = pmUtility.stackMatrix(X, p.getX());
                exoX = pmUtility.stackMatrix(exoX, p.getExogenousX());
                Y = pmUtility.stackMatrix(Y, p.getY());
                Z = pmUtility.stackMatrix(Z, p.getZ());
                price = pmUtility.stackMatrix(price, p.getPrice());
            }
        }

        pmUtility.prettyPrint(pmUtility.concatMatrix(price, Z));
        Jama.Matrix betaFirstStage = pmUtility.OLS(Z, price, false);
        System.out.println("First stage estimates:");
        pmUtility.prettyPrintVector(betaFirstStage);

        Jama.Matrix zp = Z.transpose();
        Jama.Matrix zpz = zp.times(Z);
        Jama.Matrix zpzinv = zpz.inverse();
        Jama.Matrix xhat = ((Z.times(zpzinv)).times(zp)).times(X);
        Jama.Matrix xhatp = xhat.transpose();
        Jama.Matrix xhatpx = xhatp.times(X);
        Jama.Matrix xhatpxinv = xhatpx.inverse();
        Jama.Matrix xhatpy = xhatp.times(Y);
        beta2sls = xhatpxinv.times(xhatpy);
        System.out.println("2SLS estimates:");
        pmUtility.prettyPrintVector(beta2sls);

        Jama.Matrix ivFit = X.times(beta2sls);
        Jama.Matrix ivErr = Y.minus(ivFit);
        shockVariance2SLS = 0;
        for (int i = 0; i < ivErr.getRowDimension(); i++) {
            shockVariance2SLS += Math.pow(ivErr.get(i, 0), 2);
        }
        shockVariance2SLS /= ivErr.getRowDimension();

        Jama.Matrix omega = new Jama.Matrix(Z.getRowDimension(), Z.getRowDimension());
        Jama.Matrix errors = Y.minus(X.times(beta2sls));
        for (int i = 0; i < Z.getRowDimension(); i++) {
            omega.set(i, i, errors.get(i, 0) * errors.get(i, 0));
        }

        Jama.Matrix S = zp.times(omega).times(Z);
        S.timesEquals(1.0 / Z.getRowDimension());

        W = S.inverse();

        Jama.Matrix xp = X.transpose();
        Jama.Matrix partA = (xp.times(Z).times(W).times(zp).times(X)).inverse();
        Jama.Matrix partB = xp.times(Z).times(W).times(zp).times(Y);
        betaGMM = partA.times(partB);
        System.out.println("GMM robust estimate:");
        pmUtility.prettyPrintVector(betaGMM);
        Jama.Matrix gmmFit = X.times(betaGMM);
        Jama.Matrix gmmErr = Y.minus(gmmFit);
        shockVarianceGMM = 0;
        for (int i = 0; i < gmmErr.getRowDimension(); i++) {
            shockVarianceGMM += Math.pow(gmmErr.get(i, 0), 2);
        }
        shockVarianceGMM /= gmmErr.getRowDimension();

        /**
         * Implement the LIML estimator here.
         */
        limlFunction liml = new limlFunction(X, Y, Z);
        betaLIML = liml.getEstimate(betaGMM);
        pmUtility.prettyPrintVector(betaLIML);
        shockVarianceLIML = liml.getShockVariance();
        // System.exit(0);
    }

    public double f_to_minimize(double[] x) {
        // ensure downward sloping curve
        x[2] = Math.min(-0.5, x[2]);
        // run gmm-iv
        double gmm = 0;
        Jama.Matrix g = new Jama.Matrix(NUM_MOMENTS, 1);
        // constant, state fixed effects if used, and four instruments
        for (demandDataPoint p : data) {
            g.plusEquals(p.getMoment(x));
        }
        g.timesEquals(1.0 / data.size());
        gmm = (((g.transpose()).times(W)).times(g)).get(0, 0);
        return data.size() * gmm;
    }

    public void gradient(double[] x, double[] g) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void hessian(double[] x, double[][] h) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private void computeOptimalWeightingMatrix(double[] guess) {
        Jama.Matrix v = new Jama.Matrix(NUM_MOMENTS, NUM_MOMENTS);
        for (demandDataPoint p : data) {
            Jama.Matrix g = p.getMoment(guess);
            v.plusEquals(g.times(g.transpose()));
        }
        v.timesEquals(1.0 / data.size());

        W = v.inverse();
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }
}
