/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package estimation.demand;

import Jama.Matrix;
import com.ziena.knitro.KnitroJava;
import optimization.Uncmin_f77;
import optimization.Uncmin_methods;

/**
 *
 * @author Stephen P. Ryan
 */
class limlFunction implements Uncmin_methods {

    Jama.Matrix X, Y, Z;
    Jama.Matrix pZ;
    double shockVariance;

    public limlFunction(Matrix X, Matrix Y, Matrix Z) {
        this.X = X;
        this.Y = Y;
        this.Z = Z;
        Jama.Matrix zpz = (Z.transpose()).times(Z);
        Jama.Matrix zpzinv = zpz.inverse();
        pZ = (Z.times(zpzinv)).times(Z.transpose());
    }

    public double f_to_minimize(double[] x) {
        Jama.Matrix delta = new Jama.Matrix(x.length - 1, 1);
        for (int i = 0; i < delta.getRowDimension(); i++) {
            delta.set(i, 0, x[i + 1]);
        }
        Jama.Matrix YminusXdelta = Y.minus(X.times(delta));
        Jama.Matrix YminusXdeltap = YminusXdelta.transpose();
        double top = (YminusXdeltap.times(pZ)).times(YminusXdelta).get(0, 0);
        double bottom = YminusXdeltap.times(YminusXdelta).get(0, 0);
        return top / bottom;
    }

    public void gradient(double[] x, double[] g) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void hessian(double[] x, double[][] h) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    Matrix getEstimate(Jama.Matrix startingValue) {
        int numParams = X.getColumnDimension();

        Uncmin_f77 minimizer = new Uncmin_f77(false);

        double[] guess = new double[numParams + 1];
        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        double[] fscale = {0, 1E-8};
        int[] method = {0, 1};
        int[] iexp = {0, 1};
        int[] msg = {0, 1};
        int[] ndigit = {0, 7};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-8};

        for(int i=0;i<startingValue.getRowDimension();i++) {
            guess[i+1] = startingValue.get(i,0);
        }
        guess[1] = 18.5;
        guess[2] = -2.5;

//        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
//        minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
//        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));

        solveUsingKNitro(guess);
        // pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
        // System.exit(0);

        Jama.Matrix delta = new Jama.Matrix(guess.length - 1, 1);
        for (int i = 0; i < delta.getRowDimension(); i++) {
            delta.set(i, 0, guess[i + 1]);
        }

        Jama.Matrix fits = X.times(delta);
        Jama.Matrix errors = Y.minus(fits);

        shockVariance = 0;
        for(int i=0;i<errors.getRowDimension();i++) {
            shockVariance+=Math.pow(errors.get(i,0),2);
        }
        shockVariance /= (errors.getRowDimension());

        return delta;
    }

    public void solveUsingKNitro(double[] guess) {
        //---- DEFINE THE OPTIMIZATION TEST PROBLEM.
        //---- FOR MORE INFORMATION ABOUT THE PROBLEM DEFINITION, REFER
        //---- TO THE KNITRO MANUAL, ESPECIALLY THE SECTION ON THE
        //---- CALLABLE LIBRARY.
        int n = guess.length - 1;
        int objGoal = KnitroJava.KTR_OBJGOAL_MINIMIZE;
        int objType = KnitroJava.KTR_OBJTYPE_GENERAL;

        // make sure that the problem is defined as unbounded
        double[] bndsLo = new double[n];
        double[] bndsUp = new double[n];
        for (int i = 0; i < n; i++) {
            bndsLo[i] = -KnitroJava.KTR_INFBOUND;
            bndsUp[i] = KnitroJava.KTR_INFBOUND;
        }

        // number of constraints
        int m = 0;
        int[] cType = {        };
        double[] cBndsLo = {};
        double[] cBndsUp = {};
        int nnzJ = 0;
        int[] jacIxConstr = {};
        int[] jacIxVar = {};
        int nnzH = 0;
        int[] hessRow = {};
        int[] hessCol = {};

        double[] daXInit = new double[guess.length - 1];
        for (int i = 0; i < guess.length - 1; i++) {
            daXInit[i] = guess[i + 1];
        }


        //---- SETUP AND RUN KNITRO TO SOLVE THE PROBLEM.

        //---- CREATE A SOLVER INSTANCE.
        KnitroJava solver = null;
        try {
            solver = new KnitroJava();
        } catch (java.lang.Exception e) {
            System.err.println(e);
            return;
        }

        // ---- DEMONSTRATE HOW TO SET KNITRO PARAMETERS.
        /**
         * Options: all, summary, none
         */
        if (solver.setCharParamByName("outlev", "none") == false) {
            System.err.println("Error setting parameter 'outlev'");
            return;
        }
//        solver.setIntParamByName("outlev", 0);
//        solver.setIntParamByName("debug", 0);
        if (solver.setDoubleParamByName("feastol", 1.0E-6) == false) {
            System.err.println("Error setting parameter 'feastol'");
            return;
        }
        if (solver.setIntParamByName("gradopt", 2) == false) {
            System.err.println("Error setting parameter 'gradopt'");
            return;
        }
        if (solver.setIntParamByName("hessopt", 6) == false) {
            System.err.println("Error setting parameter 'hessopt'");
            return;
        }

        //---- INITIALIZE KNITRO WITH THE PROBLEM DEFINITION.
        if (solver.initProblem(n, objGoal, objType, bndsLo, bndsUp,
                m, cType, cBndsLo, cBndsUp,
                nnzJ, jacIxVar, jacIxConstr,
                nnzH, hessRow, hessCol,
                daXInit, null) == false) {
            System.err.println("Error initializing the problem, " + "KNITRO status = " + solver.getKnitroStatusCode());
            return;
        }

        //---- ALLOCATE ARRAYS FOR REVERSE COMMUNICATIONS OPERATION.
        double[] daX = new double[n];
        double[] daLambda = new double[m + n];
        double[] daObj = new double[1];
        double[] daC = new double[m];
        double[] daObjGrad = new double[n];
        double[] daJac = new double[nnzJ];
        double[] daHess = new double[nnzH];

        // exampleHS15 thisObject = new exampleHS15();

        //---- SOLVE THE PROBLEM.  IN REVERSE COMMUNICATIONS MODE, KNITRO
        //---- RETURNS WHENEVER IT NEEDS MORE PROBLEM INFORMATION.  THE CALLING
        //---- PROGRAM MUST INTERPRET KNITRO'S RETURN STATUS AND CONTINUE
        //---- SUPPLYING PROBLEM INFORMATION UNTIL KNITRO IS COMPLETE.
        int nKnStatus;
        int nEvalStatus = 0;
        do {
            nKnStatus = solver.solve(nEvalStatus, daObj, daC,
                    daObjGrad, daJac, daHess);
            if (nKnStatus == KnitroJava.KTR_RC_EVALFC) {
                //---- KNITRO WANTS daObj AND daC EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // daObj[0] = thisObject.evaluateFC(daX, daC);
                daObj[0] = evaluateFC(daX, daC);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALGA) {
                //---- KNITRO WANTS daObjGrad AND daJac EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // thisObject.evaluateGA(daX, daObjGrad, daJac);
                evaluateGA(daX, daObjGrad, daJac);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALH) {
                //---- KNITRO WANTS daHess EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                daLambda = solver.getCurrentLambda();
                // thisObject.evaluateH(daX, daLambda, daHess);
                evaluateH(daX, daLambda, daHess);
            }

            //---- ASSUME THAT PROBLEM EVALUATION IS ALWAYS SUCCESSFUL.
            //---- IF A FUNCTION OR ITS DERIVATIVE COULD NOT BE EVALUATED
            //---- AT THE GIVEN (x, lambda), THEN SET nEvalStatus = 1 BEFORE
            //---- CALLING solve AGAIN.
            nEvalStatus = 0;
        } while (nKnStatus > 0);

        //---- DISPLAY THE RESULTS.
//        System.out.print("KNITRO finished, status " + nKnStatus + ": ");
//        switch (nKnStatus) {
//            case KnitroJava.KTR_RC_OPTIMAL:
//                System.out.println("converged to optimality.");
//                break;
//            case KnitroJava.KTR_RC_ITER_LIMIT:
//                System.out.println("reached the maximum number of allowed iterations.");
//                break;
//            case KnitroJava.KTR_RC_NEAR_OPT:
//            case KnitroJava.KTR_RC_FEAS_XTOL:
//            case KnitroJava.KTR_RC_FEAS_FTOL:
//            case KnitroJava.KTR_RC_FEAS_NO_IMPROVE:
//                System.out.println("could not improve upon the current iterate.");
//                break;
//            case KnitroJava.KTR_RC_TIME_LIMIT:
//                System.out.println("reached the maximum CPU time allowed.");
//                break;
//            default:
//                System.out.println("failed.");
//        }

        //---- EXAMPLES OF OBTAINING SOLUTION INFORMATION.
        // System.out.println("  optimal value = " + daObj[0]);
        daX = solver.getCurrentX();
        daLambda = solver.getCurrentLambda();
        // System.out.println("  solution feasibility violation    = " + solver.getAbsFeasError());
        // System.out.println("           KKT optimality violation = " + solver.getAbsOptError());
        // System.out.println("  number of function evaluations    = " + solver.getNumberFCEvals());

        //---- BE CERTAIN THE NATIVE OBJECT INSTANCE IS DESTROYED.
        solver.destroyInstance();

        for (int i = 0; i < daX.length; i++) {
            guess[i + 1] = daX[i];
        }
    }

    private double evaluateFC(double[] daX, double[] daC) {
        double[] guessX = new double[daX.length + 1];
        for (int i = 0; i < daX.length; i++) {
            guessX[i + 1] = daX[i];
        }

        return f_to_minimize(guessX);
    }

    private void evaluateGA(double[] daX, double[] daObjGrad, double[] daJac) {
        throw new UnsupportedOperationException("KNitro Gradient: Not yet implemented");
    }

    private void evaluateH(double[] daX, double[] daLambda, double[] daHess) {
        throw new UnsupportedOperationException("KNitro Hessian: Not yet implemented");
    }

    double getShockVariance() {
        return shockVariance;
    }
}
