/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package estimation.production;

import estimation.ssrules.plantRecord;
import java.util.concurrent.Callable;
import nervecenter.thesisConstants;
import utility.demandCurve;

/**
 *
 * @author Stephen P. Ryan
 */
class cournotTask implements Callable<Double> {

    private plantRecord plant = null;
    private double nu;
    private double[] capacity;
    private double[] quantity;
    private int numFirms;
    private double capacityCosts;
    private double marginalCosts;
    private double[] productionCostParameters;
    static boolean USE_MR_MC_CONDITION = thesisConstants.USE_MR_MC_CONDITION;
    static boolean IGNORE_CAPACITY_COSTS = thesisConstants.IGNORE_CAPACITY_COSTS;
    private demandCurve priceThesis;

    /**
     * Have to link in market prices to each observation to use that in the
     * estimation.  Should be possible to do without too much pain; let's see
     * how it goes with the calculated price first.
     */
    // static boolean USE_EMPIRICAL_PRICE = true;
    public cournotTask(plantRecord plant, double[] productionCostParameters, demandCurve priceThesis) {
        this.plant = plant;
        this.priceThesis = priceThesis;

        capacityCosts = productionCostParameters[1];
        nu = Math.exp(productionCostParameters[2]) / (1.0 + Math.exp(productionCostParameters[2]));
        marginalCosts = productionCostParameters[3];
        if (plant.getState().contains("CA") && thesisConstants.USE_CA_SHIFTER) {
            marginalCosts += productionCostParameters[4];
        }

        numFirms = plant.getNumFirms();
        capacity = new double[numFirms];
        quantity = new double[numFirms];

        for (int i = 0; i < numFirms; i++) {
            capacity[i] = plant.getCapacity(i);
            quantity[i] = plant.getQuantity(i);
        }
        this.productionCostParameters = productionCostParameters;
    }

    public Double call() throws Exception {
        double error = 0;

        if (USE_MR_MC_CONDITION) {
            double qAll = plant.getMarketQ();
            double priceInside = priceThesis.getPrice(qAll, plant.getStateIndex());
            double b = priceThesis.getB(plant.getStateIndex());
            double pricePrime = priceInside * (1.0 / (b * qAll));

            for (int i = 0; i < numFirms; i++) {
                double mcp = 0;
                double capacityUtilizationRate = quantity[i] / capacity[i];

                double marginalRevenue = priceInside + (quantity[i] * pricePrime);

                if (capacityUtilizationRate > nu &&!IGNORE_CAPACITY_COSTS) {
                    mcp = 2 * capacityCosts * (quantity[i] - (nu * capacity[i]));
                }
                if (capacity[i] > 0) {
                    // error += Math.pow(marginalRevenue - (productionCostParameters[1] + mcp), 2);
                    // error += marginalRevenue - (marginalCosts + mcp);
                    error += Math.pow(marginalRevenue - (marginalCosts + mcp), 2);
//                if (Double.isInfinite(error)) {
//                    System.out.println("*************************");
//                    System.out.print("Capacity: ");
//                    pmUtility.prettyPrint(new Jama.Matrix(capacity, 1));
//                    System.out.print("Quantity: ");
//                    pmUtility.prettyPrint(new Jama.Matrix(quantity, 1));
//                    System.out.println("qi: " + quantity[i] + " capacity: " + capacity[i] + " marketQ: " + qAll + " MR: " + marginalRevenue + " price: " + priceInside + " pricePrime: " + pricePrime + " mcp: " + mcp + " MC:" + (productionCostParameters[3] + mcp) + " util_rate: " + capacityUtilizationRate + " nu: " + nu);
//                    System.exit(0);
//                }
                }
            }
        } else {
            cournotTestbed sc = new cournotTestbed(false, priceThesis);
            sc.getQ(plant, productionCostParameters);
            Jama.Matrix qhat = sc.getQVector();
            // pmUtility.prettyPrintVector(qhat);
            Jama.Matrix actual = plant.getMarketQuantityVector();
            for (int j = 0; j < plant.getNumFirms(); j++) {
                // error += qhat.get(j, 0) - actual.get(j, 0);
                // error += Math.pow(qhat.get(j, 0) - actual.get(j, 0), 2);
                error += Math.abs(qhat.get(j, 0) - actual.get(j, 0));
            }
        }
        return error;
    }
}
