/*
 * cournotTestbed.java
 *
 * Created on March 4, 2004, 3:45 PM
 */
package estimation.production;

import nervecenter.thesisConstants;
import com.ziena.knitro.KnitroJava;
import estimation.ssrules.plantRecord;
import mcmc.gibbsLTEGeneralized;
import optimization.Uncmin_f77;
import utility.*;

/**
 *
 * @author  Stephen P. Ryan
 */
public class cournotTestbed implements mcmc.mcmcFunction, optimization.Uncmin_methods {

    private double[] capacity;
    static private java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
    private int state;
    Jama.Matrix qVector = new Jama.Matrix(1, 1);
    double actualQ = 0;
    double shock = 0;
    int year = thesisConstants.START_YEAR;
    private double capacityCost;
    private double marginalCost;
    private double nu;
    private Jama.Matrix profitVector = null;
    private Jama.Matrix revenues = null;
    private Jama.Matrix costs = null;
    private Jama.Matrix profitMargins = null;
    private double price;
    boolean estimatePost1990Shifters = thesisConstants.USE_LATE_PERIOD_PRODUCTION_COST_SHIFTERS;
    private boolean useCaliforniaMCShifter = thesisConstants.USE_CA_SHIFTER;
    private double COST_ADJUSTMENT = 1.0;
    private boolean approximateQuantities = false;
    static boolean IGNORE_CAPACITY_COSTS = thesisConstants.IGNORE_CAPACITY_COSTS;
    private static double approximationIntercept = 10;
    private static double approximationSlope = 0.864;
    private demandCurve priceThesis;

    /** Creates a new instance of cournotTestbed */
    public cournotTestbed(boolean approximateQuantities, demandCurve priceThesis) {
        this.approximateQuantities = approximateQuantities;
        this.priceThesis = priceThesis;
    }

    public double getQ(double[] capacities, double[] productionParams, int year) {
        this.year = year;

        capacityCost = COST_ADJUSTMENT * Math.max(1E-10, productionParams[1]);
        // capacityCost = productionParams[1];
        double bindingLevel = productionParams[2];
        marginalCost = productionParams[3];

        if (year >= 1990 && estimatePost1990Shifters) {
            capacityCost += COST_ADJUSTMENT * productionParams[4];
            bindingLevel += productionParams[5];
            marginalCost += productionParams[6];
        }

        nu = Math.exp(bindingLevel) / (1.0 + Math.exp(bindingLevel));

        int numFirms = capacities.length;
        capacity = new double[numFirms + 1];

        for (int i = 0; i < numFirms; i++) {
            capacity[i + 1] = capacities[i];
        }
        return execute(numFirms);
    }

    public double getQ(plantRecord plant, double[] productionParams) {
        this.state = plant.getStateIndex();
        this.year = plant.getYear();

        capacityCost = COST_ADJUSTMENT * productionParams[1];
        double bindingLevel = productionParams[2];
        marginalCost = productionParams[3];

        if (year >= 1990 && estimatePost1990Shifters) {
            capacityCost += COST_ADJUSTMENT * productionParams[4];
            bindingLevel += productionParams[5];
            marginalCost += productionParams[6];
        }
        if (useCaliforniaMCShifter) {
            marginalCost += productionParams[4];
        }

        nu = Math.exp(bindingLevel) / (1.0 + Math.exp(bindingLevel));

        int numFirms = plant.getNumFirms();
        capacity = new double[numFirms + 1];

        for (int i = 0; i < numFirms; i++) {
            capacity[i + 1] = plant.getCapacity(i);
        }
        return execute(numFirms);
    }

    private double execute(int numFirms) {
        double guess[] = new double[numFirms + 1];

        for (int i = 0; i < numFirms; i++) {
            guess[i + 1] = 0;
        }

        Uncmin_f77 minimizer = new Uncmin_f77(false);

        int numParams = numFirms;
        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        for (int i = 1; i <= numParams; i++) {
            typsiz[i] = 1;
        }

        double[] fscale = {0, 1E-8};
        int[] method = {0, 1};
        int[] iexp = {0, 0};
        int[] msg = {0, 1};
        int[] ndigit = {0, 10};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-8};
//
//        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
//        System.out.println("Capacity vector:");
//        pmUtility.prettyPrint(new Jama.Matrix(capacity, 1));
        if (approximationSlope > 0) {
            boolean useKNITRO = true;
            if (useKNITRO) {
//            pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                // approximateQuantities = false;
//            DISREGARD_CAPACITY_COSTS = true;
                if (!approximateQuantities) {
                    solveUsingKNitro(guess);
                } else {
                    // what should I put here to approximate?
                    // 87 percent?
                    for(int i=0;i<numFirms;i++) {
                        guess[i + 1] = capacity[i + 1] * 0.87;
                    }
                }
//            pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
//            solveUsingKNitro(guess);
                // pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                // System.exit(0);
            } else {
                // minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);

//        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
//        System.out.println("After uncmin: " + f_to_minimize(guess));

                /*
                 * Set this up so that it searches adaptively for an answer or until the
                 * target level has been met.
                 */
                if (!approximateQuantities) {
                    gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 20, 0, guess, false, false);
                    // gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 1, 0, guess, true, false, 0.1*guess.length);
                }
                // gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 1, 0, guess, false, false);
                // guess = lte.getLowestPoint();

                // minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);

                // System.exit(0);
                // gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 250, 0, guess, false, false);
                // checkSolution(guess);
                // System.out.println("After LTE adjustment: " + f_to_minimize(guess));
                // System.out.println(f_to_minimize(guess));
            }
        } else {
            for (int i = 0; i < numFirms; i++) {
                guess[i + 1] = capacity[i + 1] * 0.87;
            }
        }

        double marketQ = 0;
        for (int i = 1; i < guess.length; i++) {
            marketQ += guess[i];
        }

        // System.out.println("Computing price");
        price = priceThesis.getPrice(marketQ, state);
        if (Double.isInfinite(price)) {
            // System.out.println("Detected infinite price!");
            price = 0;
        }
        // System.out.println("Price should be "+price);
        setPrice(price);

        double penalty = 0;
        qVector = new Jama.Matrix(numFirms, 1);
        profitVector = new Jama.Matrix(numFirms, 1);
        costs = new Jama.Matrix(numFirms, 1);
        revenues = new Jama.Matrix(numFirms, 1);
        profitMargins = new Jama.Matrix(numFirms, 1);

        for (int j = 0; j < numFirms; j++) {
            if (guess[j] > capacity[j] * nu && !IGNORE_CAPACITY_COSTS) {
                penalty = capacityCost * Math.pow(guess[j] - (capacity[j] * nu), 2);
            }
            profitVector.set(j, 0, guess[j + 1] * (price - marginalCost) - penalty);
            revenues.set(j, 0, guess[j + 1] * price);
            costs.set(j, 0, guess[j + 1] * marginalCost + penalty);
            profitMargins.set(j, 0, profitVector.get(j, 0) / revenues.get(j, 0));
            qVector.set(j, 0, guess[j + 1]);
        }

        // System.out.println("MarketQ: " + marketQ + " Price: " + price + " Revenue: " + revenues.get(0, 0) + " Costs: " + costs.get(0, 0));
        // pmUtility.prettyPrintVector(profitVector);

        return guess[1];
    }

    public double getShock() {
        return shock;
    }

    public Jama.Matrix getQVector() {
        return qVector;
    }

    public Jama.Matrix getProfitVector() {
        return profitVector;
    }

    public void setProfitVector(Jama.Matrix profitVector) {
        this.profitVector = profitVector;
    }

    public Jama.Matrix getProfitMargins() {
        return profitMargins;
    }

    public void setProfitMargins(Jama.Matrix profitMargins) {
        this.profitMargins = profitMargins;
    }

    public Jama.Matrix getCosts() {
        return costs;
    }

    public void setCosts(Jama.Matrix costs) {
        this.costs = costs;
    }

    public Jama.Matrix getRevenues() {
        return revenues;
    }

    public void setRevenues(Jama.Matrix revenues) {
        this.revenues = revenues;
    }

    public double getPrice() {
        return price;
    }

    public void setPrice(double price) {
        this.price = price;
    }

    public double f_to_minimize(double[] x) {
        double error = 0;
        double qAll = 0;
//        System.out.println("----");
//        System.out.print("  x:\t");
//        pmUtility.prettyPrint(new Jama.Matrix(x, 1));
//        System.out.print("cap:\t");
//        pmUtility.prettyPrint(new Jama.Matrix(capacity, 1));
//        double[] vecMR = new double[capacity.length];
//        double[] vecMC = new double[capacity.length];
//        double[] vecNu = new double[capacity.length];
        for (int i = 1; i < x.length; i++) {
//            x[i] = Math.max(1E-10, x[i]);
//            x[i] = Math.min(x[i], capacity[i]);
            qAll += x[i];
        }
//        System.out.println("  q:\t" + qAll);
        double priceInside = priceThesis.getPrice(qAll, state);

        double b = priceThesis.getB(state);
        double pricePrime = priceInside * (1.0 / (b * qAll));

        for (int i = 1; i < x.length; i++) {
            double mcp = 0;
            double capacityUtilizationRate = x[i] / capacity[i];

            double marginalRevenue = priceInside + x[i] * pricePrime;

            if (capacityUtilizationRate > nu && !IGNORE_CAPACITY_COSTS) {
                // mcp = 2 * capacityCost * (capacityUtilizationRate - nu) * (1.0 / capacity[i]);
                mcp = 2 * capacityCost * (x[i] - (nu * capacity[i]));
            }

            if (capacity[i] > 0) {
                boolean capacityConstrained = false;
                if (Math.abs(x[i] - capacity[i]) < 1.0) {
                    capacityConstrained = true;
                    // System.out.println("Detecting capacity constrained");
                }
                if (capacityConstrained && marginalRevenue > mcp + marginalCost) {
                    // do nothing
                } else {
                    error += Math.pow(marginalRevenue - (marginalCost + mcp), 2);
                    // error += Math.abs(marginalRevenue - (marginalCost + mcp));
                }
                // error += Math.pow(marginalRevenue - (marginalCost + mcp), 2);
            }
//            vecMR[i] = marginalRevenue;
//            vecMC[i] = marginalCost + mcp;
//            vecNu[i] = nu * capacity[i];
        }
//        System.out.print(" nu:\t");
//        pmUtility.prettyPrint(new Jama.Matrix(vecNu, 1));
//        System.out.print(" MR:\t");
//        pmUtility.prettyPrint(new Jama.Matrix(vecMR, 1));
//        System.out.print(" MC:\t");
//        pmUtility.prettyPrint(new Jama.Matrix(vecMC, 1));
//        System.out.println("f: " + error);
        return error;
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }

    public void gradient(double[] x, double[] g) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void hessian(double[] x, double[][] h) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void solveUsingKNitro(double[] guess) {
        //---- DEFINE THE OPTIMIZATION TEST PROBLEM.
        //---- FOR MORE INFORMATION ABOUT THE PROBLEM DEFINITION, REFER
        //---- TO THE KNITRO MANUAL, ESPECIALLY THE SECTION ON THE
        //---- CALLABLE LIBRARY.
        int n = guess.length - 1;
        int objGoal = KnitroJava.KTR_OBJGOAL_MINIMIZE;
        int objType = KnitroJava.KTR_OBJTYPE_GENERAL;

        // make sure that the problem is defined as unbounded
        double[] bndsLo = new double[n];
        double[] bndsUp = new double[n];
        for (int i = 0; i < n; i++) {
            // bndsLo[i] = -KnitroJava.KTR_INFBOUND;
            // bndsUp[i] = KnitroJava.KTR_INFBOUND;
            bndsLo[i] = 0;
            bndsUp[i] = capacity[i + 1];
        }

        // number of constraints
        int m = 0;
        int[] cType = {};
        double[] cBndsLo = {};
        double[] cBndsUp = {};
        int nnzJ = 0;
        int[] jacIxConstr = {};
        int[] jacIxVar = {};
        int nnzH = 0;
        int[] hessRow = {};
        int[] hessCol = {};

        double[] daXInit = new double[guess.length - 1];
        for (int i = 0; i < guess.length - 1; i++) {
            daXInit[i] = guess[i + 1];
        }


        //---- SETUP AND RUN KNITRO TO SOLVE THE PROBLEM.

        //---- CREATE A SOLVER INSTANCE.
        KnitroJava solver = null;
        try {
            solver = new KnitroJava();
        } catch (java.lang.Exception e) {
            System.err.println(e);
            return;
        }

        // ---- DEMONSTRATE HOW TO SET KNITRO PARAMETERS.
        /**
         * Options: all, summary, none
         */
        if (solver.setCharParamByName("outlev", "none") == false) {
            System.err.println("Error setting parameter 'outlev'");
            return;
        }
//        solver.setIntParamByName("outlev", 0);
//        solver.setIntParamByName("debug", 0);
        if (solver.setDoubleParamByName("feastol", 1.0E-3) == false) {
            System.err.println("Error setting parameter 'feastol'");
            return;
        }
        if (solver.setIntParamByName("gradopt", 2) == false) {
            System.err.println("Error setting parameter 'gradopt'");
            return;
        }
        if (solver.setIntParamByName("hessopt", 6) == false) {
            System.err.println("Error setting parameter 'hessopt'");
            return;
        }

        //---- INITIALIZE KNITRO WITH THE PROBLEM DEFINITION.
        if (solver.initProblem(n, objGoal, objType, bndsLo, bndsUp,
                m, cType, cBndsLo, cBndsUp,
                nnzJ, jacIxVar, jacIxConstr,
                nnzH, hessRow, hessCol,
                daXInit, null) == false) {
            System.err.println("Error initializing the problem, " + "KNITRO status = " + solver.getKnitroStatusCode());
            return;
        }

        //---- ALLOCATE ARRAYS FOR REVERSE COMMUNICATIONS OPERATION.
        double[] daX = new double[n];
        double[] daLambda = new double[m + n];
        double[] daObj = new double[1];
        double[] daC = new double[m];
        double[] daObjGrad = new double[n];
        double[] daJac = new double[nnzJ];
        double[] daHess = new double[nnzH];

        // exampleHS15 thisObject = new exampleHS15();

        //---- SOLVE THE PROBLEM.  IN REVERSE COMMUNICATIONS MODE, KNITRO
        //---- RETURNS WHENEVER IT NEEDS MORE PROBLEM INFORMATION.  THE CALLING
        //---- PROGRAM MUST INTERPRET KNITRO'S RETURN STATUS AND CONTINUE
        //---- SUPPLYING PROBLEM INFORMATION UNTIL KNITRO IS COMPLETE.
        int nKnStatus;
        int nEvalStatus = 0;
        do {
            nKnStatus = solver.solve(nEvalStatus, daObj, daC,
                    daObjGrad, daJac, daHess);
            if (nKnStatus == KnitroJava.KTR_RC_EVALFC) {
                //---- KNITRO WANTS daObj AND daC EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // daObj[0] = thisObject.evaluateFC(daX, daC);
                daObj[0] = evaluateFC(daX, daC);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALGA) {
                //---- KNITRO WANTS daObjGrad AND daJac EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // thisObject.evaluateGA(daX, daObjGrad, daJac);
                evaluateGA(daX, daObjGrad, daJac);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALH) {
                //---- KNITRO WANTS daHess EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                daLambda = solver.getCurrentLambda();
                // thisObject.evaluateH(daX, daLambda, daHess);
                evaluateH(daX, daLambda, daHess);
            }

            //---- ASSUME THAT PROBLEM EVALUATION IS ALWAYS SUCCESSFUL.
            //---- IF A FUNCTION OR ITS DERIVATIVE COULD NOT BE EVALUATED
            //---- AT THE GIVEN (x, lambda), THEN SET nEvalStatus = 1 BEFORE
            //---- CALLING solve AGAIN.
            nEvalStatus = 0;
        } while (nKnStatus > 0);

        //---- DISPLAY THE RESULTS.
//        System.out.print("KNITRO finished, status " + nKnStatus + ": ");
//        switch (nKnStatus) {
//            case KnitroJava.KTR_RC_OPTIMAL:
//                System.out.println("converged to optimality.");
//                break;
//            case KnitroJava.KTR_RC_ITER_LIMIT:
//                System.out.println("reached the maximum number of allowed iterations.");
//                break;
//            case KnitroJava.KTR_RC_NEAR_OPT:
//            case KnitroJava.KTR_RC_FEAS_XTOL:
//            case KnitroJava.KTR_RC_FEAS_FTOL:
//            case KnitroJava.KTR_RC_FEAS_NO_IMPROVE:
//                System.out.println("could not improve upon the current iterate.");
//                break;
//            case KnitroJava.KTR_RC_TIME_LIMIT:
//                System.out.println("reached the maximum CPU time allowed.");
//                break;
//            default:
//                System.out.println("failed.");
//        }

        //---- EXAMPLES OF OBTAINING SOLUTION INFORMATION.
        // System.out.println("  optimal value = " + daObj[0]);
        daX = solver.getCurrentX();
        daLambda = solver.getCurrentLambda();
        // System.out.println("  solution feasibility violation    = " + solver.getAbsFeasError());
        // System.out.println("           KKT optimality violation = " + solver.getAbsOptError());
        // System.out.println("  number of function evaluations    = " + solver.getNumberFCEvals());

        //---- BE CERTAIN THE NATIVE OBJECT INSTANCE IS DESTROYED.
        solver.destroyInstance();

        for (int i = 0; i < daX.length; i++) {
            guess[i + 1] = daX[i];
        }
    }

    private double evaluateFC(double[] daX, double[] daC) {
        // daX is what we usually call guess or x
        // daC is the constraint vector: you fill it it here and that is how knitro knows if something is violated or not
        // daX starts counting at 0 and not 1, so we'll need to move everything over by one and pass it to f_to_minimize
        // this could potentially work out very nicely!
        double[] guessX = new double[daX.length + 1];
        for (int i = 0; i < daX.length; i++) {
            guessX[i + 1] = daX[i];
        }

        // no need to fill daC since I have no constraints

        return f_to_minimize(guessX);
    }

    private void evaluateGA(double[] daX, double[] daObjGrad, double[] daJac) {
        throw new UnsupportedOperationException("KNitro Gradient: Not yet implemented");
    }

    private void evaluateH(double[] daX, double[] daLambda, double[] daHess) {
        throw new UnsupportedOperationException("KNitro Hessian: Not yet implemented");
    }

        /**
     * @return the approximationIntercept
     */
    public static double getApproximationIntercept() {
        return approximationIntercept;
    }

    /**
     * @param aApproximationIntercept the approximationIntercept to set
     */
    public static void setApproximationIntercept(double aApproximationIntercept) {
        approximationIntercept = aApproximationIntercept;
    }

    /**
     * @return the approximationSlope
     */
    public static double getApproximationSlope() {
        return approximationSlope;
    }

    /**
     * @param aApproximationSlope the approximationSlope to set
     */
    public static void setApproximationSlope(double aApproximationSlope) {
        approximationSlope = aApproximationSlope;
    }

}
