/*
 * stageGameCED.java
 *
 * Created on November 17, 2003, 4:47 PM
 */
package estimation.production;

import nervecenter.thesisConstants;
import com.ziena.knitro.KnitroJava;
import estimation.ssrules.plantRecord;
import optimization.*;
import java.io.*;
import javax.swing.*;
import java.util.*;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import utility.*;

/*
 * @author Stephen P. Ryan
 */
public class stageGameCED implements Uncmin_methods, mcmc.mcmcFunction {

    private int numParameters = 3;
    private double[] bestStart;
    private ArrayList<plantRecord> dataMaster = new ArrayList<plantRecord>();
    private ArrayList<plantRecord> data = new ArrayList<plantRecord>();
    private double[] stageGameParameters;
    // private JTextAreaAutoscroll jt = new JTextAreaAutoscroll();
    int lasti = 0;
    double lastxa = 0;
    ExecutorService tpes;
    demandCurve priceThesis;
    // JFrame f;

    /** Creates a new instance of stageGameCED */
    public stageGameCED(demandCurve priceThesis) {
        this.priceThesis = priceThesis;
        tpes = Executors.newFixedThreadPool(thesisConstants.NUM_THREADS);
        if (thesisConstants.USE_LATE_PERIOD_PRODUCTION_COST_SHIFTERS) {
            numParameters = 6;
        }
        if (thesisConstants.USE_CA_SHIFTER) {
            numParameters = 4;
        }
//        f = new JFrame("Production Game Estimation Safety");
//        jt.setEditable(false);
//        f.getContentPane().add(new JScrollPane(jt));
//        f.setBounds(300, 300, 500, 400);
//        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
//        f.setVisible(true);

        stageGameParameters = new double[numParameters + 1];
        bestStart = new double[numParameters + 1];

        try {
            String lateString = "";
            if (thesisConstants.LATE_PERIOD) {
                lateString = "late";
            }
            ObjectInputStream in = new ObjectInputStream(new FileInputStream(new File("data/stageGame" + lateString + ".dat")));
            int numObs = (Integer) in.readObject();

            for (int i = 0; i < numObs; i++) {
                dataMaster.add((plantRecord) in.readObject());
            }
            in.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        double[] guess = new double[numParameters + 1];
        for (int i = 0; i <= numParameters; i++) {
            bestStart[i] = 0;
            guess[i] = bestStart[i];
        }

        System.out.print("Determining starting values...");
        try {
            String late = "";
            if (thesisConstants.LATE_PERIOD) {
                late = "late";
            }
            ObjectInputStream in = new ObjectInputStream(new FileInputStream("data/stageParams" + late + ".dat"));
            for (int i = 0; i < numParameters; i++) {
                guess[i + 1] = in.readDouble();
            }
            in.close();
            System.out.print("loaded from file, using:\n" + pmUtility.stringPrettyPrint(new Jama.Matrix(guess, 1)) + "\n");
        } catch (Exception e) {
            System.out.print("using best guess for starting parameters\n");
            guess[1] = 1.23;
            guess[2] = 1.93;
            guess[3] = 30.113;
        }

        for (int i = 0; i < numParameters; i++) {
            stageGameParameters[i + 1] = guess[i + 1];
        }

        if (thesisConstants.SUBSAMPLE_PARAMETERS) {
            guess[1] = 1.992768;  // 113 with E7
            guess[2] = 1.931648;
            guess[3] = 31.431290;
            if (thesisConstants.USE_CA_SHIFTER) {
                guess[4] = 0;
            }
        }

        optimization.Uncmin_f77 minimizer = new Uncmin_f77(true);

        int numParams = numParameters;
        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        for (int i = 0; i < numParams + 1; i++) {
            typsiz[i] = 1;
        }
        double[] fscale = {0, 1E-8};
        int[] method = {0, 3};
        int[] iexp = {0, 1};
        int[] msg = {0, 1};
        int[] ndigit = {0, 15};
        int[] itnlim = {0, 15};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-8};

        System.out.println("Production costs: intercept = " + priceThesis.getA() + " slope = " + priceThesis.getB(0));

        data = new ArrayList<plantRecord>();
        TreeSet<String> tree = new TreeSet<String>();
        for (int i = 0; i < dataMaster.size(); i++) {
            plantRecord plant = dataMaster.get(i);
            if (!tree.contains(plant.getYear() + plant.getState())) {
                data.add(dataMaster.get(i));
                tree.add(plant.getYear() + plant.getState());
            }
        }

        boolean useKNITRO = false;
        if (useKNITRO) {
            solveUsingKNitro(guess);
        } else {
            boolean uselte = false;
            if (uselte) {
                System.out.print("Calling CH minimizer\n");
                String[] parameterNames = {"Capacity Cost", "Binding Level", "Marginal Cost", "CA Marginal Shifter"};
                // mcmc.gibbsLTEGeneralized lte = new mcmc.gibbsLTEGeneralized(this, 0, 0, guess, parameterNames, true, true);
                mcmc.gibbsLTEGeneralized lte = new mcmc.gibbsLTEGeneralized(this, 3, 0, guess, parameterNames, true, false);
                guess = lte.getLowestPoint();
                // minimizer.optif9_f77(numParameters, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
            } else {
                long t1 = System.currentTimeMillis();
                System.out.println("Using data size: " + data.size());
                System.out.println("Pre-minimization f: " + f_to_minimize(guess));
                // minimizer.optif9_f77(numParameters, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
                System.out.println("Post-minimization f: " + f_to_minimize(guess));
                long t2 = System.currentTimeMillis();
                System.out.println(itrmcd[1] + " -> " + pmUtility.stringPrettyPrint(new Jama.Matrix(guess, 1)) + " f: " + fpls[1] + " t: " + (t2 - t1) / 1000.0 + " s.");
                // System.exit(0);
            }
        }

        // f_to_minimize(guess);

        System.out.print("Writing stage game parameters to vector...\n");
        stageGameParameters = new double[guess.length];
        for (int i = 0; i < guess.length; i++) {
            stageGameParameters[i] = guess[i];
        }
        System.out.print(pmUtility.stringPrettyPrint(new Jama.Matrix(stageGameParameters, 1)) + "\n");

        try {
            String late = "";
            if (thesisConstants.LATE_PERIOD) {
                late = "late";
            }
            ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream("data/stageParams" + late + ".dat"));
            for (int i = 0; i < numParameters; i++) {
                out.writeDouble(guess[i + 1]);
            }
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        boolean writeFits = true;
        double sse = 0;

        System.out.println("Cournot f: " + f_to_minimize(guess));

        double[] guess2 = new double[numParams + 1];
        for (int k = 0; k < numParams; k++) {
            guess2[k + 1] = guess[k + 1];
        }

        if (writeFits && !thesisConstants.SUBSAMPLE_PARAMETERS) {
            ArrayList<Double> capacities = new ArrayList<Double>();
            ArrayList<Double> quantities = new ArrayList<Double>();
            int countCorner = 0;
            int countNotCorner = 0;
            JFrame fitDialog = new JFrame("Quantity Fits");
            fitDialog.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
            fitDialog.setBounds(1200, 200, 500, 500);
            XYSeries xy = new XYSeries("Fit");
            XYSeriesCollection xyc = new XYSeriesCollection(xy);
            JFreeChart chart = ChartFactory.createScatterPlot("Fits", "Q", "Qhat", xyc, PlotOrientation.VERTICAL, false, false, false);
            XYPlot plot = (XYPlot) chart.getPlot();
            plot.getDomainAxis().setRange(0, 2600);
            plot.getRangeAxis().setRange(0, 2600);
            fitDialog.getContentPane().add(new ChartPanel(chart));
            System.out.println("Calculating shocks and writing stage game fits to qfit.csv...");
            ProgressMonitor mon = new ProgressMonitor(fitDialog, "Writing Fits...", "0 / " + data.size(), 0, data.size() - 1);
            try {
                BufferedWriter out = new BufferedWriter(new FileWriter("data/qfit.csv"));
                out.write("qi,qfit,error,numFirms,capacity,profit,margin,costs,revenues,price\n");
                TreeSet<String> tree2 = new TreeSet<String>();

                for (int i = 0; i < data.size(); i++) {
                    plantRecord plant = data.get(i);
                    int numFirms = plant.getNumFirms();
                    String nextMarket = plant.getState() + plant.getYear();
                    if (!tree2.contains(nextMarket)) {
                        // System.out.print("Calling sc...");
                        cournotTestbed sc = new cournotTestbed(false, priceThesis);
                        sc.getQ(plant, guess2);
                        // System.out.println("back");
                        Jama.Matrix qhat = sc.getQVector();
                        Jama.Matrix profitVec = sc.getProfitVector();

                        for (int j = 0; j < numFirms; j++) {
                            double xa = qhat.get(j, 0);
                            double qi = plant.getMarketQuantityVector().get(j, 0);
                            xy.add(qi, xa);
                            double error = xa - qi;
                            double profit = profitVec.get(j, 0);
                            sse += Math.pow(error, 2);
                            quantities.add(qi);
                            capacities.add(plant.getMarketCapacityMatrix().get(j, 0));
                            if (plant.getMarketCapacityMatrix().get(j, 0) == qi) {
                                countCorner++;
                            } else {
                                countNotCorner++;
                            }
                            out.write(qi + "," + xa + "," + error + "," + plant.getNumFirms() + "," + plant.getMarketCapacityMatrix().get(j, 0) + "," + profit + "," + sc.getProfitMargins().get(j, 0) + "," + sc.getCosts().get(j, 0) + "," + sc.getRevenues().get(j, 0) + "," + sc.getPrice() + "\n");
                            // System.out.print(plant.getID()+" "+qi+","+xa+","+error+","+plant.getNumFirms()+","+plant.getAnnualCapacity()+","+profit+","+sc.getProfitMargins().get(j,0)+","+sc.getCosts().get(j,0)+","+sc.getRevenues().get(j,0)+","+sc.getPrice()+"\n");
                        }
                        tree2.add(nextMarket);
                    }
                    mon.setProgress(i + 1);
                    mon.setNote((i + 1) + " / " + data.size());
                }
                fitDialog.setVisible(true);
                System.out.println("sse: " + sse);
                out.close();

                Jama.Matrix quant = new Jama.Matrix(quantities.size(), 1);
                Jama.Matrix cap = new Jama.Matrix(capacities.size(), 1);
                for (int i = 0; i < capacities.size(); i++) {
                    quant.set(i, 0, quantities.get(i));
                    cap.set(i, 0, capacities.get(i));
                }
                // now regress one on the other
                Jama.Matrix beta = pmUtility.OLSsvd(cap, quant, true);
                pmUtility.prettyPrintVector(beta);
                cournotTestbed.setApproximationIntercept(beta.get(0, 0));
                cournotTestbed.setApproximationSlope(beta.get(1, 0));
            } catch (Exception e) {
                e.printStackTrace();
            }
            System.out.println("Finished writing qfit.csv.");
            System.out.println("Corner: " + countCorner + " Not Corner: " + countNotCorner);
        }
        data = dataMaster;

        // f.dispose();
        tpes.shutdown();
        // System.exit(0);
    }

    public ArrayList<plantRecord> getData() {
        return data;
    }

    public double[] getParameters() {
        return stageGameParameters;
    }

    public double f_to_minimize(double[] x) {
        double sumError = 0;

        // long t1 = System.currentTimeMillis();

//        for(int i=0;i<params.length;i++) {
//            params[i] = Math.max(0, params[i]);
//        }

        // in one specification test we:
        // restrict capacity cost shifter in second period to zero
//        for (int i = 0; i < 3; i++) {
//            x[i + 1] = Math.max(0, x[i + 1]);
//        }
        // x[1] = Math.max(0, x[1]);
        // x[3] = Math.max(0, x[3]);

        // fix the threshold
        // x[2] = 1.8;

        TreeSet<String> tree = new TreeSet<String>();

        /**
         * This is a great place to parallelize the calculation of the predicted
         * quantities.
         */
        ArrayList<Future<Double>> futureList = new ArrayList<Future<Double>>();

        for (plantRecord plant : data) {
            String nextMarket = plant.getState() + plant.getYear();
            if (!tree.contains(nextMarket)) {
                futureList.add(tpes.submit(new cournotTask(plant, x, priceThesis)));
                tree.add(nextMarket);
            }
        }

        try {
            for (int ik = 0; ik < futureList.size(); ik++) {
                sumError += futureList.get(ik).get();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        sumError = sumError * sumError;
        sumError /= (2.0 * data.size());

        // long t2 = System.currentTimeMillis();

        // System.out.print((t2 - t1) + " err: " + sumError + " ");
        // pmUtility.prettyPrint(new Jama.Matrix(x, 1));
        return sumError;
    }

    public void gradient(double[] x, double[] g) {
    }

    public void hessian(double[] x, double[][] h) {
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }

    public void solveUsingKNitro(double[] guess) {
        //---- DEFINE THE OPTIMIZATION TEST PROBLEM.
        //---- FOR MORE INFORMATION ABOUT THE PROBLEM DEFINITION, REFER
        //---- TO THE KNITRO MANUAL, ESPECIALLY THE SECTION ON THE
        //---- CALLABLE LIBRARY.
        int n = guess.length - 1;
        int objGoal = KnitroJava.KTR_OBJGOAL_MINIMIZE;
        int objType = KnitroJava.KTR_OBJTYPE_GENERAL;

        // make sure that the problem is defined as unbounded
        double[] bndsLo = new double[n];
        double[] bndsUp = new double[n];
        for (int i = 0; i < n; i++) {
            bndsLo[i] = -KnitroJava.KTR_INFBOUND;
            bndsUp[i] = KnitroJava.KTR_INFBOUND;
        }

        // bound capacity cost and marginal cost to be positive
        // bound marginal cost to be less than price?  doesn't appear to be a worry
        bndsLo[0] = 0;
        bndsLo[2] = 0;

        // number of constraints
        int m = 0;
        int[] cType = {};
        double[] cBndsLo = {};
        double[] cBndsUp = {};
        int nnzJ = 0;
        int[] jacIxConstr = {};
        int[] jacIxVar = {};
        int nnzH = 0;
        int[] hessRow = {};
        int[] hessCol = {};

        double[] daXInit = new double[guess.length - 1];
        for (int i = 0; i < guess.length - 1; i++) {
            daXInit[i] = guess[i + 1];
        }


        //---- SETUP AND RUN KNITRO TO SOLVE THE PROBLEM.

        //---- CREATE A SOLVER INSTANCE.
        KnitroJava solver = null;
        try {
            solver = new KnitroJava();
        } catch (java.lang.Exception e) {
            System.err.println(e);
            return;
        }

        // ---- DEMONSTRATE HOW TO SET KNITRO PARAMETERS.
        /**
         * Options: all, summary, none
         */
        if (solver.setCharParamByName("outlev", "all") == false) {
            System.err.println("Error setting parameter 'outlev'");
            return;
        }
//        solver.setIntParamByName("outlev", 0);
//        solver.setIntParamByName("debug", 0);
        if (solver.setDoubleParamByName("feastol", 1.0E-3) == false) {
            System.err.println("Error setting parameter 'feastol'");
            return;
        }
        if (solver.setIntParamByName("gradopt", 2) == false) {
            System.err.println("Error setting parameter 'gradopt'");
            return;
        }
        if (solver.setIntParamByName("hessopt", 2) == false) {
            System.err.println("Error setting parameter 'hessopt'");
            return;
        }

        //---- INITIALIZE KNITRO WITH THE PROBLEM DEFINITION.
        if (solver.initProblem(n, objGoal, objType, bndsLo, bndsUp,
                m, cType, cBndsLo, cBndsUp,
                nnzJ, jacIxVar, jacIxConstr,
                nnzH, hessRow, hessCol,
                daXInit, null) == false) {
            System.err.println("Error initializing the problem, " + "KNITRO status = " + solver.getKnitroStatusCode());
            return;
        }

        //---- ALLOCATE ARRAYS FOR REVERSE COMMUNICATIONS OPERATION.
        double[] daX = new double[n];
        double[] daLambda = new double[m + n];
        double[] daObj = new double[1];
        double[] daC = new double[m];
        double[] daObjGrad = new double[n];
        double[] daJac = new double[nnzJ];
        double[] daHess = new double[nnzH];

        // exampleHS15 thisObject = new exampleHS15();

        //---- SOLVE THE PROBLEM.  IN REVERSE COMMUNICATIONS MODE, KNITRO
        //---- RETURNS WHENEVER IT NEEDS MORE PROBLEM INFORMATION.  THE CALLING
        //---- PROGRAM MUST INTERPRET KNITRO'S RETURN STATUS AND CONTINUE
        //---- SUPPLYING PROBLEM INFORMATION UNTIL KNITRO IS COMPLETE.
        int nKnStatus;
        int nEvalStatus = 0;
        do {
            nKnStatus = solver.solve(nEvalStatus, daObj, daC,
                    daObjGrad, daJac, daHess);
            if (nKnStatus == KnitroJava.KTR_RC_EVALFC) {
                //---- KNITRO WANTS daObj AND daC EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // daObj[0] = thisObject.evaluateFC(daX, daC);
                daObj[0] = evaluateFC(daX, daC);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALGA) {
                //---- KNITRO WANTS daObjGrad AND daJac EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // thisObject.evaluateGA(daX, daObjGrad, daJac);
                evaluateGA(daX, daObjGrad, daJac);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALH) {
                //---- KNITRO WANTS daHess EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                daLambda = solver.getCurrentLambda();
                // thisObject.evaluateH(daX, daLambda, daHess);
                evaluateH(daX, daLambda, daHess);
            }

            //---- ASSUME THAT PROBLEM EVALUATION IS ALWAYS SUCCESSFUL.
            //---- IF A FUNCTION OR ITS DERIVATIVE COULD NOT BE EVALUATED
            //---- AT THE GIVEN (x, lambda), THEN SET nEvalStatus = 1 BEFORE
            //---- CALLING solve AGAIN.
            nEvalStatus = 0;
        } while (nKnStatus > 0);

        //---- DISPLAY THE RESULTS.
        System.out.print("KNITRO finished, status " + nKnStatus + ": ");
        switch (nKnStatus) {
            case KnitroJava.KTR_RC_OPTIMAL:
                System.out.println("converged to optimality.");
                break;
            case KnitroJava.KTR_RC_ITER_LIMIT:
                System.out.println("reached the maximum number of allowed iterations.");
                break;
            case KnitroJava.KTR_RC_NEAR_OPT:
            case KnitroJava.KTR_RC_FEAS_XTOL:
            case KnitroJava.KTR_RC_FEAS_FTOL:
            case KnitroJava.KTR_RC_FEAS_NO_IMPROVE:
                System.out.println("could not improve upon the current iterate.");
                break;
            case KnitroJava.KTR_RC_TIME_LIMIT:
                System.out.println("reached the maximum CPU time allowed.");
                break;
            default:
                System.out.println("failed.");
        }

        //---- EXAMPLES OF OBTAINING SOLUTION INFORMATION.
        System.out.println("  optimal value = " + daObj[0]);
        daX = solver.getCurrentX();
        daLambda = solver.getCurrentLambda();
        System.out.println("  solution feasibility violation    = " + solver.getAbsFeasError());
        System.out.println("           KKT optimality violation = " + solver.getAbsOptError());
        System.out.println("  number of function evaluations    = " + solver.getNumberFCEvals());

        //---- BE CERTAIN THE NATIVE OBJECT INSTANCE IS DESTROYED.
        solver.destroyInstance();

        for (int i = 0; i < daX.length; i++) {
            guess[i + 1] = daX[i];
        }
    }

    private double evaluateFC(double[] daX, double[] daC) {
        double[] guessX = new double[daX.length + 1];
        for (int i = 0; i < daX.length; i++) {
            guessX[i + 1] = daX[i];
        }

        // no need to fill daC since I have no constraints

        return f_to_minimize(guessX);
    }

    private void evaluateGA(double[] daX, double[] daObjGrad, double[] daJac) {
        throw new UnsupportedOperationException("KNitro Gradient: Not yet implemented");
    }

    private void evaluateH(double[] daX, double[] daLambda, double[] daHess) {
        throw new UnsupportedOperationException("KNitro Hessian: Not yet implemented");
    }
}
