/*
 * To change this template, choose Tools | Templates and open the template in
 * the editor.
 */
package estimation.simulateW;

import JSci.maths.statistics.NormalDistribution;
import Jama.Matrix;
import com.ziena.knitro.KnitroJava;
import java.text.NumberFormat;
import mcmc.gibbsLTEGeneralized;
import optimization.Uncmin_methods;
import utility.bsplineBasisGenerator;

/**
 *
 * @author Stephen P. Ryan
 */
public class InvertDistributionCosts implements Uncmin_methods, mcmc.mcmcFunction {

    private static bsplineBasisGenerator unitIntervalBSpline = new bsplineBasisGenerator(0, 1);
    private Jama.Matrix bSplineCoefficients;
    private boolean isDecreasing = true;
    NumberFormat nf = NumberFormat.getInstance();
    private double[] guess;

    static double[] computeCosts(Matrix investmentBSplineCoefficients, String string, boolean isDecreasing) {
        InvertDistributionCosts idc = new InvertDistributionCosts(investmentBSplineCoefficients, string, isDecreasing);
        return idc.getGuess();
    }

    public InvertDistributionCosts(Jama.Matrix bSplineCoefficients, String distributionDescription, boolean isDecreasing) {
        this.bSplineCoefficients = bSplineCoefficients;
        this.isDecreasing = isDecreasing;

        int numParams = 1;  // just mu and sigma^2 in cost distribution

        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        for (int i = 0; i < numParams + 1; i++) {
            typsiz[i] = 1;
        }
        double[] fscale = {0, 300};
        int[] method = {0, 1};  // 1 line search 2 double dogleg 3 more-hebdon
        int[] iexp = {0, 0}; // 1 if expensive to evaluate
        int[] msg = {0, 0};
        int[] ndigit = {0, 8};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-15};

        guess = new double[numParams + 1];

        boolean useLTE = false;

        guess[1] = gamma(1);

        if (useLTE) {
            gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 150, 150, guess);
            guess = lte.getLowestPoint();
        } else {
            boolean useKNITRO = true;
            if (useKNITRO) {
                solveUsingKNitro(guess);
            } else {
                optimization.Uncmin_f77 minimizer = new optimization.Uncmin_f77(false);
                minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
            }
        }
        // System.out.println("Distribution of " + distributionDescription + " mu: " + gamma(1) + " sigma: " + guess[1]);
    }

    public double f_to_minimize(double[] x) {
        double value = 0;

        // keep the variance positive
        x[1] = Math.max(1E-15, x[1]);

        double mu = gamma(1);
        double sigma = x[1];

        NormalDistribution stdNormal = new NormalDistribution();

        double start = 0.001;
        double end = 0.999;

        for (double prob = start; prob <= end; prob += 0.01) {
            // match the expected value instead
            if (isDecreasing) {
                double d = mu + sigma * stdNormal.inverse(1 - prob);
                double z = (d - mu) / sigma;
                double expectedTruncatedValue = mu + (sigma * (stdNormal.probability(z) / (1.0 - stdNormal.cumulative(z))));
//                System.out.print("prob: " + nf.format(prob) + "\t sigma: " + nf.format(sigma) + "\t etv: " + nf.format(expectedTruncatedValue) + "\t gamma: " + nf.format(gamma(prob)));
//                System.out.println("\t mu: " + nf.format(mu) + "\t d: " + nf.format(d) + "\t z: " + nf.format(z) + "\t pdf(z): " + nf.format(stdNormal.probability(z)) + "\t cdf(z): " + nf.format(stdNormal.cumulative(z)) + " ");
                value += Math.pow(expectedTruncatedValue - gamma(prob), 2);
            } else {
                double d = mu + sigma * stdNormal.inverse(prob);
                double z = (d - mu) / sigma;
                double expectedTruncatedValue = mu - ((stdNormal.probability(z) / stdNormal.cumulative(z)) * sigma);
                // System.out.println("--------------");
                // System.out.println("mu: "+mu+" z: "+z+" pdf(z): "+stdNormal.probability(z)+" cdf(z): "+stdNormal.cumulative(z));
                // System.out.println("prob: "+prob+" sigma: "+sigma+" etv: "+expectedTruncatedValue);
                value += Math.pow(expectedTruncatedValue - gamma(prob), 2);
            }
        }
        return value;
    }

    private double gamma(double prob) {
        // return the estimate of gamma given prob here
        Jama.Matrix basis = unitIntervalBSpline.getBasis(prob);
        return basis.times(bSplineCoefficients).get(0, 0);
    }

    public void gradient(double[] x, double[] g) {
    }

    public void hessian(double[] x, double[][] h) {
    }

    public double[] getGuess() {
        double[] g = new double[3];
        g[0] = 0;
        g[1] = gamma(1);
        g[2] = guess[1];
        return g;
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }

    public void solveUsingKNitro(double[] guess) {
        //---- DEFINE THE OPTIMIZATION TEST PROBLEM.
        //---- FOR MORE INFORMATION ABOUT THE PROBLEM DEFINITION, REFER
        //---- TO THE KNITRO MANUAL, ESPECIALLY THE SECTION ON THE
        //---- CALLABLE LIBRARY.
        int n = guess.length - 1;
        int objGoal = KnitroJava.KTR_OBJGOAL_MINIMIZE;
        int objType = KnitroJava.KTR_OBJTYPE_GENERAL;

        // make sure that the problem is defined as unbounded
        double[] bndsLo = new double[n];
        double[] bndsUp = new double[n];
        for (int i = 0; i < n; i++) {
            // bndsLo[i] = -KnitroJava.KTR_INFBOUND;
            bndsLo[i] = 0;
            bndsUp[i] = KnitroJava.KTR_INFBOUND;
        }

        // number of constraints
        int m = 0;
        int[] cType = {};
        double[] cBndsLo = {};
        double[] cBndsUp = {};
        int nnzJ = 0;
        int[] jacIxConstr = {};
        int[] jacIxVar = {};
        int nnzH = 0;
        int[] hessRow = {};
        int[] hessCol = {};

        double[] daXInit = new double[guess.length - 1];
        for (int i = 0; i < guess.length - 1; i++) {
            daXInit[i] = guess[i + 1];
        }


        //---- SETUP AND RUN KNITRO TO SOLVE THE PROBLEM.

        //---- CREATE A SOLVER INSTANCE.
        KnitroJava solver = null;
        try {
            solver = new KnitroJava();
        } catch (java.lang.Exception e) {
            System.err.println(e);
            return;
        }

        // ---- DEMONSTRATE HOW TO SET KNITRO PARAMETERS.
        /**
         * Options: all, summary, none
         */
        if (solver.setCharParamByName("outlev", "none") == false) {
            System.err.println("Error setting parameter 'outlev'");
            return;
        }
//        solver.setIntParamByName("outlev", 0);
//        solver.setIntParamByName("debug", 0);
        if (solver.setDoubleParamByName("feastol", 1.0E-6) == false) {
            System.err.println("Error setting parameter 'feastol'");
            return;
        }
        if (solver.setIntParamByName("gradopt", 2) == false) {
            System.err.println("Error setting parameter 'gradopt'");
            return;
        }
        if (solver.setIntParamByName("hessopt", 6) == false) {
            System.err.println("Error setting parameter 'hessopt'");
            return;
        }

        //---- INITIALIZE KNITRO WITH THE PROBLEM DEFINITION.
        if (solver.initProblem(n, objGoal, objType, bndsLo, bndsUp,
                m, cType, cBndsLo, cBndsUp,
                nnzJ, jacIxVar, jacIxConstr,
                nnzH, hessRow, hessCol,
                daXInit, null) == false) {
            System.err.println("Error initializing the problem, " + "KNITRO status = " + solver.getKnitroStatusCode());
            return;
        }

        //---- ALLOCATE ARRAYS FOR REVERSE COMMUNICATIONS OPERATION.
        double[] daX = new double[n];
        double[] daLambda = new double[m + n];
        double[] daObj = new double[1];
        double[] daC = new double[m];
        double[] daObjGrad = new double[n];
        double[] daJac = new double[nnzJ];
        double[] daHess = new double[nnzH];

        // exampleHS15 thisObject = new exampleHS15();

        //---- SOLVE THE PROBLEM.  IN REVERSE COMMUNICATIONS MODE, KNITRO
        //---- RETURNS WHENEVER IT NEEDS MORE PROBLEM INFORMATION.  THE CALLING
        //---- PROGRAM MUST INTERPRET KNITRO'S RETURN STATUS AND CONTINUE
        //---- SUPPLYING PROBLEM INFORMATION UNTIL KNITRO IS COMPLETE.
        int nKnStatus;
        int nEvalStatus = 0;
        do {
            nKnStatus = solver.solve(nEvalStatus, daObj, daC,
                    daObjGrad, daJac, daHess);
            if (nKnStatus == KnitroJava.KTR_RC_EVALFC) {
                //---- KNITRO WANTS daObj AND daC EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // daObj[0] = thisObject.evaluateFC(daX, daC);
                daObj[0] = evaluateFC(daX, daC);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALGA) {
                //---- KNITRO WANTS daObjGrad AND daJac EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // thisObject.evaluateGA(daX, daObjGrad, daJac);
                evaluateGA(daX, daObjGrad, daJac);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALH) {
                //---- KNITRO WANTS daHess EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                daLambda = solver.getCurrentLambda();
                // thisObject.evaluateH(daX, daLambda, daHess);
                evaluateH(daX, daLambda, daHess);
            }

            //---- ASSUME THAT PROBLEM EVALUATION IS ALWAYS SUCCESSFUL.
            //---- IF A FUNCTION OR ITS DERIVATIVE COULD NOT BE EVALUATED
            //---- AT THE GIVEN (x, lambda), THEN SET nEvalStatus = 1 BEFORE
            //---- CALLING solve AGAIN.
            nEvalStatus = 0;
        } while (nKnStatus > 0);

        //---- DISPLAY THE RESULTS.
//        System.out.print("KNITRO finished, status " + nKnStatus + ": ");
//        switch (nKnStatus) {
//            case KnitroJava.KTR_RC_OPTIMAL:
//                System.out.println("converged to optimality.");
//                break;
//            case KnitroJava.KTR_RC_ITER_LIMIT:
//                System.out.println("reached the maximum number of allowed iterations.");
//                break;
//            case KnitroJava.KTR_RC_NEAR_OPT:
//            case KnitroJava.KTR_RC_FEAS_XTOL:
//            case KnitroJava.KTR_RC_FEAS_FTOL:
//            case KnitroJava.KTR_RC_FEAS_NO_IMPROVE:
//                System.out.println("could not improve upon the current iterate.");
//                break;
//            case KnitroJava.KTR_RC_TIME_LIMIT:
//                System.out.println("reached the maximum CPU time allowed.");
//                break;
//            default:
//                System.out.println("failed.");
//        }

        //---- EXAMPLES OF OBTAINING SOLUTION INFORMATION.
        // System.out.println("  optimal value = " + daObj[0]);
        daX = solver.getCurrentX();
        daLambda = solver.getCurrentLambda();
        // System.out.println("  solution feasibility violation    = " + solver.getAbsFeasError());
        // System.out.println("           KKT optimality violation = " + solver.getAbsOptError());
        // System.out.println("  number of function evaluations    = " + solver.getNumberFCEvals());

        //---- BE CERTAIN THE NATIVE OBJECT INSTANCE IS DESTROYED.
        solver.destroyInstance();

        for (int i = 0; i < daX.length; i++) {
            guess[i + 1] = daX[i];
        }
    }

    private double evaluateFC(double[] daX, double[] daC) {
        // daX is what we usually call guess or x
        // daC is the constraint vector: you fill it it here and that is how knitro knows if something is violated or not
        // daX starts counting at 0 and not 1, so we'll need to move everything over by one and pass it to f_to_minimize
        // this could potentially work out very nicely!
        double[] guessX = new double[daX.length + 1];
        for (int i = 0; i < daX.length; i++) {
            guessX[i + 1] = daX[i];
        }

        // no need to fill daC since I have no constraints

        return f_to_minimize(guessX);
    }

    private void evaluateGA(double[] daX, double[] daObjGrad, double[] daJac) {
        throw new UnsupportedOperationException("KNitro Gradient: Not yet implemented");
    }

    private void evaluateH(double[] daX, double[] daLambda, double[] daHess) {
        throw new UnsupportedOperationException("KNitro Hessian: Not yet implemented");
    }
}
