/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package estimation.simulateW;

import estimation.ssrulesLikelihood.investmentParametersContainer;
import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.Callable;
import nervecenter.thesisConstants;
import utility.bsplineBasisGenerator;
import utility.demandCurve;

/**
 *
 * @author Stephen P. Ryan
 */
public class SimulateWTask implements Callable<ArrayList<Jama.Matrix>> {

    private int numSimW;
    private int numParams;
    private double[] productionParameters;
    private investmentParametersContainer policyParameters;
    private int simulationHorizon;
    private static bsplineBasisGenerator unitIntervalBSpline = new bsplineBasisGenerator(0, 1);
    private boolean fixedCostHeterogeneity;
    private Random rng;
    private final double[] entryParameters;
    private final double[] exitParameters;
    private final demandCurve priceThesis;

    public SimulateWTask(long seed, int numSimW, int numParams, int simulationHorizon, boolean fixedCostHeterogeneity, demandCurve priceThesis, double[] productionParameters, investmentParametersContainer policyParameters,
            double[] exitParameters, double[] entryParameters) {
        this.numSimW = numSimW;
        this.numParams = numParams;
        this.priceThesis = priceThesis;
        this.productionParameters = productionParameters;
        this.policyParameters = policyParameters;
        this.simulationHorizon = simulationHorizon;
        this.fixedCostHeterogeneity = fixedCostHeterogeneity;
        this.entryParameters = entryParameters;
        this.exitParameters = exitParameters;
        rng = new Random(seed);
    }

    public ArrayList<Jama.Matrix> call() throws Exception {
        ArrayList<Jama.Matrix> wList = new ArrayList<Jama.Matrix>();
        // try Alabama-sized market simulation
        // int sNumFirms = 4; //
        int sNumFirms = 2; // (int) Math.floor(5 * rng.nextDouble()) + 1;
        double[] cap = new double[sNumFirms];
        for (int cn = 0; cn < 2; cn++) {
            cap[cn] = 750; // 250 + 1500 * rng.nextDouble();
            if (rng.nextDouble() < 0.25 && cn > 0) {
                // cap[cn] = 0;
            }
        }


        long simulateRNGBase = rng.nextLong();

        for (int k = 0; k < 2; k++) {
            // want to hold the RNG draws fixed across the two policies
            rng = new Random(simulateRNGBase);
            // System.out.println("k = "+k);
            double[] W = new double[numParams + 1];

            // System.out.println("Initializing state");
            State baseState = new State(rng.nextLong(), priceThesis, productionParameters, policyParameters, exitParameters, entryParameters);
            // System.out.println("State initialized");
            baseState.setCapacitiesInit(cap);
            // can test the effect of the number of simulation draws on the expectation here
            // for(numSimW=16;numSimW<=4096;numSimW*=2) {
            // need at least 2000
            for (int ci = 0; ci <= numParams; ci++) {
                W[ci] = 0;
            }
            // System.out.println("Entering numSimW loop");
            for (int j = 0; j < numSimW; j++) {
//                State s = null;
//                try {
//                    s = (State) baseState.clone();
//                } catch (Exception e) {
//                    e.printStackTrace();
//                }
                State s = baseState.getDeepCopy();
                s.refreshRNG(rng.nextLong());
                if (k == 1) {
                    s.perturb();  // make this run a permutation on the optimal policy functions observed in the data
                }
                double[] Win = new double[numParams + 1];
                for (int ci = 0; ci <= numParams; ci++) {
                    Win[ci] = 0;
                }
                // System.out.println("s.getProduction()");
                // s.getProduction(); // initializes the quantities and profits
                for (int i = 0; i < simulationHorizon; i++) {
                    // System.out.println("Inside simulationHorizon = "+i);
                    double discount = Math.pow(thesisConstants.DISCOUNT_FACTOR, i);
                    // System.out.println("s.getPolicyFirst()");
                    double policy = s.getPolicyFirst();
                    double probabilityExit = s.getProbabilityExit();
//                    if(policy>1500) {
//                        System.out.println("k = "+k+" policy: "+policy);
//                    }
                    // artificially set this to zero
                    double probabilityDivestment = s.getProbabilityDivestment();
                    double probabilityInvestment = s.getProbabilityInvestment();
                    if (k == 0) {
                        // System.out.println(i+" k: "+k+" cap: "+s.getCapacity(0)+" policy: "+policy+" prexit: "+probabilityExit+" prinv: "+probabilityInvestment+" prdiv: "+probabilityDivestment);
                    }

                    double profit = s.getProfit(true);
                    // System.out.println("s.getProfit()");
                    Win[0] += discount * profit;
                    if (policy > 0) {
                        if (fixedCostHeterogeneity) {
                            Jama.Matrix basis = unitIntervalBSpline.getBasis(probabilityInvestment);
                            for (int dm = 0; dm < 6; dm++) {
                                Win[dm + 1] -= discount * basis.get(0, dm);
                            }
                            Win[7] -= discount * (policy);
                            Win[8] -= discount * (policy * policy);
                        } else {
                            Win[1] -= discount;
                            Win[2] -= discount * policy;
                            Win[3] -= discount * policy * policy;
                        }
                    }
                    // System.out.println("s.getExit()");
                    if (policy < 0 && !s.getExit()) {
                        // also for negative adjustments
                        if (fixedCostHeterogeneity) {
                            Jama.Matrix basis = unitIntervalBSpline.getBasis(probabilityDivestment);
                            for (int dm = 0; dm < 6; dm++) {
                                Win[dm + 9] -= discount * basis.get(0, dm);
                            }
                            Win[15] -= discount * policy;
                            Win[16] -= discount * policy * policy;
                        } else {
                            Win[4] -= discount;
                            Win[5] -= discount * policy;
                            Win[6] -= discount * policy * policy;
                        }
                    }

                    double c0 = s.getCapacity(0);
                    s.updateState(policy);

                    if (s.getExit()) {
                        if (fixedCostHeterogeneity) {
                            Jama.Matrix basis = unitIntervalBSpline.getBasis(probabilityExit);
                            for (int dm = 0; dm < 6; dm++) {
                                Win[dm + 17] += discount * basis.get(0, dm);
                            }
                            Win[17] += discount;
                        } else {
                            Win[7] += discount;
                            Win[4] -= discount;
                            Win[5] -= discount * c0;
                            Win[6] -= discount * c0 * c0;
                        }
                        i = simulationHorizon + 1; // if you exit, you are done
                    }
                }
                for (int ci = 0; ci <= numParams; ci++) {
                    W[ci] += Win[ci];
                }
            }
            for (int ci = 0; ci <= numParams; ci++) {
                W[ci] /= (numSimW + 0.0);
            }
            wList.add(new Jama.Matrix(W, 1));
        }
        return wList;
    }
}
