/*
 * state.java
 *
 * Created on January 27, 2004, 9:03 PM
 */
package estimation.simulateW;

import JSci.maths.statistics.*;
import estimation.ssrulesLikelihood.investmentParametersContainer;
import nervecenter.thesisConstants;
import java.text.NumberFormat;
import java.util.Random;
import utility.*;

/**
 *
 * @author  Stephen P. Ryan
 */
public class State implements Cloneable {

    private static boolean latePeriod = thesisConstants.LATE_PERIOD;
    private static java.text.NumberFormat format = java.text.NumberFormat.getInstance();
    private static double[] productionParameters;
    private double[] capacities;
    private double marketQ = 0;
    public boolean perturbPolicies = false;
    private bsplineBasisGenerator b1;
    private bsplineBasisGenerator b2;
    private Jama.Matrix betaBand;
    private Jama.Matrix betaTarget;
    public static double entryLevel = 1.234E-10;
    private boolean exited = false;
    private static NormalDistribution normal = new NormalDistribution();
    private double iperturbtarget;
    private double iperturbband;
    private double exitperturb;
    private Jama.Matrix qEveryone;
    private NormalDistribution errorBand;
    private NormalDistribution errorTarget;
    private double profit = 0;
    private double probabilityInvestment;
    private double probabilityDivestment;
    private double probabilityExit;
    private Random rng;
    private long seed2;
    private long seed;
    static boolean verbose = false;
    static NumberFormat nf = NumberFormat.getInstance();
    boolean turnOffDivestment = false;
    private double ENTRY_POST_1990_SHIFTER;
    private double ENTRY_INTERCEPT;
    private double ENTRY_INCUMBENT_CAPACITY_COEFFICIENT;
    private double EXIT_POST_1990_SHIFTER;
    private double EXIT_INTERCEPT;
    private double EXIT_OWN_CAPACITY_COEFFICIENT;
    private double EXIT_COMPETITORS_CAPACITY_COEFFICIENT;
    private double[] exitParameters;
    private double[] entryParameters;
    Jama.Matrix demandParameters;
    demandCurve priceThesis;
    /**
     * Set this to 0 if interpreting the variance in the target/band equations
     * as measurement error, 1 if interpreting as structural errors.
     */
    double varianceTargetAdjustment = 0;
    double varianceBandAdjustment = 1;
    private final investmentParametersContainer policyParameters;

    /** Creates a new instance of state */
    public State(long seed, demandCurve priceThesis, double[] capacityCosts, investmentParametersContainer policyParameters,
            double[] exitParameters, double[] entryParameters) {
        b1 = policyParameters.getB1();
        b2 = policyParameters.getB2();
        betaBand = policyParameters.getBetaBand();
        betaTarget = policyParameters.getBetaTarget();
        this.seed = seed;
        this.policyParameters = policyParameters;
        productionParameters = capacityCosts;
        errorBand = new NormalDistribution(0, policyParameters.getVarianceBandEquation());
        errorTarget = new NormalDistribution(0, policyParameters.getVarianceTargetEquation());
        rng = new Random(seed);
        seed2 = rng.nextLong();
        iperturbtarget = normal.inverse(rng.nextDouble());
        iperturbband = normal.inverse(rng.nextDouble());
        exitperturb = normal.inverse(rng.nextDouble());
        this.entryParameters = entryParameters;
        this.exitParameters = exitParameters;
        this.priceThesis = priceThesis;

        EXIT_INTERCEPT = exitParameters[1];
        EXIT_POST_1990_SHIFTER = exitParameters[2];
        EXIT_OWN_CAPACITY_COEFFICIENT = exitParameters[3];
        EXIT_COMPETITORS_CAPACITY_COEFFICIENT = exitParameters[4];

        ENTRY_INTERCEPT = entryParameters[1];
        ENTRY_POST_1990_SHIFTER = entryParameters[2];
        ENTRY_INCUMBENT_CAPACITY_COEFFICIENT = entryParameters[3];

    }

    public static boolean getLatePeriod() {
        return latePeriod;
    }

    public State(long seed, demandCurve priceThesisPass, double[] capacityCosts, investmentParametersContainer policyParametersPass, double marketQPass, double[] exitParametersPass, double[] entryParametersPass) {
        rng = new Random(seed);
        marketQ = marketQPass;
        policyParameters = policyParametersPass;
        this.b1 = policyParametersPass.getB1();
        this.seed = seed;
        this.b2 = policyParametersPass.getB2();
        this.betaBand = policyParametersPass.getBetaBand();
        this.betaTarget = policyParametersPass.getBetaTarget();
        productionParameters = capacityCosts;
        priceThesis = priceThesisPass;
        errorBand = new NormalDistribution(0, policyParametersPass.getVarianceBandEquation());
        errorTarget = new NormalDistribution(0, policyParametersPass.getVarianceTargetEquation());

        this.entryParameters = entryParametersPass;
        this.exitParameters = exitParametersPass;

        EXIT_INTERCEPT = exitParametersPass[1];
        EXIT_POST_1990_SHIFTER = exitParametersPass[2];
        EXIT_OWN_CAPACITY_COEFFICIENT = exitParametersPass[3];
        EXIT_COMPETITORS_CAPACITY_COEFFICIENT = exitParametersPass[4];

        ENTRY_INTERCEPT = entryParametersPass[1];
        ENTRY_POST_1990_SHIFTER = entryParametersPass[2];
        ENTRY_INCUMBENT_CAPACITY_COEFFICIENT = entryParametersPass[3];
    }

    public Jama.Matrix getState() {
        return new Jama.Matrix(capacities, 1);
    }

    public void perturb() {
        perturbPolicies = true;
        Random rng2 = new Random(seed2);
        if (rng2.nextDouble() < 0.5) {
            // perturb threshold for investment
            iperturbband = 0.025 * normal.inverse(rng2.nextDouble());
        } else {
            iperturbtarget = 0.025 * normal.inverse(rng2.nextDouble());
        }
    }

    public double getPolicyFirst() {
        double value = 0;
        double x = getCapacity(0);
        double capOther = 0;
        for (int i = 1; i < capacities.length; i++) {
            capOther += capacities[i];
        }
        if (x == 0) {
            // firms that have exited have no investment
            return 0;
        }

        // find basis for capother, owncap, and shocks for particular configuration here
        if (verbose) {
            System.out.println("Using own capacity of: " + x + " and competitor capacity of: " + capOther + " to compute bands and targets.");
        }
        Jama.Matrix basisX = b1.getBasis(0 * capOther);
        Jama.Matrix basisOwnCap = b2.getBasis(x);
        basisX = pmUtility.concatMatrixLOO(basisX, basisOwnCap);

        if (capOther < 5 && verbose) {
            pmUtility.prettyPrintVector(betaTarget);
            pmUtility.prettyPrint(basisX);
        }
        // interpret the error here as measurement error, not structural error (so set to zero)
        double target = basisX.times(betaTarget).get(0, 0) + varianceTargetAdjustment * errorTarget.inverse(rng.nextDouble());
        if (perturbPolicies) {
            target += iperturbtarget;
        }
        target = Math.exp(target);

        double band = 0;

        // System.out.println(band);
        if (perturbPolicies) {
            band = Math.exp(basisX.times(betaBand).get(0, 0) + varianceBandAdjustment * errorBand.inverse(rng.nextDouble()) + iperturbband);
        } else {
            band = Math.exp(basisX.times(betaBand).get(0, 0) + varianceBandAdjustment * errorBand.inverse(rng.nextDouble()));
        }

        // can compute and report probabilities of investment and divestment here
        boolean reportInvestmentProbability = false;
        if (reportInvestmentProbability) {
            double gapMean = Math.exp(basisX.times(betaBand).get(0, 0));
            NormalDistribution bandDist = new NormalDistribution(0, errorBand.getVariance());
            double prInvZ = x - target - gapMean;
            double prDivZ = x - target + gapMean;
            double prInv = bandDist.cumulative(prInvZ);
            double prDiv = 1.0 - bandDist.cumulative(prDivZ);
            if (prInv > 0 && prDiv > 0) {
                System.out.println("-------------------");
                System.out.println("x: " + x + " gap: " + gapMean + " target: " + target);
                System.out.println("PrInv: " + prInv + " prDiv: " + prDiv);
            }
        }

        if (verbose) {
            pmUtility.prettyPrint(basisX);
            System.out.print("Current: " + nf.format(x) + " Target: " + nf.format(target) + " change: " + nf.format(target - x) + " Band: " + nf.format(band));
        }
        boolean nothing = true;
        if (x < target - band) {
            if (verbose) {
                System.out.println(" ---> Making an investment");
            }
            nothing = false;
        }
        if (x > target + band) {
            if (verbose) {
                System.out.println(" ---> Making a divestment");
            }
            nothing = false;
        }
        if (nothing) {
            if (verbose) {
                System.out.println(" ---> Nothing happens");
            }
        }

        // System.out.println(x+" target: "+target+" band: "+band);
        if (x > target + band) {
            // downward adjustment
            value = target - x;
            if (turnOffDivestment) {
                value = 0;
            }
        }
        if (x < target - band) {
            // upward adjustment
            value = target - x;
        }

        // target - band = threshold
        // x + epsilon < threshold -> investment
        // prob(epsilon < threshold - x) -> prob(investment)
        probabilityInvestment = errorBand.cumulative(target - band - x);
        // x + epsilon > threshold -> divestment
        // prob(epsilon > threshold - x) -> prob(divestment)
        probabilityDivestment = 1.0 - errorBand.cumulative(target + band - x);
        if (turnOffDivestment) {
            probabilityDivestment = 0;
        }
        // want to adjust this so that if we are a new entrant then we adjust to our desired capacity
        if (x == entryLevel) {
            value = target;
            probabilityInvestment = 1;
        }

        return value;
    }

    @Override
    public String toString() {
        String result = "";
        for (int i = 0; i < getNumFirms(); i++) {
            result = result.concat(format.format(capacities[i]) + " ");
        }
        return result;
    }

    public void updateState(double myPolicy) {
        double marketCapacity = 0;
        for (int i = 0; i < getNumFirms(); i++) {
            marketCapacity += capacities[i];
        }

        double[] plannedPolicy = new double[capacities.length];

        // want to figure out if there are any entries before determining investment
        for (int i = 1; i < getNumFirms(); i++) {
            double periodShifter = ENTRY_POST_1990_SHIFTER;
            if (!latePeriod || thesisConstants.FORCE_EARLY_DYNAMIC_PARAMETERS) {
                periodShifter = 0;
            }
            if (capacities[i] == 0) {
                double prentry = normal.cumulative(periodShifter + ENTRY_INTERCEPT + ENTRY_INCUMBENT_CAPACITY_COEFFICIENT * marketCapacity);
                if (rng.nextDouble() < prentry) {
                    capacities[i] = entryLevel;
                }
            }
        }

        for (int i = 1; i < getNumFirms(); i++) {
            double[] change = new double[capacities.length];
            for (int j = 0; j < capacities.length; j++) {
                change[j] = capacities[j];
            }
            change[0] = capacities[i];
            change[i] = capacities[0];
            if (verbose) {
                System.out.println("swapping out");
            }
            State swap = new State(seed, priceThesis, productionParameters, policyParameters, marketQ, exitParameters, entryParameters);
            swap.setCapacities(change);
            plannedPolicy[i] = swap.getPolicyFirst();
        }
        plannedPolicy[0] = myPolicy;
        if (verbose) {
            System.out.println("--------------- UPDATE STATE " + perturbPolicies + " ---------------");
        }
        // pmUtility.prettyPrint(new Jama.Matrix(plannedPolicy, 1));

        // put in entry and exit here
        double shifter = 0;
        if (perturbPolicies) {
            shifter = exitperturb;
        }
        for (int i = 0; i < getNumFirms(); i++) {
            if (i > 0) {
                shifter = 0;
            }
            double periodShifter = EXIT_POST_1990_SHIFTER;
            if (!latePeriod || thesisConstants.FORCE_EARLY_DYNAMIC_PARAMETERS) {
                periodShifter = 0;
            }
            if (capacities[i] > 1.0) {
                double firmProbabilityExit = normal.cumulative(periodShifter + shifter + EXIT_INTERCEPT + EXIT_OWN_CAPACITY_COEFFICIENT * capacities[i]
                        + EXIT_COMPETITORS_CAPACITY_COEFFICIENT * (marketCapacity - capacities[i]));
                if (i == 0) {
                    probabilityExit = firmProbabilityExit;
                }
                
                if (rng.nextDouble() < firmProbabilityExit) {
                    capacities[i] = 0;
                    if (verbose) {
                        System.out.println("**************************************** Firm " + i + " exited ****************************************");
                    }
                    if (i == 0) {
                        exited = true;
                    }
                }
            }
        }

        // update the state vector
        for (int i = 0; i < getNumFirms(); i++) {
            if (capacities[i] > 0) {
                capacities[i] += plannedPolicy[i];
            }
            if (capacities[i] < 0) {
                capacities[i] = 0;
            }
        }
    }

    public boolean getExit() {
        return exited;
    }

    private double getProduction(boolean approximateQuantities) {
        double qi = 0;
        if (capacities[0] == 0) {
            return 0;
        }
        estimation.production.cournotTestbed qGen = new estimation.production.cournotTestbed(approximateQuantities, priceThesis);
        if (latePeriod) {
            qi = qGen.getQ(capacities, productionParameters, 1990);
        } else {
            qi = qGen.getQ(capacities, productionParameters, thesisConstants.START_YEAR);
        }
        qEveryone = qGen.getQVector();
        profit = qGen.getProfitVector().get(0, 0);
        marketQ = 0;
        for (int i = 0; i < capacities.length; i++) {
            marketQ += qEveryone.get(i, 0);
        }
        // pmUtility.prettyPrintVector(qEveryone);
        // pmUtility.prettyPrintVector(qGen.getProfitVector());
        return Math.max(0, qi);
    }

    public int getNumFirms() {
        return capacities.length;
    }

    public void setCapacities(double[] cap) {
        capacities = new double[cap.length];
        for (int i = 0; i < cap.length; i++) {
            capacities[i] = cap[i];
        }
        // System.out.println("Set first firm to "+capacities[0]);
    }

    public void setCapacitiesInit(double[] cap) {
        capacities = new double[cap.length];
        for (int i = 0; i < cap.length; i++) {
            capacities[i] = cap[i];
        }
        // System.out.println("Initialized first firm to "+capacities[0]);

        estimation.production.cournotTestbed qGen = new estimation.production.cournotTestbed(true, priceThesis);
        double qi = 0;
        if (latePeriod) {
            qi = qGen.getQ(capacities, productionParameters, 1990);
        } else {
            qi = qGen.getQ(capacities, productionParameters, thesisConstants.START_YEAR);
        }
        qEveryone = qGen.getQVector();
        profit = qGen.getProfitVector().get(0, 0);
        // pmUtility.prettyPrintVector(qEveryone);
        // pmUtility.prettyPrintVector(qGen.getProfitVector());
    }

    public double getCapacity(int i) {
        return capacities[i];
    }

    public double price() {
        return priceThesis.getPrice(marketQ, 0);
    }

    public double getProfit(boolean approximateQuantities) {
        getProduction(approximateQuantities);
        return profit;
    }

    public State getDeepCopy() {
        State copy = new State(seed, priceThesis, productionParameters, policyParameters, exitParameters, entryParameters);
        copy.setCapacitiesInit(capacities);
        return copy;
    }

    public double getProbabilityInvestment() {
        return probabilityInvestment;
    }

    public double getProbabilityDivestment() {
        return probabilityDivestment;
    }

    public double getProbabilityExit() {
        return probabilityExit;
    }

    public void refreshRNG(long nextLong) {
        rng = new Random(nextLong);
        seed2 = rng.nextLong();
    }
}
