/*
 * simulateWMain.java
 *
 * Created on January 26, 2004, 11:22 AM
 */
package estimation.simulateW;

import nervecenter.thesisConstants;
import com.ziena.knitro.KnitroJava;
import estimation.ssrulesLikelihood.investmentParametersContainer;
import java.util.*;
import utility.*;
import java.io.*;
import java.text.*;
import javax.swing.*;
import java.awt.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 *
 * @author  Stephen P. Ryan
 */
public class simulateW implements optimization.Uncmin_methods, mcmc.mcmcFunction {
    // simulation constants

    static int simulationHorizon = 200;  // 25 periods at a positive discount rate should tail off to zero pretty quickly
    static int numSimW = 2;  // paths per alternative policy
    int numParams = 22;  // try out the cost of operation (binds if plant is active)
    // six parameters extra for each b-spline + quadratic costs for investment and divestment
    // need to add a few parameters here to account for b-spline on p_e, p_d, and p_i
    // static int numInequalityDraws = 3000;
    double[] productionParameters = new double[6 + 1];
    ArrayList<Jama.Matrix> wList; // = new ArrayList<Jama.Matrix>();  // list containing all the alternative policies, with the true optimal policy at location 0
//    Jama.Matrix betaBand;
//    Jama.Matrix betaTarget;
//    double varianceBand = 0;
//    double varianceTarget = 0;
    static NumberFormat formatNumber = NumberFormat.getInstance();
    static NumberFormat formatMoney = NumberFormat.getCurrencyInstance();
    boolean fixedCostHeterogeneity = true;
    int LOWER_BOUND_NUM_INEQUALITY_DRAWS = 25;
    int UPPER_BOUND_NUM_INEQUALITY_DRAWS = 38400;
    boolean setIdentifiedApproach = true;
    private simulateWResult estimatedParameters;
    private boolean IMPOSE_CONSTRAINTS_ON_INV_MARGINAL_COST = false;
    private boolean USE_KNITRO = false;
    private boolean USE_LTE = true;
    private boolean KILL_DIVESTMENT = false;

    @SuppressWarnings("unchecked")
    public simulateW(long rngSeed, demandCurve priceThesis, double[] productionCosts, investmentParametersContainer policyParameters, double[] exitParameters, double[] entryParameters) {
        formatMoney.setMaximumFractionDigits(0);
        formatNumber.setMaximumFractionDigits(0);

        String separateString = "";
        if (thesisConstants.SEPARATE_POLICY_FUNCTIONS) {
            if (thesisConstants.LATE_PERIOD) {
                separateString = "late";
            } else {
                separateString = "early";
            }
        }

        if (thesisConstants.LATE_PERIOD) {
            System.out.println("**** LATE PERIOD ****");
        } else {
            System.out.println("**** EARLY PERIOD ****");
        }

        ExecutorService tpes = Executors.newFixedThreadPool(thesisConstants.NUM_THREADS);
        if (fixedCostHeterogeneity) {
            numParams = 22;
        } else {
            numParams = 7;
        }

        if (setIdentifiedApproach) {
            numSimW = thesisConstants.NUM_SIMW;
            LOWER_BOUND_NUM_INEQUALITY_DRAWS = 1250; // multiplier * 125;
            UPPER_BOUND_NUM_INEQUALITY_DRAWS = 1250; // multiplier * 125; // 38400;
        } else {
            numSimW = 1;
            LOWER_BOUND_NUM_INEQUALITY_DRAWS = 25;
            UPPER_BOUND_NUM_INEQUALITY_DRAWS = 38400;
        }

        // sum capacity of other firms
        bsplineBasisGenerator b1 = null;
        // own capacity
        bsplineBasisGenerator b2 = null;

        pmUtility.prettyPrintVector(new Jama.Matrix(productionCosts,1));
        for(int i=0;i<3;i++) {
            // productionParameters[i+1] = productionCosts.get(i+1, 0);
            productionParameters[i+1] = productionCosts[i+1];
        }
        
        if (policyParameters.getVarianceBandEquation() == 0) {
            System.out.println("ALERT: Triggering this varianceBand=0 warning in simulateW.");
        }

        System.out.println("Parameters for band:");
        pmUtility.prettyPrintVector(policyParameters.getBetaBand());
        System.out.println("Variance of band: " + policyParameters.getVarianceBandEquation());
        System.out.println("Parameters for target:");
        pmUtility.prettyPrintVector(policyParameters.getBetaTarget());
        System.out.println("Variance of target: " + policyParameters.getVarianceTargetEquation());

        double[] masterParameters = new double[numParams];

        if (thesisConstants.LATE_PERIOD) {
            double[] startingValues = {-890, 82, 1530, 1530, 1836, 3967, 212, 0, -12, 35400, 41255, 187658, 361967, 943645, -230, 4292, 258629, 231745, 1905, -70860, -77132, -78946};
            masterParameters = startingValues;
        } else {
            double[] startingValues = {-725, 68, 761, 768, 863, 2662, 235, 0, -4, 39078, 41271, 149364, 371211, 1135363, -98, 3091, 263442, 238384, 2536, -75180, -84802, -93061};
            masterParameters = startingValues;
        }

        System.out.println("Dynamic Parameters Starting values: " + pmUtility.stringPrettyPrint(new Jama.Matrix(masterParameters, 1)));

        wList = new ArrayList<Jama.Matrix>();

        String lateString = "Early Period";
        if (State.getLatePeriod()) {
            lateString = "Late Period";
        }
        javax.swing.JFrame f = new javax.swing.JFrame("Dynamic Estimation " + lateString);
        f.getContentPane().setLayout(new BorderLayout());
        JTextArea jt = new JTextAreaAutoscroll();
        jt.setEditable(false);
        f.getContentPane().add(new JScrollPane(jt), BorderLayout.CENTER);
        f.setBounds(200, 100, 800, 500);
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        f.setVisible(true);

        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        for (int i = 0; i < numParams + 1; i++) {
            typsiz[i] = 1;
        }
        double[] fscale = {0, 300};
        int[] method = {0, 2};  // 1 line search 2 double dogleg 3 more-hebdon
        int[] iexp = {0, 1}; // 1 if expensive to evaluate
        int[] msg = {0, 0};
        int[] ndigit = {0, 8};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-8};
        double[] guess = new double[numParams + 1];

        String[] names = {"Investment B-Spline 1", "Investment B-Spline 2", "Investment B-Spline 3", "Investment B-Spline 4", "Investment B-Spline 5", "Investment B-Spline 6",
            "Investment Cost", "Investment Cost Squared",
            "Divestment B-Spline 1", "Divestment B-Spline 2", "Divestment B-Spline 3", "Divestment B-Spline 4", "Divestment B-Spline 5", "Divestment B-Spline 6",
            "Divestment Cost", "Divestment Cost Squared",
            "Exit B-Spline 1", "Exit B-Spline 2", "Exit B-Spline 3", "Exit B-Spline 4", "Exit B-Spline 5", "Exit B-Spline 6", "Investment Dist Mean", "Investment Dist StdDev",
            "Divestment Dist Mean", "Divestment Dist StdDev", "Exit Dist Mean", "Exit Dist StdDev"
        };
        if (!fixedCostHeterogeneity) {
            names = new String[8];
            String[] names2 = {"ADJPOS", "INVMCPOS", "INVMCPOS2", "ADJNEG", "INVMCNEG", "INVMCNEG2", "SCRAP"};
            for (int i = 0; i < 7; i++) {
                names[i] = names2[i];
            }
        }

        System.out.println("Random number seed: "+rngSeed);
        Random rng = new Random(rngSeed);
        System.out.println("NumSimW: " + numSimW);

        for (int i = 0; i < numParams; i++) {
            guess[i + 1] = masterParameters[i];
        }

        double[] dDist = new double[2];
        double[] iDist = new double[2];
        double[] xDist = new double[2];

        ArrayList<Jama.Matrix> wMasterList = new ArrayList<Jama.Matrix>();
        ArrayList<Jama.Matrix> wMasterListA = new ArrayList<Jama.Matrix>();

        try {
            // BufferedWriter capout = new BufferedWriter(new FileWriter("data/cap"+State.getLatePeriod()+".csv"));
            for (int numInequalityDraws = LOWER_BOUND_NUM_INEQUALITY_DRAWS; numInequalityDraws <= UPPER_BOUND_NUM_INEQUALITY_DRAWS; numInequalityDraws *= 2) {
                jt.append("\nNumber of simulation draws: " + (wMasterList.size() + numInequalityDraws) + " (" + numInequalityDraws + " new)\n");
                jt.append("Length of simulated lifetimes: " + simulationHorizon + "\n");
                jt.append("Number of simulation paths in expectations: " + numSimW + "\n");
                jt.append("Initializing simulation paths...");
                long timeStart = System.currentTimeMillis();
                javax.swing.ProgressMonitor mon = new javax.swing.ProgressMonitor(f, "Initializing Inequalities", "0 / " + numInequalityDraws, 0, numInequalityDraws - 2);

                // i should parallelize this since it takes so long
                ArrayList<Future> futureList = new ArrayList<Future>();

                for (int ik = 0; ik < numInequalityDraws; ik++) {
                    futureList.add(tpes.submit(new SimulateWTask(rng.nextLong(), numSimW, numParams, simulationHorizon, fixedCostHeterogeneity, priceThesis, productionParameters, policyParameters, exitParameters, entryParameters)));
                }

                try {
                    for (int ik = 0; ik < futureList.size(); ik++) {
                        Future<ArrayList<Jama.Matrix>> futureTask = futureList.get(ik);
                        ArrayList<Jama.Matrix> W = futureTask.get();
                        wMasterList.add(W.get(0));
                        wMasterListA.add(W.get(1));
                        mon.setProgress(ik + 1);
                        mon.setNote((ik + 1) + " / " + numInequalityDraws);
                        if (ik == 0) {
                            pmUtility.prettyPrint(W.get(0));
                        }
                        // System.out.println("ik: "+ik);
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }

                jt.append("finished.\n");
                try {
                    ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream("data/path" + numInequalityDraws + State.getLatePeriod() + ".dat"));
                    out.writeObject(wMasterList);
                    out.writeObject(wMasterListA);
                    out.close();
                } catch (Exception e) {
                    e.printStackTrace();
                }


                wList = new ArrayList<Jama.Matrix>();
                for (int draw = 0; draw < wMasterList.size(); draw++) {
                    wList.add(wMasterList.get(draw));
                    wList.add(wMasterListA.get(draw));
                }

                double globalMin = 777;

                boolean useLTE = USE_LTE;
                if (useLTE) {
                    boolean preEstimate = false;
                    if (preEstimate) {
                        boolean useKNITROPreEstimate = false;
                        if (useKNITROPreEstimate) {
                            System.out.println("KNITRO pre-estimation before LTE...");
                            solveUsingKNitro(guess);
                        } else {
                            System.out.print("Before pre-estimate: ");
                            pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                            optimization.Uncmin_f77 minimizer = new optimization.Uncmin_f77(false);
                            minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
                            System.out.print("After Pre-estimate: ");
                            pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                        }
                    }
                    int burnLength = 15000;
                    int runLength = 2500;
                    mcmc.gibbsLTEGeneralized lte = new mcmc.gibbsLTEGeneralized(this, runLength, burnLength, guess, names, false, true); // , "lteSimulateW");
                    Jama.Matrix chain = lte.getChain();
                    guess = lte.getLowestPoint();

                    boolean firstSet = true;
                    double minCost = -1;
                    double[] minParameters = new double[numParams + 1];
                    double[] maxParameters = new double[numParams + 1];
                    double minExit = 0;
                    double maxExit = 0;
                    double minInvest = 0;
                    double maxInvest = 0;
                    double minDivest = 0;
                    double maxDivest = 0;
                    double minExitSD = 0;
                    double maxExitSD = 0;
                    double minInvestSD = 0;
                    double maxInvestSD = 0;
                    double minDivestSD = 0;
                    double maxDivestSD = 0;

                    if (globalMin == 0) {
                        System.out.println("Set-identified parameter projection bounds:");
                        jt.append("\n*********\nSet-identified parameter projection bounds:\n");
                        for (int i = 0; i < numParams; i++) {
                            String o = names[i] + " \t&\t" + formatNumber.format(minParameters[i + 1]) + "\t&\t " + formatNumber.format(maxParameters[i + 1]) + " \\\\";
                            System.out.println(o);
                            jt.append(o + "\n");
                        }
                        System.out.println("\\\\");
                        jt.append("\\\\ \n");

                        System.out.print("Investment Cost Distribution Mean \t&\t" + formatNumber.format(minInvest) + "\t&\t" + formatNumber.format(maxInvest) + " \\\\ \n");
                        System.out.print("Investment Cost Distribution Standard Deviation \t&\t" + formatNumber.format(minInvestSD) + "\t&\t" + formatNumber.format(maxInvestSD) + " \\\\ \n");
                        System.out.print("Divestment Cost Distribution Mean \t&\t" + formatNumber.format(minDivest) + "\t&\t" + formatNumber.format(maxDivest) + " \\\\ \n");
                        System.out.print("Divestment Cost Distribution Standard Deviation \t&\t" + formatNumber.format(minDivestSD) + "\t&\t" + formatNumber.format(maxDivestSD) + " \\\\ \n");
                        System.out.print("Exit Cost Distribution Mean \t&\t" + formatNumber.format(minExit) + "\t&\t" + formatNumber.format(maxExit) + " \\\\ \n");
                        System.out.print("Exit Cost Distribution Standard Deviation \t&\t" + formatNumber.format(minExitSD) + "\t&\t" + formatNumber.format(maxExitSD) + " \\\\ \n");

                        jt.append("Investment Cost Distribution Mean \t&\t" + formatNumber.format(minInvest) + "\t&\t" + formatNumber.format(maxInvest) + " \\\\ \n");
                        jt.append("Investment Cost Distribution Standard Deviation \t&\t" + formatNumber.format(minInvestSD) + "\t&\t" + formatNumber.format(maxInvestSD) + " \\\\ \n");
                        jt.append("Divestment Cost Distribution Mean \t&\t" + formatNumber.format(minDivest) + "\t&\t" + formatNumber.format(maxDivest) + " \\\\ \n");
                        jt.append("Divestment Cost Distribution Standard Deviation \t&\t" + formatNumber.format(minDivestSD) + "\t&\t" + formatNumber.format(maxDivestSD) + " \\\\ \n");
                        jt.append("Exit Cost Distribution Mean \t&\t" + formatNumber.format(minExit) + "\t&\t" + formatNumber.format(maxExit) + " \\\\ \n");
                        jt.append("Exit Cost Distribution Standard Deviation \t&\t" + formatNumber.format(minExitSD) + "\t&\t" + formatNumber.format(maxExitSD) + " \\\\ \n");
                        jt.append("*********\n");
                    }

                } else {
                    boolean useKNITRO = USE_KNITRO;
                    if (useKNITRO) {
                        solveUsingKNitro(guess);
                        System.out.println("Post-KNITRO: ");
                        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                    } else {
                        System.out.println("Pre-minimization for starting values...");
                        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                        optimization.Uncmin_f77 minimizer = new optimization.Uncmin_f77(true);
                        minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
                        System.out.println("Post-minimization for starting values...");
                        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                    }
                }
                Jama.Matrix investmentBSplineCoefficients = new Jama.Matrix(6, 1);
                Jama.Matrix divestmentBSplineCoefficients = new Jama.Matrix(6, 1);
                Jama.Matrix exitBSplineCoefficients = new Jama.Matrix(6, 1);
                if (guess.length > 8) {
                    for (int i = 0; i < 6; i++) {
                        investmentBSplineCoefficients.set(i, 0, guess[i + 1]);  // think that is right
                        divestmentBSplineCoefficients.set(i, 0, guess[i + 1 + 8]);  // think that is right
                        exitBSplineCoefficients.set(i, 0, guess[i + 1 + 16]);  // think that is right
                    }
                    iDist = InvertDistributionCosts.computeCosts(investmentBSplineCoefficients, "Investment", false);
                    dDist = InvertDistributionCosts.computeCosts(divestmentBSplineCoefficients, "Divestment", false);
                    xDist = InvertDistributionCosts.computeCosts(exitBSplineCoefficients, "Exit", true);
                }

                /**
                 * Output results
                 */
                boolean outputResults = false;
                if (outputResults) {
                    for (int i = 0; i < numParams; i++) {
                        jt.append(names[i] + " & " + formatNumber.format(guess[i + 1]) + " \\\\ \n");
                        System.out.println(names[i] + " & " + formatNumber.format(guess[i + 1]) + " \\\\");
                    }
                    if (guess.length > 8) {
                        jt.append("Investment distribution mean: " + formatNumber.format(iDist[1]) + " stddev: " + formatNumber.format(iDist[2]) + " \\\\ \n");
                        jt.append("Divestment distribution mean: " + formatNumber.format(dDist[1]) + " stddev: " + formatNumber.format(dDist[2]) + " \\\\ \n");
                        jt.append("Exit distribution mean: " + formatNumber.format(xDist[1]) + " stddev: " + formatNumber.format(xDist[2]) + " \\\\ \n");
                        System.out.print("Investment distribution mean: " + formatNumber.format(iDist[1]) + " stddev: " + formatNumber.format(iDist[2]) + " \\\\ \n");
                        System.out.print("Divestment distribution mean: " + formatNumber.format(dDist[1]) + " stddev: " + formatNumber.format(dDist[2]) + " \\\\ \n");
                        System.out.print("Exit distribution mean: " + formatNumber.format(xDist[1]) + " stddev: " + formatNumber.format(xDist[2]) + " \\\\ \n");
                    }
                }

                boolean outputProfits = false;
                if (outputProfits) {
                    System.out.println("-------------");
                    Jama.Matrix theta = new Jama.Matrix(numParams + 1, 1);
                    theta.set(0, 0, 1);
                    for (int ci = 1; ci <= numParams; ci++) {
                        theta.set(ci, 0, guess[ci]);
                    }

                    for (int i = 0; i < wMasterList.size(); i++) {
                        Jama.Matrix draw = wMasterList.get(i);
                        double profit = draw.times(theta).get(0, 0);
                        Jama.Matrix drawA = wMasterListA.get(i);
                        double profitA = drawA.times(theta).get(0, 0);

                        if (profitA > profit) {
                            System.out.println("------------");
                            pmUtility.prettyPrint(draw);
                            System.out.println(i + ". profit: " + profit);
                            pmUtility.prettyPrint(drawA);
                            System.out.println("Alt. profit: " + profitA);
                        }
                    }
                }

                if (outputResults) {
                    System.out.println("n: " + numInequalityDraws + " guess: " + pmUtility.stringPrettyPrint(new Jama.Matrix(guess, 1)));
                    double cost = guess[2] * 1500 + (guess[3] * 1500 * 1500);
                    if (fixedCostHeterogeneity) {
                        cost = guess[7] * 1500 + (guess[8] * 1500 * 1500);
                    }

                    jt.append("Variable Cost of 1.5M ton per year plant: " + formatMoney.format(cost) + "\n");
                    System.out.println("Variable Cost of 1.5M ton per year plant: " + formatMoney.format(cost));
                    jt.append("Binding: " + binding(guess) + " ");
                    long timeEnd = System.currentTimeMillis();
                    jt.append("Total execution time: " + (timeEnd - timeStart) / 1000.0 + " seconds.\n");
                }

                // only write the answers from the true model
                try {
                    // System.out.print("Writing dynamic parameters...");
                    ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream("data/dynamic" + numParams + State.getLatePeriod() + ".dat"));
                    for (int i = 1; i < guess.length; i++) {
                        out.writeDouble(guess[i]);
                    }
                    out.close();
                    // System.out.println("success.");
                } catch (Exception e) {
                    System.out.println("failure.");
                    e.printStackTrace();
                }

            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        
        System.out.print("Guess: ");
        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
        System.out.print("iDist: ");
        pmUtility.prettyPrint(new Jama.Matrix(iDist, 1));
        System.out.print("dDist: ");
        pmUtility.prettyPrint(new Jama.Matrix(dDist, 1));
        System.out.print("xDist: ");
        pmUtility.prettyPrint(new Jama.Matrix(xDist, 1));
        // System.exit(0);
        estimatedParameters = new simulateWResult(guess, iDist, dDist, xDist, f_to_minimize(guess));
        // }
        // monB.close();

        // pmUtility.prettyPrint(bootResults);

        System.out.println("******************* Estimation complete *******************\n");
        f.dispose();
        tpes.shutdown();
    }

    public int binding(double[] x) {
        int numBind = 0;
        Jama.Matrix theta = new Jama.Matrix(numParams + 1, 1);
        theta.set(0, 0, 1);
        for (int ci = 1; ci <= numParams; ci++) {
            theta.set(ci, 0, x[ci]);
        }
        for (int i = 0; i < wList.size(); i += 2) {
            Jama.Matrix optimal = wList.get(i);
            Jama.Matrix draw = wList.get(i + 1);
            Jama.Matrix diff = optimal.minus(draw);
            double g = diff.times(theta).get(0, 0);
            if (g < 0) {
                numBind++;
            }
        }
        return numBind;
    }

    public double f_to_minimize(double[] x) {
        double value = 0;

        if (!USE_KNITRO) {
            if (x.length > 8) {
                if (IMPOSE_CONSTRAINTS_ON_INV_MARGINAL_COST) {
                    x[7] = Math.max(x[7], 100);
                    x[7] = Math.min(x[7], 300);
                } else {
                    x[7] = Math.max(0, x[7]);
                }
                x[8] = 0;
                if (KILL_DIVESTMENT) {
                    for (int i = 0; i < 8; i++) {
                        x[9 + i] = 0;
                    }
                }
                // x[16] = 0;
            } else {
                // x[1] = Math.max(0, x[1]);   // want to keep the cost of adjustment positive
                // x[4] = Math.max(0, x[4]);
                // x[3] = 0;
                // x[6] = 0;
            }


            // bsplines 1-6 are for the investment
            // bsplines 9-14 for divestment
            // bspline 17-22 are for exit
            if (x.length > 8) {
                for (int i = 2; i <= 6; i++) {
                    x[i] = Math.max(x[i], x[i - 1] + 1E-15);
                }
                for (int i = 10; i <= 14; i++) {
                    x[i] = Math.max(x[i], x[i - 1] + 1E-15);
                }
                for (int i = 21; i >= 17; i--) {
                    x[i] = Math.max(x[i], x[i + 1] + 1E-15);
                }
            }
        }

        Jama.Matrix theta = new Jama.Matrix(numParams + 1, 1);
        theta.set(0, 0, 1);
        for (int ci = 1; ci <= numParams; ci++) {
            theta.set(ci, 0, x[ci]);
        }

        Jama.Matrix optimal;
        Jama.Matrix draw;
        Jama.Matrix diff;
        double g = 0;

        for (int i = 0; i < wList.size(); i += 2) {
            optimal = wList.get(i);
            draw = wList.get(i + 1);
            diff = optimal.minus(draw);
            g = (diff.times(theta)).get(0, 0);
            if (g < 0) {
                value += g * g;
            }
        }
        value /= (2.0 * wList.size());
//        System.out.print(value + " ");
//        pmUtility.prettyPrint(new Jama.Matrix(x, 1));
        return value;
    }

    public void gradient(double[] x, double[] g) {
    }

    public void hessian(double[] x, double[][] h) {
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }

    public simulateWResult getParameters() {
        return estimatedParameters;
    }

    private void solveUsingKNitro(double[] guess) {
        //---- DEFINE THE OPTIMIZATION TEST PROBLEM.
        //---- FOR MORE INFORMATION ABOUT THE PROBLEM DEFINITION, REFER
        //---- TO THE KNITRO MANUAL, ESPECIALLY THE SECTION ON THE
        //---- CALLABLE LIBRARY.
        int n = guess.length - 1;
        int objGoal = KnitroJava.KTR_OBJGOAL_MINIMIZE;
        int objType = KnitroJava.KTR_OBJTYPE_GENERAL;

        // make sure that the problem is defined as unbounded
        double[] bndsLo = new double[n];
        double[] bndsUp = new double[n];
        for (int i = 0; i < n; i++) {
            bndsLo[i] = -KnitroJava.KTR_INFBOUND;
            bndsUp[i] = KnitroJava.KTR_INFBOUND;
        }

        // bounds on investment cost
        if (IMPOSE_CONSTRAINTS_ON_INV_MARGINAL_COST) {
            bndsLo[6] = 100;
            bndsUp[6] = 300;
        } else {
            // at least require that the costs are positive
            bndsLo[6] = 0;
        }
        // set squared costs to zero
        bndsLo[7] = 0;
        bndsUp[7] = 0;
        // potentially turn off the divestment costs
        if (KILL_DIVESTMENT) {
            for (int j = 0; j < 8; j++) {
                bndsLo[8 + j] = 0;
                bndsUp[8 + j] = 0;
            }
        }

        // number of constraints
        int m = 15;
        int[] cType = {KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR,
            KnitroJava.KTR_CONTYPE_LINEAR
        };
        double[] cBndsLo = {-KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND,
            -KnitroJava.KTR_INFBOUND};
        double[] cBndsUp = {0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0};
        int nnzJ = 0;
        int[] jacIxConstr = {};
        int[] jacIxVar = {};
        int nnzH = 0;
        int[] hessRow = {};
        int[] hessCol = {};

        double[] daXInit = new double[guess.length - 1];
        for (int i = 0; i < guess.length - 1; i++) {
            daXInit[i] = guess[i + 1];
        }


        //---- SETUP AND RUN KNITRO TO SOLVE THE PROBLEM.

        //---- CREATE A SOLVER INSTANCE.
        KnitroJava solver = null;
        try {
            solver = new KnitroJava();
        } catch (java.lang.Exception e) {
            System.err.println(e);
            return;
        }

        // ---- DEMONSTRATE HOW TO SET KNITRO PARAMETERS.
        /**
         * Options: all, summary, none
         */
        if (solver.setCharParamByName("outlev", "summary") == false) {
            System.err.println("Error setting parameter 'outlev'");
            return;
        }
//        solver.setIntParamByName("outlev", 0);
//        solver.setIntParamByName("debug", 0);
        if (solver.setDoubleParamByName("feastol", 1.0E-6) == false) {
            System.err.println("Error setting parameter 'feastol'");
            return;
        }
        if (solver.setIntParamByName("gradopt", 2) == false) {
            System.err.println("Error setting parameter 'gradopt'");
            return;
        }
        if (solver.setIntParamByName("hessopt", 6) == false) {
            System.err.println("Error setting parameter 'hessopt'");
            return;
        }

        //---- INITIALIZE KNITRO WITH THE PROBLEM DEFINITION.
        if (solver.initProblem(n, objGoal, objType, bndsLo, bndsUp,
                m, cType, cBndsLo, cBndsUp,
                nnzJ, jacIxVar, jacIxConstr,
                nnzH, hessRow, hessCol,
                daXInit, null) == false) {
            System.err.println("Error initializing the problem, " + "KNITRO status = " + solver.getKnitroStatusCode());
            return;
        }

        //---- ALLOCATE ARRAYS FOR REVERSE COMMUNICATIONS OPERATION.
        double[] daX = new double[n];
        double[] daLambda = new double[m + n];
        double[] daObj = new double[1];
        double[] daC = new double[m];
        double[] daObjGrad = new double[n];
        double[] daJac = new double[nnzJ];
        double[] daHess = new double[nnzH];

        // exampleHS15 thisObject = new exampleHS15();

        //---- SOLVE THE PROBLEM.  IN REVERSE COMMUNICATIONS MODE, KNITRO
        //---- RETURNS WHENEVER IT NEEDS MORE PROBLEM INFORMATION.  THE CALLING
        //---- PROGRAM MUST INTERPRET KNITRO'S RETURN STATUS AND CONTINUE
        //---- SUPPLYING PROBLEM INFORMATION UNTIL KNITRO IS COMPLETE.
        int nKnStatus;
        int nEvalStatus = 0;
        do {
            nKnStatus = solver.solve(nEvalStatus, daObj, daC,
                    daObjGrad, daJac, daHess);
            if (nKnStatus == KnitroJava.KTR_RC_EVALFC) {
                //---- KNITRO WANTS daObj AND daC EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // daObj[0] = thisObject.evaluateFC(daX, daC);
                daObj[0] = evaluateFC(daX, daC);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALGA) {
                //---- KNITRO WANTS daObjGrad AND daJac EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // thisObject.evaluateGA(daX, daObjGrad, daJac);
                evaluateGA(daX, daObjGrad, daJac);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALH) {
                //---- KNITRO WANTS daHess EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                daLambda = solver.getCurrentLambda();
                // thisObject.evaluateH(daX, daLambda, daHess);
                evaluateH(daX, daLambda, daHess);
            }

            //---- ASSUME THAT PROBLEM EVALUATION IS ALWAYS SUCCESSFUL.
            //---- IF A FUNCTION OR ITS DERIVATIVE COULD NOT BE EVALUATED
            //---- AT THE GIVEN (x, lambda), THEN SET nEvalStatus = 1 BEFORE
            //---- CALLING solve AGAIN.
            nEvalStatus = 0;
        } while (nKnStatus > 0);

        //---- DISPLAY THE RESULTS.
//        System.out.print("KNITRO finished, status " + nKnStatus + ": ");
//        switch (nKnStatus) {
//            case KnitroJava.KTR_RC_OPTIMAL:
//                System.out.println("converged to optimality.");
//                break;
//            case KnitroJava.KTR_RC_ITER_LIMIT:
//                System.out.println("reached the maximum number of allowed iterations.");
//                break;
//            case KnitroJava.KTR_RC_NEAR_OPT:
//            case KnitroJava.KTR_RC_FEAS_XTOL:
//            case KnitroJava.KTR_RC_FEAS_FTOL:
//            case KnitroJava.KTR_RC_FEAS_NO_IMPROVE:
//                System.out.println("could not improve upon the current iterate.");
//                break;
//            case KnitroJava.KTR_RC_TIME_LIMIT:
//                System.out.println("reached the maximum CPU time allowed.");
//                break;
//            default:
//                System.out.println("failed.");
//        }

        //---- EXAMPLES OF OBTAINING SOLUTION INFORMATION.
        // System.out.println("  optimal value = " + daObj[0]);
        daX = solver.getCurrentX();
        daLambda = solver.getCurrentLambda();
        // System.out.println("  solution feasibility violation    = " + solver.getAbsFeasError());
        // System.out.println("           KKT optimality violation = " + solver.getAbsOptError());
        // System.out.println("  number of function evaluations    = " + solver.getNumberFCEvals());

        //---- BE CERTAIN THE NATIVE OBJECT INSTANCE IS DESTROYED.
        solver.destroyInstance();

        for (int i = 0; i < daX.length; i++) {
            guess[i + 1] = daX[i];
        }
    }

    private double evaluateFC(double[] daX, double[] daC) {
        // daX is what we usually call guess or x
        // daC is the constraint vector: you fill it it here and that is how knitro knows if something is violated or not
        // daX starts counting at 0 and not 1, so we'll need to move everything over by one and pass it to f_to_minimize
        // this could potentially work out very nicely!
        double[] guessX = new double[daX.length + 1];
        for (int i = 0; i < daX.length; i++) {
            guessX[i + 1] = daX[i];
        }

        // no need to fill daC since I have no constraints
        // investment b-spline is increasing
        daC[0] = daX[0] - daX[1];
        daC[1] = daX[1] - daX[2];
        daC[2] = daX[2] - daX[3];
        daC[3] = daX[3] - daX[4];
        daC[4] = daX[4] - daX[5];
        // so is the divestment b-spline
        daC[5] = daX[8] - daX[9];
        daC[6] = daX[9] - daX[10];
        daC[7] = daX[10] - daX[11];
        daC[8] = daX[11] - daX[12];
        daC[9] = daX[12] - daX[13];
        // exit b-spline is decreasing
        daC[10] = daX[20] - daX[19];
        daC[11] = daX[19] - daX[18];
        daC[12] = daX[18] - daX[17];
        daC[13] = daX[17] - daX[16];
        daC[14] = daX[21] - daX[20];


        return f_to_minimize(guessX);
    }

    private void evaluateGA(double[] daX, double[] daObjGrad, double[] daJac) {
        throw new UnsupportedOperationException("KNitro Gradient: Not yet implemented");
    }

    private void evaluateH(double[] daX, double[] daLambda, double[] daHess) {
        throw new UnsupportedOperationException("KNitro Hessian: Not yet implemented");
    }
}
