package estimation.simulateWEntry;

import JSci.maths.statistics.NormalDistribution;
import nervecenter.thesisConstants;
import com.ziena.knitro.KnitroJava;
import java.awt.BorderLayout;
import java.util.*;
import utility.*;
import java.text.*;
import javax.swing.*;
import org.jfree.chart.*;
import org.jfree.data.xy.*;
import estimation.simulateW.State;
import estimation.ssrulesLikelihood.investmentParametersContainer;

/**
 * 
 * @author Stephen P. Ryan
 */
public class simulateWEntryCostsHeterogeneity implements optimization.Uncmin_methods, mcmc.mcmcFunction {

    // simulation constants
    static int simulationHorizon = 200;  // 25 periods at a positive discount rate should tail off to zero pretty quickly
    static int numSimW = thesisConstants.NUM_SIMW;  // paths per alternative policy (1 if there are no probabilistic transitions)
    static int numInequalityDraws = 10;
    // static double[] productionParameters = new double[7];  // capacity cost and shifter
    static Jama.Matrix betaBand;
    static Jama.Matrix betaTarget;
    static double varianceBand;
    static double varianceTarget;
    static NumberFormat format = NumberFormat.getInstance();
    static Jama.Matrix results = null;
    static javax.swing.JFrame f = null;
    static JTextArea jt = null;
    private double estimatedMu;
    private double estimatedSigma;
    Jama.Matrix data;
    static XYSeries xy = new XYSeries("Expected value", true, false);
    static XYSeriesCollection collection = new XYSeriesCollection(xy);
    int numComp = 1;
    static boolean latePeriod = State.getLatePeriod();
    int numParams = 22;
    private static bsplineBasisGenerator unitIntervalBSpline = new bsplineBasisGenerator(0, 1);
    java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
    static boolean useGUI = false;

    private void initializeData(demandCurve priceThesis, double[] productionCosts,
            investmentParametersContainer policyParameters, double[] exitParameters, double[] entryParameters, Jama.Matrix dynamicParameters) {
        // for (numSimW = 1500; numSimW <= 1500; numSimW *= 2) {
        System.out.println("Late period: " + latePeriod);
        System.out.println("NumSimW: " + numSimW);
        // javax.swing.ProgressMonitor mon = new javax.swing.ProgressMonitor(f, "Initializing Inequalities", "0 / " + numInequalityDraws, 0, numInequalityDraws - 1);

        ArrayList<Jama.Matrix> wMasterList = new ArrayList<Jama.Matrix>();

        double ENTRY_POST_1990_SHIFTER = entryParameters[2];
        double ENTRY_INTERCEPT = entryParameters[1];
        double ENTRY_INCUMBENT_CAPACITY_COEFFICIENT = entryParameters[3];

        Jama.Matrix theta = new Jama.Matrix(numParams + 1, 1);
        // have to set first element of theta to 1 to pick up profits

        pmUtility.prettyPrint(dynamicParameters);
        theta.set(0, 0, 1);
        for(int i=0;i<22;i++) {
            theta.set(i+1, 0, dynamicParameters.get(0, i));
        }
        pmUtility.prettyPrintVector(theta);

        if (useGUI) {
            jt.append("Loaded the following parameters:\n");
            jt.append(pmUtility.stringPrettyPrintVector(theta) + "\n");
        }

        data = new Jama.Matrix(numInequalityDraws, 2);

        Random rng = new Random(911);
        
        Jama.Matrix profits = new Jama.Matrix(numInequalityDraws, 1);
        Jama.Matrix capacityCompetitor = new Jama.Matrix(numInequalityDraws, 1);
        Jama.Matrix capacityOwnFirm = new Jama.Matrix(numInequalityDraws, 1);

        for (int ik = 0; ik < numInequalityDraws; ik++) {
            double[] cap = new double[1 + numComp];
            cap[0] = State.entryLevel;

            // vary these parameters depending on whether it is early or late
            // if early, a = 600, b = 10
            double a = 500;
            double b = 125;
            // if late, a = 60, b = 5?
            if (thesisConstants.LATE_PERIOD) {
                a = 500;
                b = 125;
            }

            for (int i = 0; i < numComp; i++) {
                cap[i + 1] = a + b * ik;
            }

            double[] W = new double[numParams + 1];
            for (int ci = 0; ci < numParams; ci++) {
                W[ci] = 0;
            }
            State baseState = new State(rng.nextLong(), priceThesis, productionCosts, policyParameters, exitParameters, entryParameters);
            baseState.setCapacitiesInit(cap);
            double profitJ = 0;
            double capacityCompetitorJ = 0;
            double capacityOwnFirmJ = 0;
            ProgressMonitor simMon = null;
            if (useGUI) {
                simMon = new ProgressMonitor(f, "Simulating Expected Value of Entry", "0 / " + numSimW, 0, numSimW);
            }
            for (int j = 0; j < numSimW; j++) {
                State s = baseState.getDeepCopy();
                s.refreshRNG(rng.nextLong());
                double[] Win = new double[numParams + 1];
                for (int ci = 0; ci < numParams; ci++) {
                    Win[ci] = 0;
                }
                double discountedSumCompetitorCapacity = 0;
                double discountedSumOwnCapacity = 0;
                for (int i = 0; i < simulationHorizon; i++) {
                    double sumCompetitorCapacity = 0;
                    for (int ui = 0; ui < numComp; ui++) {
                        sumCompetitorCapacity += s.getCapacity(ui + 1);
                    }
                    double discount = Math.pow(thesisConstants.DISCOUNT_FACTOR, i);
                    discountedSumCompetitorCapacity += discount * sumCompetitorCapacity;
                    discountedSumOwnCapacity += discount * s.getCapacity(0);

                    // System.out.println("s.getPolicyFirst()");
                    double policy = s.getPolicyFirst();
                    double probabilityExit = s.getProbabilityExit();
                    double probabilityDivestment = s.getProbabilityDivestment();
                    double probabilityInvestment = s.getProbabilityInvestment();

                    double profit = s.getProfit(true);

                    if (profit <= 0) {
                        System.out.println("Price: "+s.price());
                        System.out.println("ownCap: " + s.getCapacity(0) + " capOther: " + sumCompetitorCapacity + " Profit: " + profit);
                    }

                    Win[0] += discount * profit;
                    if (policy > 0) {
                        // here is where i need to implement the investment cost function
                        // right now have a constant to indicate that an investment occured
                        // want to replace that with the bs-spline evaluated at prob_investment, right?
                        // think so.  this will be a vector of basis functions now instead of a constant
                        Jama.Matrix basis = unitIntervalBSpline.getBasis(probabilityInvestment);
                        for (int dm = 0; dm < 6; dm++) {
                            Win[dm + 1] -= discount * basis.get(0, dm);
                        }
                        Win[7] -= discount * (policy);
                        Win[8] -= discount * (policy * policy);
                    }
                    // System.out.println("s.getExit()");
                    if (policy < 0 && !s.getExit()) {
                        // also for negative adjustments
                        Jama.Matrix basis = unitIntervalBSpline.getBasis(probabilityDivestment);
                        for (int dm = 0; dm < 6; dm++) {
                            Win[dm + 9] -= discount * basis.get(0, dm);
                        }
                        Win[15] -= discount * policy;
                        Win[16] -= discount * policy * policy;
                    }

                    s.updateState(policy);

                    if (s.getExit()) {
                        Jama.Matrix basis = unitIntervalBSpline.getBasis(probabilityExit);
                        for (int dm = 0; dm < 6; dm++) {
                            Win[dm + 17] += discount * basis.get(0, dm);
                        }
                        Win[17] += discount;

                        i = simulationHorizon + 1; // if you exit, you are done
                    }
                }
                for (int ci = 0; ci < W.length; ci++) {
                    W[ci] += Win[ci];
                }
                capacityCompetitorJ += discountedSumCompetitorCapacity;
                profitJ += W[0];
                capacityOwnFirmJ += discountedSumOwnCapacity;
                if (useGUI) {
                    simMon.setProgress(j + 1);
                    simMon.setNote((j + 1) + " / " + numSimW);
                }
            }
            capacityCompetitorJ /= numSimW;
            profitJ /= numSimW;
            capacityOwnFirmJ /= numSimW;

            profits.set(ik, 0, profitJ);
            capacityCompetitor.set(ik, 0, capacityCompetitorJ);
            capacityOwnFirm.set(ik, 0, capacityOwnFirmJ);


            for (int ci = 0; ci < W.length; ci++) {
                W[ci] /= numSimW;
            }
            Jama.Matrix wMat = new Jama.Matrix(W, 1);
            if (useGUI) {
                jt.append(ik + "." + pmUtility.stringPrettyPrint(wMat) + "\n");
            }
            JSci.maths.statistics.NormalDistribution normal = new JSci.maths.statistics.NormalDistribution();

            double shifter = ENTRY_POST_1990_SHIFTER;
            if (!latePeriod || thesisConstants.FORCE_EARLY_DYNAMIC_PARAMETERS) {
                shifter = 0;
            }
            double incumbentCapacity = 0;
            for (int i = 0; i < numComp; i++) {
                incumbentCapacity += cap[i + 1];
            }

            // System.out.print("Force: "+thesisConstants.FORCE_EARLY_DYNAMIC_PARAMETERS+" Entry Shifter: "+shifter+"\n");

            double prentry = normal.cumulative(shifter + ENTRY_INTERCEPT + ENTRY_INCUMBENT_CAPACITY_COEFFICIENT * incumbentCapacity);
            data.set(ik, 0, prentry);
            data.set(ik, 1, wMat.times(theta).get(0, 0));
            // xy.add((cap[1] * numComp), data.get(ik, 1));
            if (useGUI) {
                jt.append(ik + ". Incumbent capacity: " + format.format(cap[1] * numComp) + " Expected profit: " + format.format(wMat.times(theta).get(0, 0)) + " Prob entry: " + format.format(prentry) + "\n");
                String[] names = {"Product Market Profit", "Investment B-Spline 1", "Investment B-Spline 2", "Investment B-Spline 3", "Investment B-Spline 4", "Investment B-Spline 5", "Investment B-Spline 6",
                    "Investment Cost", "Investment Cost Squared",
                    "Divestment B-Spline 1", "Divestment B-Spline 2", "Divestment B-Spline 3", "Divestment B-Spline 4", "Divestment B-Spline 5", "Divestment B-Spline 6",
                    "Divestment Cost", "Divestment Cost Squared",
                    "Exit B-Spline 1\t", "Exit B-Spline 2\t", "Exit B-Spline 3\t", "Exit B-Spline 4\t", "Exit B-Spline 5\t", "Exit B-Spline 6\t", "Investment Dist Mean", "Investment Dist StdDev",
                    "Divestment Dist Mean", "Divestment Dist StdDev", "Exit Dist Mean", "Exit Dist StdDev"};
                for (int mt = 0; mt < theta.getRowDimension(); mt++) {
                    jt.append(mt + "\t" + names[mt] + "\t" + nf.format(wMat.get(0, mt)) + "\t" + nf.format(theta.get(mt, 0)) + "\t" + nf.format(wMat.get(0, mt) * theta.get(mt, 0)) + "\n");
                }
            }
            wMasterList.add(new Jama.Matrix(W, 1));
        }

        for (int i = 0; i < wMasterList.size(); i++) {
            Jama.Matrix draw = wMasterList.get(i);
            double profit = draw.times(theta).get(0, 0);
            results.set(i, 0, profit);
        }

        System.out.println("numSimW: " + numSimW);
        pmUtility.prettyPrint(data);
        // }
    }


    public void execute(demandCurve priceThesis, double[] productionCosts,
            investmentParametersContainer policyParameters, double[] exitParameters, double[] entryParameters, Jama.Matrix dynamicParameters) {

        if (useGUI) {
            f = new javax.swing.JFrame("Entry Cost Simulations");
            jt = new JTextAreaAutoscroll();
            jt.setEditable(false);
            f.getContentPane().setLayout(new BorderLayout());
            // f.getContentPane().add(new JScrollPane(jt), BorderLayout.SOUTH);
            String lateString = "Early Period";
            if (State.getLatePeriod()) {
                lateString = "Late Period";
            }
            JFreeChart chart = ChartFactory.createScatterPlot("Entry Values - " + lateString, "Incumbent Capacity", "Dollars", collection, org.jfree.chart.plot.PlotOrientation.VERTICAL, true, true, true);
            ChartPanel panel = new ChartPanel(chart);
            f.getContentPane().add(panel, BorderLayout.CENTER);
            f.getContentPane().add(new JScrollPane(jt), BorderLayout.CENTER);
            f.setBounds(100, 100, 660, 600);
            // f.pack();
            f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            f.setVisible(true);

            jt.append("Entry Cost Distribution Estimation\n");
        }

        bsplineBasisGenerator b1 = null;
        bsplineBasisGenerator b2 = null;

        String separateString = "";
        if (thesisConstants.SEPARATE_POLICY_FUNCTIONS) {
            if (thesisConstants.LATE_PERIOD) {
                separateString = "late";
            } else {
                separateString = "early";
            }
        }


        results = new Jama.Matrix(numInequalityDraws, 1);
        initializeData(priceThesis, productionCosts, policyParameters, exitParameters, entryParameters, dynamicParameters);

        int numEntryParams = 2;
        double[] guess = new double[numEntryParams + 1];
        double[] xpls = new double[numEntryParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numEntryParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numEntryParams + 1][numEntryParams + 1];
        double[] udiag = new double[numEntryParams + 1];
        double[] typsiz = new double[numEntryParams + 1];
        for (int i = 1; i <= numEntryParams; i++) {
            typsiz[i] = 1;
            guess[i] = 1;
        }
        double[] fscale = {0, 1E-8};
        int[] method = {0, 3};
        int[] iexp = {0, 0};
        int[] msg = {0, 1};
        int[] ndigit = {0, 15};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-6};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-8};

        //   421,280    19,750,211,376  early
        //   460,942    16,494,896,160  late

        guess[1] = 178970;
        guess[2] = 100000;

        optimization.Uncmin_f77 minimizer = new optimization.Uncmin_f77(true);

        Jama.Matrix dataMaster = data.copy();
        data = new Jama.Matrix(dataMaster.getRowDimension(), 2);
        for (int j = 0; j < dataMaster.getRowDimension(); j++) {
            data.set(j, 0, dataMaster.get(j, 0));
            data.set(j, 1, dataMaster.get(j, 1));
        }
//            minimizer.optif9_f77(numEntryParams, guess, wec, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
        System.out.print("Before LTE: ");
        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
        // gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 2000, 2000, guess);
        // guess = lte.getLowestPoint();
        System.out.print("After LTE: ");
        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
        System.out.println("Objective function: " + f_to_minimize(guess));

        // minimizer.optif9_f77(numEntryParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
        solveUsingKNitro(guess);
        System.out.println("After KNITRO:");
        pmUtility.prettyPrint(new Jama.Matrix(guess, 1));

        NormalDistribution stdNormal = new NormalDistribution();
        // for 0.01
        NormalDistribution reasonableTest = new NormalDistribution(guess[1], guess[2] * guess[2]);
        System.out.println("Firm entering with 1 percent chance draws below: " + reasonableTest.inverse(0.01));
        System.out.println("Firm entering with 4 percent chance draws below: " + reasonableTest.inverse(0.04));
        double d = reasonableTest.inverse(0.01);
        double z = (d - guess[1]) / guess[2];
        double expectedTruncatedValue = guess[1] - ((stdNormal.probability(z) / stdNormal.cumulative(z)) * guess[2]);
        System.out.println("0.01 pays " + expectedTruncatedValue);
        // for 0.04
        d = reasonableTest.inverse(0.04);
        z = (d - guess[1]) / guess[2];
        expectedTruncatedValue = guess[1] - ((stdNormal.probability(z) / stdNormal.cumulative(z)) * guess[2]);
        System.out.println("0.04 pays " + expectedTruncatedValue);


        if (State.getLatePeriod()) {
            System.out.print("LATE PERIOD ");
        } else {
            System.out.print("EARLY PERIOD ");
        }
        System.out.println(" Entry Distribution Mu: " + guess[1] + " Sigma: " + guess[2]);
        estimatedMu = guess[1];
        estimatedSigma = guess[2];
        if (useGUI) {
            f.dispose();
        }
    }

    public simulateWEntryCostsHeterogeneity(demandCurve priceThesis, double[] costParameters, investmentParametersContainer policyParameters,
            double[] exitParameters, double[] entryParameters, Jama.Matrix dynamicParameters) {

        if (!thesisConstants.SUBSAMPLE_PARAMETERS) {
            useGUI = true;
        }
        execute(priceThesis, costParameters, policyParameters, exitParameters, entryParameters, dynamicParameters);
    }

    public double f_to_minimize(double[] x) {
        // x[1] is mu
        // x[2] is std.dev.
        // match probs to empirical prob of entry
        double value = 0;
        x[2] = Math.max(1E-15, x[2]);
        // what if restrict the late sigma to be equal to early sigma?
        // x[2] = 64306;
        JSci.maths.statistics.ProbabilityDistribution pr = new JSci.maths.statistics.NormalDistribution(x[1], x[2] * x[2]);
        // JSci.maths.statistics.ProbabilityDistribution pr = new JSci.maths.statistics.CauchyDistribution(x[1], Math.max(1E-15, x[2]));
        for (int i = 0; i < data.getRowDimension(); i++) {
            // System.out.println("x[1]: "+x[1]+" x[2]: "+x[2]+" prEntry: "+data.get(i,0)+" profit: "+data.get(i,1)+" F: "+pr.cumulative(data.get(i,1)));
            // value += Math.pow(data.get(i, 0) - pr.cumulative(data.get(i, 1)), 2);
            value += data.get(i, 0) - pr.cumulative(data.get(i, 1));
        }
        // System.out.println(pmUtility.stringPrettyPrint(new Jama.Matrix(x, 1))+" "+value);
        value = value * value;
        return value;
    }

    public void gradient(double[] x, double[] g) {
    }

    public void hessian(double[] x, double[][] h) {
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }

    /**
     * Getter for property data.
     * @return Value of property data.
     */
    public Jama.Matrix getData() {
        return data;
    }

    /**
     * Setter for property data.
     * @param data New value of property data.
     */
    public void setData(Jama.Matrix data) {
        this.data = data;
    }

    /**
     * @return the estimatedMu
     */
    public double getEstimatedMu() {
        return estimatedMu;
    }

    /**
     * @return the estimatedSigma
     */
    public double getEstimatedSigma() {
        return estimatedSigma;
    }

    public void solveUsingKNitro(double[] guess) {
        //---- DEFINE THE OPTIMIZATION TEST PROBLEM.
        //---- FOR MORE INFORMATION ABOUT THE PROBLEM DEFINITION, REFER
        //---- TO THE KNITRO MANUAL, ESPECIALLY THE SECTION ON THE
        //---- CALLABLE LIBRARY.
        int n = guess.length - 1;
        int objGoal = KnitroJava.KTR_OBJGOAL_MINIMIZE;
        int objType = KnitroJava.KTR_OBJTYPE_GENERAL;

        // make sure that the problem is defined as unbounded
        double[] bndsLo = new double[n];
        double[] bndsUp = new double[n];
        for (int i = 0; i < n; i++) {
            bndsLo[i] = 0; // -KnitroJava.KTR_INFBOUND;
            bndsUp[i] = KnitroJava.KTR_INFBOUND;
        }

        // number of constraints
        int m = 0;
        int[] cType = {};
        double[] cBndsLo = {};
        double[] cBndsUp = {};
        int nnzJ = 0;
        int[] jacIxConstr = {};
        int[] jacIxVar = {};
        int nnzH = 0;
        int[] hessRow = {};
        int[] hessCol = {};

        double[] daXInit = new double[guess.length - 1];
        for (int i = 0; i < guess.length - 1; i++) {
            daXInit[i] = guess[i + 1];
        }


        //---- SETUP AND RUN KNITRO TO SOLVE THE PROBLEM.

        //---- CREATE A SOLVER INSTANCE.
        KnitroJava solver = null;
        try {
            solver = new KnitroJava();
        } catch (java.lang.Exception e) {
            System.err.println(e);
            return;
        }

        // ---- DEMONSTRATE HOW TO SET KNITRO PARAMETERS.
        /**
         * Options: all, summary, none
         */
        if (solver.setCharParamByName("outlev", "all") == false) {
            System.err.println("Error setting parameter 'outlev'");
            return;
        }
//        solver.setIntParamByName("outlev", 0);
//        solver.setIntParamByName("debug", 0);
        if (solver.setDoubleParamByName("feastol", 1.0E-3) == false) {
            System.err.println("Error setting parameter 'feastol'");
            return;
        }
        if (solver.setIntParamByName("gradopt", 2) == false) {
            System.err.println("Error setting parameter 'gradopt'");
            return;
        }
        if (solver.setIntParamByName("hessopt", 6) == false) {
            System.err.println("Error setting parameter 'hessopt'");
            return;
        }

        //---- INITIALIZE KNITRO WITH THE PROBLEM DEFINITION.
        if (solver.initProblem(n, objGoal, objType, bndsLo, bndsUp,
                m, cType, cBndsLo, cBndsUp,
                nnzJ, jacIxVar, jacIxConstr,
                nnzH, hessRow, hessCol,
                daXInit, null) == false) {
            System.err.println("Error initializing the problem, " + "KNITRO status = " + solver.getKnitroStatusCode());
            return;
        }

        //---- ALLOCATE ARRAYS FOR REVERSE COMMUNICATIONS OPERATION.
        double[] daX = new double[n];
        double[] daLambda = new double[m + n];
        double[] daObj = new double[1];
        double[] daC = new double[m];
        double[] daObjGrad = new double[n];
        double[] daJac = new double[nnzJ];
        double[] daHess = new double[nnzH];

        // exampleHS15 thisObject = new exampleHS15();

        //---- SOLVE THE PROBLEM.  IN REVERSE COMMUNICATIONS MODE, KNITRO
        //---- RETURNS WHENEVER IT NEEDS MORE PROBLEM INFORMATION.  THE CALLING
        //---- PROGRAM MUST INTERPRET KNITRO'S RETURN STATUS AND CONTINUE
        //---- SUPPLYING PROBLEM INFORMATION UNTIL KNITRO IS COMPLETE.
        int nKnStatus;
        int nEvalStatus = 0;
        do {
            nKnStatus = solver.solve(nEvalStatus, daObj, daC,
                    daObjGrad, daJac, daHess);
            if (nKnStatus == KnitroJava.KTR_RC_EVALFC) {
                //---- KNITRO WANTS daObj AND daC EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // daObj[0] = thisObject.evaluateFC(daX, daC);
                daObj[0] = evaluateFC(daX, daC);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALGA) {
                //---- KNITRO WANTS daObjGrad AND daJac EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                // thisObject.evaluateGA(daX, daObjGrad, daJac);
                evaluateGA(daX, daObjGrad, daJac);
            } else if (nKnStatus == KnitroJava.KTR_RC_EVALH) {
                //---- KNITRO WANTS daHess EVALUATED AT THE POINT x.
                daX = solver.getCurrentX();
                daLambda = solver.getCurrentLambda();
                // thisObject.evaluateH(daX, daLambda, daHess);
                evaluateH(daX, daLambda, daHess);
            }

            //---- ASSUME THAT PROBLEM EVALUATION IS ALWAYS SUCCESSFUL.
            //---- IF A FUNCTION OR ITS DERIVATIVE COULD NOT BE EVALUATED
            //---- AT THE GIVEN (x, lambda), THEN SET nEvalStatus = 1 BEFORE
            //---- CALLING solve AGAIN.
            nEvalStatus = 0;
        } while (nKnStatus > 0);

        //---- DISPLAY THE RESULTS.
//        System.out.print("KNITRO finished, status " + nKnStatus + ": ");
//        switch (nKnStatus) {
//            case KnitroJava.KTR_RC_OPTIMAL:
//                System.out.println("converged to optimality.");
//                break;
//            case KnitroJava.KTR_RC_ITER_LIMIT:
//                System.out.println("reached the maximum number of allowed iterations.");
//                break;
//            case KnitroJava.KTR_RC_NEAR_OPT:
//            case KnitroJava.KTR_RC_FEAS_XTOL:
//            case KnitroJava.KTR_RC_FEAS_FTOL:
//            case KnitroJava.KTR_RC_FEAS_NO_IMPROVE:
//                System.out.println("could not improve upon the current iterate.");
//                break;
//            case KnitroJava.KTR_RC_TIME_LIMIT:
//                System.out.println("reached the maximum CPU time allowed.");
//                break;
//            default:
//                System.out.println("failed.");
//        }

        //---- EXAMPLES OF OBTAINING SOLUTION INFORMATION.
        // System.out.println("  optimal value = " + daObj[0]);
        daX = solver.getCurrentX();
        daLambda = solver.getCurrentLambda();
        // System.out.println("  solution feasibility violation    = " + solver.getAbsFeasError());
        // System.out.println("           KKT optimality violation = " + solver.getAbsOptError());
        // System.out.println("  number of function evaluations    = " + solver.getNumberFCEvals());

        //---- BE CERTAIN THE NATIVE OBJECT INSTANCE IS DESTROYED.
        solver.destroyInstance();

        for (int i = 0; i < daX.length; i++) {
            guess[i + 1] = daX[i];
        }
    }

    private double evaluateFC(double[] daX, double[] daC) {
        // daX is what we usually call guess or x
        // daC is the constraint vector: you fill it it here and that is how knitro knows if something is violated or not
        // daX starts counting at 0 and not 1, so we'll need to move everything over by one and pass it to f_to_minimize
        // this could potentially work out very nicely!
        double[] guessX = new double[daX.length + 1];
        for (int i = 0; i < daX.length; i++) {
            guessX[i + 1] = daX[i];
        }

        // no need to fill daC since I have no constraints

        return f_to_minimize(guessX);
    }

    private void evaluateGA(double[] daX, double[] daObjGrad, double[] daJac) {
        throw new UnsupportedOperationException("KNitro Gradient: Not yet implemented");
    }

    private void evaluateH(double[] daX, double[] daLambda, double[] daHess) {
        throw new UnsupportedOperationException("KNitro Hessian: Not yet implemented");
    }
}
