/*
 * plantRecord.java
 *
 * Created on February 3, 2004, 3:46 PM
 */
package estimation.ssrules;

import java.io.Serializable;
import java.util.ArrayList;

/**
 *
 * @author  Stephen P. Ryan
 */
public class plantRecord implements Serializable, Comparable<plantRecord> {
    private String name;
    private String state;
    private int year;
    private double quantity;
    private int numFirms;
    private String id;
    private double investment = 0;
    private ArrayList<Double> marketCapacities;
    private ArrayList<Double> marketQuantities;
    private double marketQ = 0;
    private double dailyCapacity;
    private double annualCapacity;
    private double wage96;
    private double coal96;
    private double elec96;
    private double gas96;
    private double population;
    private int stateIndex = 0;
    private Jama.Matrix lastGuess = null;
    private boolean entered = false;
    private boolean exited = false;
    
    public plantRecord deepCopy() {
        plantRecord p = new plantRecord(name, state, year);
        p.quantity = quantity;
        p.numFirms = numFirms;
        p.id = new String(id);
        p.investment = investment;
        p.marketCapacities = new ArrayList<Double>(marketCapacities);
        p.marketQuantities = new ArrayList<Double>(marketQuantities);
        p.dailyCapacity = dailyCapacity;
        p.annualCapacity = annualCapacity;
        p.marketQ = marketQ;
        p.wage96 = wage96;
        p.coal96 = coal96;
        p.elec96 = elec96;
        p.gas96 = gas96;
        p.population = population;
        p.stateIndex = stateIndex;
        return p;
    }
    
    public plantRecord(String pname, String pstate, int pyear) {
        name = new String(pname);
        state = new String(pstate);
        year = pyear;
        if(state.equalsIgnoreCase("AKHIORWA")) {
            stateIndex = 1;
        }
        if(state.equalsIgnoreCase("AZNVNM")) {
            stateIndex = 2;
        }
        if(state.equalsIgnoreCase("AROK")) {
            stateIndex = 3;
        }
        if(state.equalsIgnoreCase("CA")) {
            // have to add dilineation based on zip code or something here
            stateIndex = 4;
        }
        if(state.equalsIgnoreCase("COWY")) {
            stateIndex = 5;
        }
        if(state.equalsIgnoreCase("FL")) {
            stateIndex = 6;
        }
        if(state.equalsIgnoreCase("GATN")) {
            stateIndex = 7;
        }
        if(state.equalsIgnoreCase("IDMTUT")) {
            stateIndex = 8;
        }
        if(state.equalsIgnoreCase("IL")) {
            stateIndex = 9;
        }
        if(state.equalsIgnoreCase("IN")) {
            stateIndex = 10;
        }
        if(state.equalsIgnoreCase("IANBSD")) {
            stateIndex = 11;
        }
        if(state.equalsIgnoreCase("KS")) {
            stateIndex = 12;
        }
        if(state.equalsIgnoreCase("KYMSNCLA")) {
            stateIndex = 13;
        }
        if(state.equalsIgnoreCase("MDVAWV")) {
            stateIndex = 14;
        }
        if(state.equalsIgnoreCase("MIWI")) {
            stateIndex = 15;
        }
        if(state.equalsIgnoreCase("MO")) {
            stateIndex = 16;
        }
        if(state.equalsIgnoreCase("NYME")) {
            stateIndex = 17;
        }
        if(state.equalsIgnoreCase("OH")) {
            stateIndex = 18;
        }
        if(state.equalsIgnoreCase("PA")) {
            // again differentiate
            stateIndex = 19;
        }
        if(state.equalsIgnoreCase("SC")) {
            stateIndex = 20;
        }
        if(state.equalsIgnoreCase("TX")) {
            // differentiate three ways
            stateIndex = 21;
        }
    }
    
    public double getCapOtherDouble() {
        double value = 0;
        for(int i=0;i<marketCapacities.size();i++) {
                value += marketCapacities.get(i);
        }
        value -= annualCapacity;
        return value;
    }
    
    public Jama.Matrix getMarketCapacityMatrix() {
        Jama.Matrix matrix = new Jama.Matrix(numFirms, 1);
        for(int i=0;i<numFirms;i++) {
            matrix.set(i, 0, marketCapacities.get(i));
        }
        return matrix;
    }
    
    public Jama.Matrix getMarketQuantityVector() {
        Jama.Matrix matrix = new Jama.Matrix(numFirms, 1);
        for(int i=0;i<numFirms;i++) {
            matrix.set(i, 0, marketQuantities.get(i));
        }
        return matrix;
    }
    
    public double getCapacity(int firmIndex) {
        return marketCapacities.get(firmIndex);
    }

    public double getQuantity(int firmIndex) {
        return marketCapacities.get(firmIndex);
    }
    
    public void setMarketQ(double x) {
        marketQ = x;
    }
    
    public double getMarketQ() {
        return marketQ;
    }
    
    public void setAnnualCapacity(double x) {
        annualCapacity = x;
    }
    
    public double getAnnualCapacity() {
        return annualCapacity;
    }
    
    public void setDailyCapacity(double x) {
        dailyCapacity = x;
    }
    
    public double getDailyCapacity() {
        return dailyCapacity;
    }
    
    public String getState() {
        return new String(state);
    }
    
    public void setState(String s) {
        state = new String(s);
    }
    
    public void setMarketQuantities(ArrayList<Double> x) {
        marketQuantities = new ArrayList<Double>(x);
    }
    
    public ArrayList getMarketQuantities() {
        return marketQuantities;
    }
    
    public void setMarketCapacities(ArrayList<Double> x) {
        marketCapacities = new ArrayList<Double>(x);
    }
    
    public ArrayList getMarketCapacities() {
        return marketCapacities;
    }
    
    public void setInvestment(double x) {
        investment = x;
    }
    
    public double getInvestment() {
        return investment;
    }
    
    public void setQ(double q) {
        quantity = q;
    }
    
    public double getQ() {
        return quantity;
    }
    
    public void setNumFirms(int i) {
        numFirms = i;
    }
    
    public int getNumFirms() {
        return numFirms;
    }
    
    public void setID(String s) {
        id = new String(s);
    }
    
    public String getID() {
        return new String(id);
    }
    
    public void setName(String s) {
        name = new String(s);
    }
    
    public String getName() {
        return new String(name);
    }
    
    public void setYear(int i) {
        year = i;
    }
    
    public int getYear() {
        return year;
    }
    
    @Override
    public String toString() {
        return id+","+name+","+state+","+annualCapacity+","+year+","+quantity+","+numFirms+","+investment+","+
                ","+marketQ+","+dailyCapacity+","+wage96+","+coal96+","+elec96+","+gas96+","+population+","+getCapOtherDouble();
    }
    
    public int compareTo(plantRecord comp) {
        int c = -1*comp.getState().compareTo(getState());
        if(c==0) {
            if(comp.getYear()<getYear()) {
                return 1;
            }
            if(comp.getYear()>getYear()) {
                return -1;
            }
            return 0;
        }
        else {
            return c;
        }
    }
    
    /**
     * Getter for property gas96.
     * @return Value of property gas96.
     */
    public double getGas96() {
        return gas96;
    }
    
    /**
     * Setter for property gas96.
     * @param gas96 New value of property gas96.
     */
    public void setGas96(double gas96) {
        this.gas96 = gas96;
    }
    
    /**
     * Getter for property elec96.
     * @return Value of property elec96.
     */
    public double getElec96() {
        return elec96;
    }
    
    /**
     * Setter for property elec96.
     * @param elec96 New value of property elec96.
     */
    public void setElec96(double elec96) {
        this.elec96 = elec96;
    }
    
    /**
     * Getter for property coal96.
     * @return Value of property coal96.
     */
    public double getCoal96() {
        return coal96;
    }
    
    /**
     * Setter for property coal96.
     * @param coal96 New value of property coal96.
     */
    public void setCoal96(double coal96) {
        this.coal96 = coal96;
    }
    
    /**
     * Getter for property wage96.
     * @return Value of property wage96.
     */
    public double getWage96() {
        return wage96;
    }
    
    /**
     * Setter for property wage96.
     * @param wage96 New value of property wage96.
     */
    public void setWage96(double wage96) {
        this.wage96 = wage96;
    }
    
    /**
     * Getter for property population.
     * @return Value of property population.
     */
    public double getPopulation() {
        return population;
    }
    
    /**
     * Setter for property population.
     * @param population New value of property population.
     */
    public void setPopulation(double population) {
        this.population = population;
    }
    
    /**
     * Getter for property stateIndex.
     * @return Value of property stateIndex.
     */
    public int getStateIndex() {
        return stateIndex;
    }
    
    /**
     * Setter for property stateIndex.
     * @param stateIndex New value of property stateIndex.
     */
    public void setStateIndex(int stateIndex) {
        this.stateIndex = stateIndex;
    }

    public Jama.Matrix getLastGuess() {
        return lastGuess;
    }

    public void setLastGuess(Jama.Matrix lastGuess) {
        this.lastGuess = lastGuess;
    }

    public boolean isEntered() {
        return entered;
    }

    public void setEntered(boolean entered) {
        this.entered = entered;
    }

    public boolean isExited() {
        return exited;
    }

    public void setExited(boolean exited) {
        this.exited = exited;
    }

}
