package nervecenter;

/*
 * thesisEstimationMain.java
 *
 * Created on February 9, 2004, 12:08 PM
 */
import estimation.dataManipulation.stageGameDataPrep;
import estimation.demand.estimateDemand;
import estimation.entry.estimateEntryPolicy;
import estimation.exit.estimateExitPolicy;
import estimation.production.stageGameCED;
import estimation.simulateW.simulateW;
import estimation.simulateW.simulateWResult;
import estimation.simulateWEntry.simulateWEntryCostsHeterogeneity;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Random;
import javax.swing.*;
import simulation.mpneSimulation.utility.simulationPrimitives;
import simulation.mpneSolver.SimulationResults;
import simulation.mpneSolver.centralControl;
import utility.demandCurve;
import utility.pmUtility;

/** This class integrates the estimation routines for my
 * thesis and passes the relevant data from one component
 * to another.
 * @author Stephen P. Ryan
 */
public class thesisEstimationMain {

    private static boolean COMPUTE_SOCIAL = false;

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        try {
            // UIManager.setLookAndFeel(new com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel());
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            e.printStackTrace();
        }

        cementGUI gui = new cementGUI();
        NumberFormat nf = NumberFormat.getInstance();
        nf.setMaximumFractionDigits(5);

        JDialog demandProductionTableDialog = new JDialog(gui, "Demand and Production Parameters");
        demandProductionTableDialog.setBounds(1000, 50, 500, 500);

        JDialog dynamicTableDialog = new JDialog(gui, "Dynamic Parameters");
        dynamicTableDialog.setBounds(50, 600, 1500, 500);

        gui.addString("LATE PERIOD: " + thesisConstants.LATE_PERIOD + "\n");
        gui.addString("Number of Simulations in Computing W: " + thesisConstants.NUM_SIMW + "\n");

        /**
         * Need some way to bootstrap all of the data in here, then run the whole
         * minimization process, collect results, including counterfactual outcomes.
         */
        int numBoots = 1;
        if (thesisConstants.SUBSAMPLE_PARAMETERS) {
            numBoots = 500;
        } else {
            thesisConstants.SUBSAMPLE_SIZE = 22;
        }

        JTable demandProductionTable = new JTable(new CementTableModel(numBoots, CementTableModel.DEMAND_PRODUCTION));
        demandProductionTableDialog.getContentPane().add(new JScrollPane(demandProductionTable));

        JTable dynamicTable = new JTable(new CementTableModel(numBoots, CementTableModel.DYNAMIC));
        dynamicTableDialog.getContentPane().add(new JScrollPane(dynamicTable));

        if (thesisConstants.SUBSAMPLE_PARAMETERS) {
            // demandProductionTableDialog.pack();
            demandProductionTableDialog.setVisible(true);
            // dynamicTableDialog.pack();
            dynamicTableDialog.setVisible(true);

            // thesisConstants.NUM_SIMW = 3000;
        }
        if (thesisConstants.TRUE_BOOTSTRAP) {
            thesisConstants.SUBSAMPLE_SIZE = 22;
        }

        Jama.Matrix demandParameters = new Jama.Matrix(numBoots, 2);
        /**
         * Collect information on estimated production parameters as well.
         */
        Jama.Matrix productionCosts = new Jama.Matrix(numBoots, 4);
        /**
         * 22 parameters here: six b-splines for investment, divestment, and exit each (+18)
         * investment cost and cost squared, and divestment cost and cost squared (+4)
         */
        Jama.Matrix dynamicParameters = new Jama.Matrix(numBoots, 22 + 6);
        /**
         * Six parameters for (implied) distributions
         * Investment distribution mean: 847 stddev: 130
         * Divestment distribution mean: 235,351 stddev: 114,010
         * Exit distribution mean: -70,881 stddev: 58,005
         */
        Random rngMaster = new Random(7478);
        /**
         * Two entry parameters: mu and sigma of the entry cost distribution
         * But they show up twice, since there is early and then late
         */
        Jama.Matrix entryParameters = new Jama.Matrix(numBoots, 4);
        /**
         * Several outcomes from the six counterfactual simulations
         * Times two for early and late
         */
        boolean computeCounterfactuals = true;

        Jama.Matrix[] counterfactualResultsDeNovo = new Jama.Matrix[2]; // (numBoots, 13);
        Jama.Matrix[] counterfactualResultsMature = new Jama.Matrix[2]; // (numBoots, 13);
        Jama.Matrix[] counterfactualResultsDeNovoSocial = new Jama.Matrix[2]; // (numBoots, 13);
        Jama.Matrix[] counterfactualResultsMatureSocial = new Jama.Matrix[2]; // (numBoots, 13);
        for (int t = 0; t < 2; t++) {
            counterfactualResultsDeNovo[t] = new Jama.Matrix(numBoots, 13);
            counterfactualResultsMature[t] = new Jama.Matrix(numBoots, 13);
            counterfactualResultsDeNovoSocial[t] = new Jama.Matrix(numBoots, 13);
            counterfactualResultsMatureSocial[t] = new Jama.Matrix(numBoots, 13);
        }

        ProgressMonitor mon = new ProgressMonitor(null, "Bootstrapping Dynamic Model", "0 / " + numBoots, 0, numBoots);

        for (int r = 0; r < numBoots; r++) {
            demandCurve priceThesis = new demandCurve();
            thesisConstants.FORCE_EARLY_DYNAMIC_PARAMETERS = true;
            // rngMaster = new Random(747);
            System.gc();
            /**
             * This is where we set up the bootstrap data.
             *
             * Modify this slightly so that we redraw samples if we get a demand slope smaller in absolute value than -1.
             */
            boolean validData = false;
            while (!validData) {
                if (!validData) {
                    System.out.print("***** Preparing Data *****\n");
                    stageGameDataPrep dataPreparation = new stageGameDataPrep(rngMaster.nextLong());
                    // stageGameDataPrep dataPreparation = new stageGameDataPrep(747400);
                    dataPreparation.prepareData();
                    /**
                     * Run the demand estimation on this subset of data.
                     */
                    estimateDemand demandEstimator = new estimateDemand(priceThesis);
                    demandEstimator.execute();
                    demandParameters.set(r, 0, priceThesis.getA());
                    demandParameters.set(r, 1, priceThesis.getB(0));
                    demandProductionTable.setValueAt(priceThesis.getA(), r, 0);
                    demandProductionTable.setValueAt(priceThesis.getB(0), r, 1);
                    /**
                     * Need to enforce that this slope is less than -1.  Otherwise
                     * MR is upward sloping, and this causes all kinds of havoc.
                     */
                    if (priceThesis.getB(0) < -1) {
                        validData = true;
                    }
                }
            }

            /**
             * Just need to perturb the data sets available to the following objects at this point
             * And we'll get our standard errors correct; want to re-sample from
             * all of the data sets in the paper.
             */
            // stage game estimation
            // returns the parameters mu and mc from the stage game
            System.out.print("\n***** Stage Estimation *****\n");
            stageGameCED stage = new stageGameCED(priceThesis);
            double[] estimatedCosts = stage.getParameters();
//            gui.addString(pmUtility.stringPrettyPrint(new Jama.Matrix(estimatedCosts, 1)) + "\n");
            for (int i = 0; i < estimatedCosts.length - 1; i++) {
                productionCosts.set(r, i, estimatedCosts[i + 1]);
                demandProductionTable.setValueAt(estimatedCosts[i + 1], r, 2 + i);
            }
            /**
             * Note: stage writes the parameters to stageParams.dat file; that is how
             * they are propagated to simulateW below
             * In the bootstrap case, pass directly, so that the same data is used
             * to compute both production costs and investment policy function.
             */
            // (s,S)-rule estimation
            // needs shocks input from the stage game estimation
            // returns the parameters governing the adjustment of the capital stock
            System.out.print("\n***** Policy Estimation *****\n");
            estimation.ssrulesLikelihood.ssrulesMain policy = new estimation.ssrulesLikelihood.ssrulesMain(stage.getData());
            /**
             * Can estimate the exit and entry policy functions here
             */
            System.out.println("Estimating exit policy");
            estimateExitPolicy exit = new estimateExitPolicy();
            exit.execute();

            System.out.println("Estimating entry policy");
            estimateEntryPolicy entry = new estimateEntryPolicy();
            entry.execute();

            // dynamic parameter estimation
            // needs the other parameters in the profit function to calculate profits
            // and the investment rule to determine the transition between states
            // returns investment costs and adjustment costs
            System.out.print("\n***** Dynamic Parameter Estimation *****\n");
            // simulateW dynamic = new simulateW(demand.getDemandParameters(), stage.getParameters(), policy.getParameters());
            simulateW dynamic = new simulateW(rngMaster.nextLong(), priceThesis, estimatedCosts, policy.getParameters(), exit.getExitParameters(), entry.getEntryParameters());
            System.out.println("Estimation finished.");
            simulateWResult dynamicParametersResult = dynamic.getParameters();
            gui.addString("SimulateW value: " + dynamicParametersResult.getObjectiveFunction() + "\n");
            for (int i = 0; i < 22; i++) {
                dynamicParameters.set(r, i, dynamicParametersResult.getGuess()[i + 1]);
            }
            for (int i = 0; i < 2; i++) {
                dynamicParameters.set(r, 22 + i, dynamicParametersResult.getiDist()[i + 1]);
            }
            for (int i = 0; i < 2; i++) {
                dynamicParameters.set(r, 24 + i, dynamicParametersResult.getdDist()[i + 1]);
            }
            for (int i = 0; i < 2; i++) {
                dynamicParameters.set(r, 26 + i, dynamicParametersResult.getxDist()[i + 1]);
            }
            dynamicTable.setValueAt(dynamicParameters.get(r, 6), r, 0);
            for (int i = 0; i < 6; i++) {
                dynamicTable.setValueAt(dynamicParameters.get(r, 22 + i), r, i + 1);
            }
            mon.setProgress(r + 1);
            mon.setNote((r + 1) + " / " + numBoots);
//            double variablePlantCost = result.getGuess()[7] * 1500;
//            gui.addString("x: " + result.getGuess()[7] + " Cost of plant: " + nf.format(variablePlantCost) + "\n");


            for (int t = 0; t < 2; t++) {
                /**
                 * New thoughts for getting standard errors right.  In the case where
                 * we restrict the parameters to be the same for early and late,
                 * it would be nice to control for numerical variance in those estimates
                 * by using the same parameters for early and late entry costs.
                 *
                 * So run the entry cost calculations and simulations twice; once
                 * with the early flag on and once with the late flag on.
                 */
                if (t == 1) {
                    thesisConstants.FORCE_EARLY_DYNAMIC_PARAMETERS = false;
                }
                
                simulateWEntryCostsHeterogeneity entryCostEstimator = new simulateWEntryCostsHeterogeneity(priceThesis, estimatedCosts, policy.getParameters(), exit.getExitParameters(), entry.getEntryParameters(), dynamicParameters.getMatrix(r, r, 0, 27));
                entryParameters.set(r, 0 + t * 2, entryCostEstimator.getEstimatedMu());
                entryParameters.set(r, 1 + t * 2, entryCostEstimator.getEstimatedSigma());
                dynamicTable.setValueAt(entryCostEstimator.getEstimatedMu(), r, 7 + t * 2);
                dynamicTable.setValueAt(entryCostEstimator.getEstimatedSigma(), r, 8 + t * 2);

                /**
                 * This is the place to put calls to the counterfactual solver and
                 * simulations.  Collect those outputs here as well, so we can get
                 * standard errors on those outcomes.
                 *
                 * Need 6 counterfactuals: new market and incumbent market
                 * Split 3 ways: high sunk costs, low sunk costs, and social planner
                 *
                 * Okay, four of the counterfactuals only apply when the period is LATE
                 * The other two are for comparison purposes
                 */
                if (computeCounterfactuals) {
                    simulationPrimitives.SOCIAL_PLANNER = false;
                    centralControl counterfactualControl = new centralControl(priceThesis,
                            estimatedCosts, policy.getParameters(), dynamicParametersResult, entryCostEstimator.getEstimatedMu(), entryCostEstimator.getEstimatedSigma());
                    counterfactualControl.solve();
                    SimulationResults cfDeNovo = counterfactualControl.getSimulationResults(centralControl.NEW_MARKET);
                    SimulationResults cfMature = counterfactualControl.getSimulationResults(centralControl.MATURE_MARKET);
                    /**
                     * Only compute the social planner when in the late period
                     */
                    SimulationResults cfSocialDeNovo = null;
                    SimulationResults cfSocialMature = null;
                    if (thesisConstants.LATE_PERIOD && COMPUTE_SOCIAL) {
                        simulationPrimitives.SOCIAL_PLANNER = true;
                        centralControl counterfactualControlSocial = new centralControl(priceThesis,
                                estimatedCosts, policy.getParameters(), dynamicParametersResult, entryCostEstimator.getEstimatedMu(), entryCostEstimator.getEstimatedSigma());
                        counterfactualControlSocial.solve();
                        cfSocialDeNovo = counterfactualControlSocial.getSimulationResults(centralControl.NEW_MARKET);
                        cfSocialMature = counterfactualControlSocial.getSimulationResults(centralControl.MATURE_MARKET);
                        simulationPrimitives.SOCIAL_PLANNER = true;

                        counterfactualResultsDeNovoSocial[t].set(r, 0, cfSocialDeNovo.getAverageProducerProfit());
                        counterfactualResultsDeNovoSocial[t].set(r, 1, cfSocialDeNovo.getProfitFirm1());
                        counterfactualResultsDeNovoSocial[t].set(r, 2, cfSocialDeNovo.getSimCS());
                        counterfactualResultsDeNovoSocial[t].set(r, 3, cfSocialDeNovo.getTotalWelfare());
                        counterfactualResultsDeNovoSocial[t].set(r, 4, cfSocialDeNovo.getNofirms());
                        counterfactualResultsDeNovoSocial[t].set(r, 5, cfSocialDeNovo.getOnefirm());
                        counterfactualResultsDeNovoSocial[t].set(r, 6, cfSocialDeNovo.getTwofirms());
                        counterfactualResultsDeNovoSocial[t].set(r, 7, cfSocialDeNovo.getThreefirms());
                        counterfactualResultsDeNovoSocial[t].set(r, 8, cfSocialDeNovo.getFourfirms());
                        counterfactualResultsDeNovoSocial[t].set(r, 9, cfSocialDeNovo.getAverageSizeOverall());
                        counterfactualResultsDeNovoSocial[t].set(r, 10, cfSocialDeNovo.getMarketSizeOverall());
                        counterfactualResultsDeNovoSocial[t].set(r, 11, cfSocialDeNovo.getQuantityProduced());
                        counterfactualResultsDeNovoSocial[t].set(r, 12, cfSocialDeNovo.getAveragePrice());

                        counterfactualResultsMatureSocial[t].set(r, 0, cfSocialMature.getAverageProducerProfit());
                        counterfactualResultsMatureSocial[t].set(r, 1, cfSocialMature.getProfitFirm1());
                        counterfactualResultsMatureSocial[t].set(r, 2, cfSocialMature.getSimCS());
                        counterfactualResultsMatureSocial[t].set(r, 3, cfSocialMature.getTotalWelfare());
                        counterfactualResultsMatureSocial[t].set(r, 4, cfSocialMature.getNofirms());
                        counterfactualResultsMatureSocial[t].set(r, 5, cfSocialMature.getOnefirm());
                        counterfactualResultsMatureSocial[t].set(r, 6, cfSocialMature.getTwofirms());
                        counterfactualResultsMatureSocial[t].set(r, 7, cfSocialMature.getThreefirms());
                        counterfactualResultsMatureSocial[t].set(r, 8, cfSocialMature.getFourfirms());
                        counterfactualResultsMatureSocial[t].set(r, 9, cfSocialMature.getAverageSizeOverall());
                        counterfactualResultsMatureSocial[t].set(r, 10, cfSocialMature.getMarketSizeOverall());
                        counterfactualResultsMatureSocial[t].set(r, 11, cfSocialMature.getQuantityProduced());
                        counterfactualResultsMatureSocial[t].set(r, 12, cfSocialMature.getAveragePrice());
                    }

                    counterfactualResultsDeNovo[t].set(r, 0, cfDeNovo.getAverageProducerProfit());
                    counterfactualResultsDeNovo[t].set(r, 1, cfDeNovo.getProfitFirm1());
                    counterfactualResultsDeNovo[t].set(r, 2, cfDeNovo.getSimCS());
                    counterfactualResultsDeNovo[t].set(r, 3, cfDeNovo.getTotalWelfare());
                    counterfactualResultsDeNovo[t].set(r, 4, cfDeNovo.getNofirms());
                    counterfactualResultsDeNovo[t].set(r, 5, cfDeNovo.getOnefirm());
                    counterfactualResultsDeNovo[t].set(r, 6, cfDeNovo.getTwofirms());
                    counterfactualResultsDeNovo[t].set(r, 7, cfDeNovo.getThreefirms());
                    counterfactualResultsDeNovo[t].set(r, 8, cfDeNovo.getFourfirms());
                    counterfactualResultsDeNovo[t].set(r, 9, cfDeNovo.getAverageSizeOverall());
                    counterfactualResultsDeNovo[t].set(r, 10, cfDeNovo.getMarketSizeOverall());
                    counterfactualResultsDeNovo[t].set(r, 11, cfDeNovo.getQuantityProduced());
                    counterfactualResultsDeNovo[t].set(r, 12, cfDeNovo.getAveragePrice());

                    counterfactualResultsMature[t].set(r, 0, cfMature.getAverageProducerProfit());
                    counterfactualResultsMature[t].set(r, 1, cfMature.getProfitFirm1());
                    counterfactualResultsMature[t].set(r, 2, cfMature.getSimCS());
                    counterfactualResultsMature[t].set(r, 3, cfMature.getTotalWelfare());
                    counterfactualResultsMature[t].set(r, 4, cfMature.getNofirms());
                    counterfactualResultsMature[t].set(r, 5, cfMature.getOnefirm());
                    counterfactualResultsMature[t].set(r, 6, cfMature.getTwofirms());
                    counterfactualResultsMature[t].set(r, 7, cfMature.getThreefirms());
                    counterfactualResultsMature[t].set(r, 8, cfMature.getFourfirms());
                    counterfactualResultsMature[t].set(r, 9, cfMature.getAverageSizeOverall());
                    counterfactualResultsMature[t].set(r, 10, cfMature.getMarketSizeOverall());
                    counterfactualResultsMature[t].set(r, 11, cfMature.getQuantityProduced());
                    counterfactualResultsMature[t].set(r, 12, cfMature.getAveragePrice());
                }
            }
        }


        String[] demandNames = {"Intercept", "Slope"};
        for (int i = 0; i < 2; i++) {
            // System.out.println(productionNames[i] + ": " + nf.format(pmUtility.mean(productionCosts, i)) + " (" + nf.format(pmUtility.standardDeviation(productionCosts, i)) + ")");
            System.out.println(demandNames[i] + ": mean: " + nf.format(pmUtility.mean(demandParameters, i)) + " median: " + nf.format(pmUtility.median(demandParameters, i)) + " se: " + nf.format(pmUtility.standardDeviation(demandParameters, i)) + " iqr: " + nf.format(pmUtility.IQR(demandParameters, i)));
            gui.addString(demandNames[i] + ": mean: " + nf.format(pmUtility.mean(demandParameters, i)) + " median: " + nf.format(pmUtility.median(demandParameters, i)) + " se: " + nf.format(pmUtility.standardDeviation(demandParameters, i)) + " iqr: " + nf.format(pmUtility.IQR(demandParameters, i)));
            gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(demandParameters, i, 0.05)) + ", " + nf.format(pmUtility.percentile(demandParameters, i, 0.95)));
            gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(demandParameters, i, 0.025)) + ", " + nf.format(pmUtility.percentile(demandParameters, i, 0.975)) + "\n");
        }

        /**
         * Output production parameters.
         */
        String[] productionNames = {"Capacity Cost", "Binding Level", "Marginal Cost", "CA Shifter", "Capacity Cost Post 1990 Shifter", "Binding Level Post 1990 Shifter", "Marginal Cost Post 1990 Shifter"};
        for (int i = 0; i < 4; i++) {
            // System.out.println(productionNames[i] + ": " + nf.format(pmUtility.mean(productionCosts, i)) + " (" + nf.format(pmUtility.standardDeviation(productionCosts, i)) + ")");
            System.out.println(productionNames[i] + ": mean: " + nf.format(pmUtility.mean(productionCosts, i)) + " median: " + nf.format(pmUtility.median(productionCosts, i)) + " se: " + nf.format(pmUtility.standardDeviation(productionCosts, i)) + " iqr: " + nf.format(pmUtility.IQR(productionCosts, i)));
            gui.addString(productionNames[i] + ": mean: " + nf.format(pmUtility.mean(productionCosts, i)) + " median: " + nf.format(pmUtility.median(productionCosts, i)) + " se: " + nf.format(pmUtility.standardDeviation(productionCosts, i)) + " iqr: " + nf.format(pmUtility.IQR(productionCosts, i)));
            gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(productionCosts, i, 0.05)) + ", " + nf.format(pmUtility.percentile(productionCosts, i, 0.95)));
            gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(productionCosts, i, 0.025)) + ", " + nf.format(pmUtility.percentile(productionCosts, i, 0.975)) + "\n");
        }

        String[] names = {"Investment B-Spline 1", "Investment B-Spline 2", "Investment B-Spline 3", "Investment B-Spline 4", "Investment B-Spline 5", "Investment B-Spline 6",
            "Investment Cost", "Investment Cost Squared",
            "Divestment B-Spline 1", "Divestment B-Spline 2", "Divestment B-Spline 3", "Divestment B-Spline 4", "Divestment B-Spline 5", "Divestment B-Spline 6",
            "Divestment Cost", "Divestment Cost Squared",
            "Exit B-Spline 1", "Exit B-Spline 2", "Exit B-Spline 3", "Exit B-Spline 4", "Exit B-Spline 5", "Exit B-Spline 6", "Investment Dist Mean", "Investment Dist StdDev",
            "Divestment Dist Mean", "Divestment Dist StdDev", "Exit Dist Mean", "Exit Dist StdDev"
        };
        for (int i = 0; i < 28; i++) {
            System.out.println(names[i] + ": mean: " + nf.format(pmUtility.mean(dynamicParameters, i)) + " median: " + nf.format(pmUtility.median(dynamicParameters, i)) + " se: " + nf.format(pmUtility.standardDeviation(dynamicParameters, i)) + " iqr: " + nf.format(pmUtility.IQR(dynamicParameters, i)));
            gui.addString(names[i] + ": mean: " + nf.format(pmUtility.mean(dynamicParameters, i)) + " median: " + nf.format(pmUtility.median(dynamicParameters, i)) + " se: " + nf.format(pmUtility.standardDeviation(dynamicParameters, i)) + " iqr: " + nf.format(pmUtility.IQR(dynamicParameters, i)));
            gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(dynamicParameters, i, 0.05)) + ", " + nf.format(pmUtility.percentile(dynamicParameters, i, 0.95)));
            gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(dynamicParameters, i, 0.025)) + ", " + nf.format(pmUtility.percentile(dynamicParameters, i, 0.975)) + "\n");
        }

        String[] entryNames = {"Entry Mu Early", "Entry Sigma Early", "Entry Mu Late", "Entry Sigma Late"};
        for (int i = 0; i < 4; i++) {
            System.out.println(entryNames[i] + ": mean: " + nf.format(pmUtility.mean(entryParameters, i)) + " median: " + nf.format(pmUtility.median(entryParameters, i)) + " se: " + nf.format(pmUtility.standardDeviation(entryParameters, i)) + " iqr: " + nf.format(pmUtility.IQR(entryParameters, i)));
            gui.addString(entryNames[i] + ": mean: " + nf.format(pmUtility.mean(entryParameters, i)) + " median: " + nf.format(pmUtility.median(entryParameters, i)) + " se: " + nf.format(pmUtility.standardDeviation(entryParameters, i)) + " iqr: " + nf.format(pmUtility.IQR(entryParameters, i)));
            gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(entryParameters, i, 0.05)) + ", " + nf.format(pmUtility.percentile(entryParameters, i, 0.95)));
            gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(entryParameters, i, 0.025)) + ", " + nf.format(pmUtility.percentile(entryParameters, i, 0.975)) + "\n");
        }

        Jama.Matrix entryDiff = new Jama.Matrix(numBoots, 2);
        for (int i = 0; i < entryParameters.getRowDimension(); i++) {
            entryDiff.set(i, 0, entryParameters.get(i, 2) - entryParameters.get(i, 0));
            entryDiff.set(i, 1, entryParameters.get(i, 3) - entryParameters.get(i, 1));
        }

        String[] entryDiffNames = {"Entry Mu Difference", "Entry Sigma Difference"};
        for (int i = 0; i < 2; i++) {
            System.out.println(entryDiffNames[i] + ": mean: " + nf.format(pmUtility.mean(entryDiff, i)) + " median: " + nf.format(pmUtility.median(entryDiff, i)) + " se: " + nf.format(pmUtility.standardDeviation(entryDiff, i)) + " iqr: " + nf.format(pmUtility.IQR(entryDiff, i)));
            gui.addString(entryDiffNames[i] + ": mean: " + nf.format(pmUtility.mean(entryDiff, i)) + " median: " + nf.format(pmUtility.median(entryDiff, i)) + " se: " + nf.format(pmUtility.standardDeviation(entryDiff, i)) + " iqr: " + nf.format(pmUtility.IQR(entryDiff, i)));
            gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(entryDiff, i, 0.05)) + ", " + nf.format(pmUtility.percentile(entryDiff, i, 0.95)));
            gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(entryDiff, i, 0.025)) + ", " + nf.format(pmUtility.percentile(entryDiff, i, 0.975)) + "\n");
        }

        if (computeCounterfactuals) {
            String[] counterfactualNames = {"Total Producer Profit (\\$ in NPV) & ",
                "Profit Firm 1 (\\$ in NPV)& ",
                "Total Net Consumer Surplus (\\$ in NPV) & ",
                "Total Welfare  (\\$ in NPV) & ",
                "Periods with no firms (periods) &",
                "Periods with one firm (periods) &",
                "Periods with two firms (periods) & ",
                "Periods with three firms (periods) & ",
                "Periods with four firms (periods) & ",
                "Average Size of Active Firm (tons) &",
                "Average Market Capacity (tons) &",
                "Average Market Quantity (tons) &",
                "Average Market Price"};

            for (int t = 0; t < 2; t++) {
                String modifier = "early";
                if (t == 1) {
                    modifier = "late";
                }
                System.out.println("\nDeNovo Market: " + modifier);
                gui.addString("\nDeNovo Market: " + modifier + "\n");
                for (int i = 0; i < counterfactualNames.length; i++) {
                    System.out.println(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsDeNovo[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsDeNovo[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsDeNovo[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsDeNovo[t], i)));
                    gui.addString(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsDeNovo[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsDeNovo[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsDeNovo[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsDeNovo[t], i)));
                    gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsDeNovo[t], i, 0.05)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsDeNovo[t], i, 0.95)));
                    gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsDeNovo[t], i, 0.025)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsDeNovo[t], i, 0.975)) + "\n");
                }

                System.out.println("\nMature Market: " + modifier);
                gui.addString("\nMature Market: " + modifier + "\n");
                for (int i = 0; i < counterfactualNames.length; i++) {
                    System.out.println(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsMature[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsMature[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsMature[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsMature[t], i)));
                    gui.addString(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsMature[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsMature[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsMature[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsMature[t], i)));
                    gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsMature[t], i, 0.05)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsMature[t], i, 0.95)));
                    gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsMature[t], i, 0.025)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsMature[t], i, 0.975)) + "\n");
                }

                if (thesisConstants.LATE_PERIOD && COMPUTE_SOCIAL) {
                    System.out.println("Social DeNovo Market: " + modifier);
                    gui.addString("Social DeNovo Market: " + modifier + "\n");
                    for (int i = 0; i < counterfactualNames.length; i++) {
                        System.out.println(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsDeNovoSocial[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsDeNovoSocial[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsDeNovoSocial[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsDeNovoSocial[t], i)));
                        gui.addString(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsDeNovoSocial[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsDeNovoSocial[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsDeNovoSocial[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsDeNovoSocial[t], i)));
                        gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsDeNovoSocial[t], i, 0.05)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsDeNovoSocial[t], i, 0.95)));
                        gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsDeNovoSocial[t], i, 0.025)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsDeNovoSocial[t], i, 0.975)) + "\n");
                    }

                    System.out.println("Social Mature Market: " + modifier);
                    gui.addString("Social Mature Market: " + modifier + "\n");
                    for (int i = 0; i < counterfactualNames.length; i++) {
                        System.out.println(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsMatureSocial[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsMatureSocial[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsMatureSocial[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsMatureSocial[t], i)));
                        gui.addString(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(counterfactualResultsMatureSocial[t], i)) + " median: " + nf.format(pmUtility.median(counterfactualResultsMatureSocial[t], i)) + " se: " + nf.format(pmUtility.standardDeviation(counterfactualResultsMatureSocial[t], i)) + " iqr: " + nf.format(pmUtility.IQR(counterfactualResultsMatureSocial[t], i)));
                        gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsMatureSocial[t], i, 0.05)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsMatureSocial[t], i, 0.95)));
                        gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(counterfactualResultsMatureSocial[t], i, 0.025)) + ", " + nf.format(pmUtility.percentile(counterfactualResultsMatureSocial[t], i, 0.975)) + "\n");
                    }
                }
            }


            /**
             * Output the differences in the counterfactuals for standard error
             * purposes.
             */
            String modifier = "Differences";
            Jama.Matrix cfnewDiff = counterfactualResultsDeNovo[1].minus(counterfactualResultsDeNovo[0]);

            System.out.println("\nDeNovo Market: " + modifier);
            gui.addString("\nDeNovo Market: " + modifier + "\n");
            for (int i = 0; i < counterfactualNames.length; i++) {
                System.out.println(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(cfnewDiff, i)) + " median: " + nf.format(pmUtility.median(cfnewDiff, i)) + " se: " + nf.format(pmUtility.standardDeviation(cfnewDiff, i)) + " iqr: " + nf.format(pmUtility.IQR(cfnewDiff, i)));
                gui.addString(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(cfnewDiff, i)) + " median: " + nf.format(pmUtility.median(cfnewDiff, i)) + " se: " + nf.format(pmUtility.standardDeviation(cfnewDiff, i)) + " iqr: " + nf.format(pmUtility.IQR(cfnewDiff, i)));
                gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(cfnewDiff, i, 0.05)) + ", " + nf.format(pmUtility.percentile(cfnewDiff, i, 0.95)));
                gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(cfnewDiff, i, 0.025)) + ", " + nf.format(pmUtility.percentile(cfnewDiff, i, 0.975)) + "\n");
            }

            Jama.Matrix cfMatureDiff = counterfactualResultsMature[1].minus(counterfactualResultsMature[0]);
            System.out.println("\nMature Market: " + modifier);
            gui.addString("\nMature Market: " + modifier + "\n");
            for (int i = 0; i < counterfactualNames.length; i++) {
                System.out.println(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(cfMatureDiff, i)) + " median: " + nf.format(pmUtility.median(cfMatureDiff, i)) + " se: " + nf.format(pmUtility.standardDeviation(cfMatureDiff, i)) + " iqr: " + nf.format(pmUtility.IQR(cfMatureDiff, i)));
                gui.addString(counterfactualNames[i] + ": mean: " + nf.format(pmUtility.mean(cfMatureDiff, i)) + " median: " + nf.format(pmUtility.median(cfMatureDiff, i)) + " se: " + nf.format(pmUtility.standardDeviation(cfMatureDiff, i)) + " iqr: " + nf.format(pmUtility.IQR(cfMatureDiff, i)));
                gui.addString("\t90\\% CI: " + nf.format(pmUtility.percentile(cfMatureDiff, i, 0.05)) + ", " + nf.format(pmUtility.percentile(cfMatureDiff, i, 0.95)));
                gui.addString("\t95\\% CI: " + nf.format(pmUtility.percentile(cfMatureDiff, i, 0.025)) + ", " + nf.format(pmUtility.percentile(cfMatureDiff, i, 0.975)) + "\n");
            }
        }

        /**
         * Save outputs to use as inputs in the counterfactual simulations.
         */
        if (numBoots > 1) {
            try {
                Date today = new Date();
                SimpleDateFormat dateFormat = new SimpleDateFormat("MMM.dd.yy.hh.mm.ss");
                String filename = "bootResults" + "_" + dateFormat.format(today);
                System.out.println("Writing " + filename + ".dat");
                if (thesisConstants.LATE_PERIOD) {
                    filename = filename.concat("late");
                } else {
                    filename = filename.concat("early");
                }
                ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream("data/" + filename + ".dat"));
                out.writeObject(demandParameters);
                out.writeObject(productionCosts);
                out.writeObject(dynamicParameters);
                out.writeObject(entryParameters);
                out.writeObject(counterfactualResultsDeNovo);
                out.writeObject(counterfactualResultsDeNovoSocial);
                out.writeObject(counterfactualResultsMature);
                out.writeObject(counterfactualResultsMatureSocial);
                out.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
}
