/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package simulation.mpneSimulation;

import JSci.maths.statistics.NormalDistribution;
import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.Callable;
import simulation.mpneSimulation.utility.simulationPrimitives;
import simulation.mpneSolver.Interpolate;
import simulation.mpneSolver.coefficientValues;
import simulation.mpneSolver.stateSpace;
import simulation.mpneSolver.stateSpacePoint;

/**
 *
 * @author Stephen P. Ryan
 */
public class SimulationTask implements Callable<SimulationTaskResult> {

    private int timeMax = simulationPrimitives.SIMULATION_LENGTH;
    private int numFirms = simulationPrimitives.NUM_FIRMS;
    NormalDistribution standardNormal = new NormalDistribution();
    private Jama.Matrix ws;
    // policy coefficients
    private coefficientValues coefInvestActive;
    private coefficientValues coefInvestEntrant;
    private coefficientValues coefExit;
    private coefficientValues coefEntryValueThreshold;
    private coefficientValues coefExpectedMax;
    // market outcomes coefficients
    private coefficientValues coefProfit;
    private coefficientValues coefWelfare;
    private coefficientValues coefQuantity;
    private coefficientValues coefPrice;
    // value function coefficients
    private coefficientValues coefValueDoingNothing;
    private coefficientValues coefValueOptimalInvestment;
    Random rng;
    Random entryShock;
    Random exitShock;
    Random investShock;
    stateSpace s;

    public SimulationTask(stateSpace s, long seed) {
        this.s = s;
        rng = new Random(seed);

        entryShock = new Random(rng.nextLong());
        exitShock = new Random(rng.nextLong());
        investShock = new Random(rng.nextLong());

        Jama.Matrix w;
        w = new Jama.Matrix(numFirms, 2, 0);
        for (int i = 0; i < numFirms; i++) {
            w.set(i, 0, simulationPrimitives.CAPACITY[i]);
            w.set(i, 1, simulationPrimitives.TYPE[i]);
            if (simulationPrimitives.CAPACITY[i] == 0) {
                w.set(i, 1, 0);
            }
        }

        ArrayList<stateSpacePoint> list = s.getList();
        determineCoefficients(list, s);

        ws = w.copy();
    }

    public SimulationTaskResult call() throws Exception {
        double[] runProfit = new double[numFirms];
        double runCS = 0;
        double runQuantity = 0;
        for (int i = 0; i < numFirms; i++) {
            runProfit[i] = 0;
        }
        double runOneFirm = 0;
        double runTwoFirms = 0;
        double runNoFirms = 0;
        double runThreeFirms = 0;
        double runFourFirms = 0;
        double runFiveFirms = 0;
        double runSixFirms = 0;
        double averageSize = 0;
        double marketSize = 0;
        double periodsWithAnyFirms = 0;
        double averagePriceRun = 0;
        double numEntries = 0;
        double numPotentialEntries = 0;
        int N = simulationPrimitives.CAPACITY_POINTS;

        for (int time = 0; time < timeMax; time++) {
            // System.out.print("t: " + time + " ");
            // pmUtility.prettyPrintVector(ws);
            boolean[] exit = new boolean[numFirms];
            double[] investment = new double[numFirms];
            double[] profit = new double[numFirms];
            boolean[] entry = new boolean[numFirms];
            double consumerCS = 0;

            for (int j = 0; j < numFirms; j++) {
                Jama.Matrix swap = ws.copy();
                swap.set(j, 0, ws.get(0, 0));
                swap.set(0, 0, ws.get(j, 0));
                swap.set(j, 1, ws.get(0, 1));
                swap.set(0, 1, ws.get(j, 1));
                int numActive = 0;
                int ownActive = 0;
                if (swap.get(0, 0) > 0) {
                    ownActive = 1;
                }
                for (int k = 0; k < numFirms; k++) {
                    if (swap.get(k, 0) > 0) {
                        numActive += 1;
                    }
                }
                Jama.Matrix wsActive = new Jama.Matrix(numActive, 1);
                int count = 0;
                for (int k = 0; k < numFirms; k++) {
                    if (swap.get(k, 0) > 0) {
                        wsActive.set(count, 0, swap.get(k, 0));
                        count += 1;
                    }
                }
                int offset = s.getExpandedIndex(swap.getMatrix(0, numFirms - 1, 1, 1));
                int indexC = numActive - 1;
                // Compute common aggregate measures
                if (j == 0) {
                    if (numActive > 0) { // need to fix quantity computation
                        periodsWithAnyFirms++;
                        runQuantity += Interpolate.getValue(wsActive, coefQuantity.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                                numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                        averagePriceRun += Interpolate.getValue(wsActive, coefPrice.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                                numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                        consumerCS = Interpolate.getValue(wsActive, coefWelfare.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                                numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                        // System.out.println("runQuantity: "+runQuantity);
                    }
                }
                /**
                 * Resolve optimal discrete choice here.  Draw values of private
                 * information, figure out which option (invest, exit, do nothing)
                 * gives highest payoff, then go with that.
                 */
                double drawGamma = s.getADJUSTMENT_MU() + s.getADJUSTMENT_SIGMA() * standardNormal.inverse(investShock.nextDouble());
                double drawPhi = s.getSCRAP_MU() + s.getSCRAP_SIGMA() * standardNormal.inverse(exitShock.nextDouble());
                double valueOptimalInvestment = -drawGamma;
                double valueDoingNothing = 0;
                
                double expectedMaxOptions = 0;

                /**
                 * Determine optimal discrete choice values.
                 */
                if (ws.get(j, 0) > 0) {
                    indexC = numActive - 1;
                    valueOptimalInvestment += Interpolate.getValue(wsActive, coefValueOptimalInvestment.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                            numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                    valueDoingNothing += Interpolate.getValue(wsActive, coefValueDoingNothing.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                            numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                    expectedMaxOptions += Interpolate.getValue(wsActive, coefExpectedMax.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                            numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                    indexC = numActive - 1;
                    profit[j] = Interpolate.getValue(wsActive, coefProfit.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                            numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                } else {
                    profit[j] = 0;
                }

                /**
                 * Figure out optimal investments if we are going to make one.
                 */
                if (ownActive > 0) {
                    indexC = numActive - 1;
                    investment[j] = Interpolate.getValue(wsActive, coefInvestActive.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                            numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                } else if (ownActive == 0 && numActive > 0) {
                    indexC = numActive;
                    // System.out.println("indexC: "+indexC+" rowEnd: "+((int) java.lang.Math.pow(N + 2, numActive) - 1)+" rowdim: "+coefInvestEntrant.getCoefficients(offset).getRowDimension()+" coldim: "+coefInvestEntrant.getCoefficients(offset).getColumnDimension());
                    investment[j] = Interpolate.getValue(wsActive, coefInvestEntrant.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                            numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                } else {
                    indexC = numActive;
                    investment[j] = coefInvestEntrant.getCoefficients(offset).get(0, indexC);
                }

                /**
                 * Censor negative adjustments.
                 */
                if (investment[j] < 0) {
                    investment[j] = 0;
                }

                if (ws.get(j, 0) > 0) {
                    /**
                     *  Case: Incumbent firms
                     */
                    if (drawPhi > expectedMaxOptions) {
                        /**
                         * Firm exits.  Give the firm scrap value and indicate to the
                         * program that the firm is gone.
                         */
                        exit[j] = true;
                        profit[j] += drawPhi;
                    } else {
                        if (valueOptimalInvestment > valueDoingNothing) {
                            /**
                             * Firm makes optimal positive investment.  For some reason
                             * I reckon all the investments below, so just give the
                             * fixed cost right now.
                             */
                            profit[j] -= drawGamma;
                            exit[j] = false;
                        }
                        else {
                            /**
                             * Firm does nothing, stays pat.  Drop investment to zero.
                             */
                            investment[j] = 0;
                            exit[j] = false;
                        }
                    }
                }

                /**
                 * Case: Potential Entrants
                 */
                if (ws.get(j, 0) == 0) {
                    numPotentialEntries++;
                    double entryDraw = s.getENTRY_MU() + s.getENTRY_SIGMA() * standardNormal.inverse(entryShock.nextDouble());
                    double entryValueCutoff = 0;
                    if (numActive > 0) {
                        entryValueCutoff = Interpolate.getValue(wsActive, coefEntryValueThreshold.getCoefficients(offset).getMatrix(0, (int) java.lang.Math.pow(N + 2, numActive) - 1, indexC, indexC),
                                numActive, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                    } else {
                        entryValueCutoff = coefEntryValueThreshold.getCoefficients(offset).get(0, indexC);
                    }
                    // System.out.println("EntryDraw: " + entryDraw + " EntryValue: " + entryValueCutoff);
                    if (entryDraw + drawGamma < entryValueCutoff) {
                        entry[j] = true;
                        profit[j] -= drawGamma;
                        profit[j] -= entryDraw;
                        numEntries++;
                        // System.out.println("Showing an entry at investment level: " + investment[j]);
                    } else {
                        entry[j] = false;
                        investment[j] = 0;
                    }
                }
            }

            int active = 0;
            double sumSize = 0;
            //ws.print(0,0);
            for (int i = 0; i < numFirms; i++) {
                if (ws.get(i, 0) > 0) {
                    active++;
                    sumSize += ws.get(i, 0);
                }
            }
            averageSize += sumSize / Math.max(1, active * timeMax);
            marketSize += sumSize / timeMax;
            if (active == 0) {
                runNoFirms++;
            }
            if (active == 1) {
                runOneFirm++;
            }
            if (active == 2) {
                runTwoFirms++;
            }
            if (numFirms > 2) {
                if (active == 3) {
                    runThreeFirms++;
                }
            }
            if (numFirms > 3) {
                if (active == 4) {
                    runFourFirms++;
                }
            }
            if (numFirms > 4) {
                if (active == 5) {
                    runFiveFirms++;
                }
            }
            if (numFirms > 5) {
                if (active == 6) {
                    runSixFirms++;
                }
            }
            // update the state now that i have a vector describing the payoffs and policies of all the firms
            double[] outcomes = new double[numFirms];
            double[] types = new double[numFirms];
            for (int j = 0; j < numFirms; j++) {
                if ((ws.get(j, 0) > 0 || entry[j]) && !exit[j]) {
                    outcomes[j] = ws.get(j, 0) + investment[j];
                } else {
                    outcomes[j] = 0;
                }
                if (exit[j]) {
                    types[j] = 0;
                } else {
                    types[j] = ws.get(j, 1);
                }
                if (investment[j] > 0 && (ws.get(j, 0) > 0 || entry[j])) {
                    profit[j] -= investment[j] * s.getINVESTMENT_MARGINAL_COST();
                    profit[j] -= Math.pow(investment[j], 2) * s.getINVESTMENT_MARGINAL_COST2();
                }
            }
            Jama.Matrix outcomesMat = new Jama.Matrix(outcomes, numFirms);
            Jama.Matrix typesMat = new Jama.Matrix(types, numFirms);
            ws.setMatrix(0, numFirms - 1, 0, 0, outcomesMat);
            ws.setMatrix(0, numFirms - 1, 1, 1, typesMat);
            for (int i = 0; i < numFirms; i++) {
                runProfit[i] += Math.pow(simulationPrimitives.DISCOUNT_FACTOR, time) * profit[i];
            }
            runCS += Math.pow(simulationPrimitives.DISCOUNT_FACTOR, time) * consumerCS;
            // System.out.println("runQuantity: "+runQuantity);
        }
        runQuantity /= timeMax;
        averagePriceRun /= timeMax;
        // System.out.println("runQuantity: "+runQuantity);

        // System.out.println("Num entries: " + numEntries + " Num potential entries: " + numPotentialEntries + " Rate: " + (numEntries / numPotentialEntries));

        return new SimulationTaskResult(runProfit, runQuantity, averageSize, marketSize, runCS, runNoFirms, runOneFirm, runTwoFirms, runThreeFirms, runFourFirms, runFiveFirms, runSixFirms, averagePriceRun);
    }

    private void determineCoefficients(ArrayList<stateSpacePoint> list, stateSpace s) {
        // compute coefficients for each of the policies
        Jama.Matrix Invest = new Jama.Matrix(list.size(), 1);
        Jama.Matrix ValueOptimalInvestment = new Jama.Matrix(list.size(), 1);
        Jama.Matrix ValueDoingNothing = new Jama.Matrix(list.size(), 1);
        Jama.Matrix EntryValueCutoff = new Jama.Matrix(list.size(), 1);
        Jama.Matrix Exit = new Jama.Matrix(list.size(), 1);
        Jama.Matrix Profit = new Jama.Matrix(list.size(), 1);
        Jama.Matrix Welfare = new Jama.Matrix(list.size(), 1);
        Jama.Matrix Quantity = new Jama.Matrix(list.size(), 1);
        Jama.Matrix Price = new Jama.Matrix(list.size(), 1);
        Jama.Matrix ExpectedMaxOptions = new Jama.Matrix(list.size(), 1);

        // put everything in a matrix and run routine similar to update coefficients for all of them
        for (int i = 0; i < list.size(); i++) {
            stateSpacePoint point = list.get(i);
            Invest.set(i, 0, point.getPolicy());
            ValueOptimalInvestment.set(i, 0, point.getValueOptimalInvestment());
            ValueDoingNothing.set(i, 0, point.getValueDoingNothing());
            EntryValueCutoff.set(i, 0, point.getEntryValueCutoff());
            Exit.set(i, 0, point.getExitCutoff());
            Profit.set(i, 0, point.getProfit());
            Welfare.set(i, 0, point.getConsumerWelfare());
            Quantity.set(i, 0, point.getMarketQ());
            Price.set(i, 0, point.getPrice());
            ExpectedMaxOptions.set(i, 0, point.getExpectedMaxOptions());
        }

        coefInvestActive = new coefficientValues(s, true);
        coefInvestActive.updateCoefficient(Invest);
        coefInvestEntrant = new coefficientValues(s, false);
        coefInvestEntrant.updateCoefficient(Invest);

        coefValueOptimalInvestment = new coefficientValues(s, true);
        coefValueOptimalInvestment.updateCoefficient(ValueOptimalInvestment);
        coefValueDoingNothing = new coefficientValues(s, true);
        coefValueDoingNothing.updateCoefficient(ValueDoingNothing);
        coefExpectedMax = new coefficientValues(s, true);
        coefExpectedMax.updateCoefficient(ExpectedMaxOptions);

        coefEntryValueThreshold = new coefficientValues(s, false);
        coefEntryValueThreshold.updateCoefficient(EntryValueCutoff);

        coefExit = new coefficientValues(s, true);
        coefExit.updateCoefficient(Exit);
        coefProfit = new coefficientValues(s, true);
        coefProfit.updateCoefficient(Profit);
        coefWelfare = new coefficientValues(s, true);
        coefWelfare.updateCoefficient(Welfare);
        coefQuantity = new coefficientValues(s, true);
        coefQuantity.updateCoefficient(Quantity);
        coefPrice = new coefficientValues(s, true);
        coefPrice.updateCoefficient(Price);
    }
}
