package simulation.mpneSimulation;

/*
 * simulation.java
 *
 * Created on June 9, 2004, 6:51 AM
 */
import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.swing.JTextArea;
import javax.swing.ProgressMonitor;
import simulation.mpneSimulation.utility.simulationPrimitives;
import simulation.mpneSolver.SimulationResults;
import simulation.mpneSolver.stateSpace;
import utility.pmUtility;

/**
 *
 * @author  Stephen P. Ryan
 */
public class simulationInterpolation {

    java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
    static int numFirms = simulationPrimitives.NUM_FIRMS;
    static int numSimulations = simulationPrimitives.NUM_SIMULATIONS;
    private SimulationResults results;

    /** Creates a new instance of simulation */
    public simulationInterpolation(JTextArea jtCommon, stateSpace s) {
        nf.setMinimumFractionDigits(2);
        nf.setMaximumFractionDigits(2);

        boolean socialPlanner = simulationPrimitives.SOCIAL_PLANNER;
        String fn = "";
        String description = "";
        if (simulationPrimitives.LATE_PERIOD) {
            fn = fn.concat("late");
            description = "Late Period";
        }
        if (simulationPrimitives.COUNTERFACTUAL) {
            fn = fn.concat("cf");
            description = description.concat(" Counterfactual");
        }
        if (simulationPrimitives.SOCIAL_PLANNER) {
            fn = fn.concat("sp");
            description = description.concat(" Social Planner");
        }
        // jt.append("Simulating " + description + "\n");

        System.out.println("Check on parameters:");
        System.out.println("\t Scrap Mu: " + s.getSCRAP_MU());
        System.out.println("\t Scrap Sigma: " + s.getSCRAP_SIGMA());
        System.out.println("\t Adjustment Mu: " + s.getADJUSTMENT_MU());
        System.out.println("\t Adjustment Sigma: " + s.getADJUSTMENT_SIGMA());
        System.out.println("\t Entry Mu: " + s.getENTRY_MU());
        System.out.println("\t Entry Sigma: " + s.getENTRY_SIGMA());
        System.out.println("\t Investment MC: "+s.getINVESTMENT_MARGINAL_COST());
        System.out.println("\t Investment MC2: "+s.getINVESTMENT_MARGINAL_COST2());

        simulationPrimitives.SOCIAL_PLANNER = socialPlanner;

        ProgressMonitor mon = new ProgressMonitor(null, "Simulating Industry Paths", "0 / " + numSimulations, 0, numSimulations);

        Jama.Matrix simProfit = new Jama.Matrix(numSimulations, numFirms, 0);
        double averageSizeOverall = 0;
        double marketSizeOverall = 0;
        double quantityProduced = 0;
        double averagePrice = 0;
        double averagePriceEmissions = 0;
        double averageEmissions = 0;
        double simCS = 0;
        double simCarbon = 0;
        double simEmissions = 0;
        double nofirms = 0;
        double onefirm = 0;
        double twofirms = 0;
        double threefirms = 0;
        double fourfirms = 0;
        double fivefirms = 0;
        double sixfirms = 0;

        Random rng = new Random(777);
        ExecutorService tpes = Executors.newFixedThreadPool(simulationPrimitives.NUM_THREADS);

        ArrayList<Future<SimulationTaskResult>> futureList = new ArrayList<Future<SimulationTaskResult>>();
        for (int simulations = 0; simulations < numSimulations; simulations++) {
            futureList.add(tpes.submit(new SimulationTask(s, rng.nextLong())));
        }

        try {
            for (int simulations = 0; simulations < numSimulations; simulations++) {
                SimulationTaskResult r = futureList.get(simulations).get();
                
                for (int i = 0; i < numFirms; i++) {
                    simProfit.set(simulations, i, r.getProfit()[i]);
                }
                averageSizeOverall += r.getAverageSize() / numSimulations;
                marketSizeOverall += r.getMarketSize() / numSimulations;
                simCS += r.getCS() / numSimulations;
                // System.out.println("simulations: "+r.getQuantity());
                quantityProduced += r.getQuantity() / numSimulations;
                nofirms += r.getNoFirms() / numSimulations;
                onefirm += r.getOneFirm() / numSimulations;
                twofirms += r.getTwoFirms() / numSimulations;
                threefirms += r.getThreeFirms() / numSimulations;
                fourfirms += r.getFourFirms() / numSimulations;
                fivefirms += r.getFiveFirms() / numSimulations;
                sixfirms += r.getSixFirms() / numSimulations;
                averagePrice += r.getAveragePrice() / numSimulations;

                mon.setProgress(simulations+1);
                mon.setNote((simulations + 1) + " / " + numSimulations);
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(0);
        }

        Jama.Matrix piVec = new Jama.Matrix(numFirms, 1);
        double piTotal = 0;
        for (int i = 0; i < numFirms; i++) {
            piVec.set(i, 0, pmUtility.mean(simProfit, i));
            piTotal += piVec.get(i, 0);
        }

        Jama.Matrix w;
        w = new Jama.Matrix(numFirms, 2, 0);
        for (int i = 0; i < numFirms; i++) {
            w.set(i, 0, simulationPrimitives.CAPACITY[i]);
            w.set(i, 1, simulationPrimitives.TYPE[i]);
            if (simulationPrimitives.CAPACITY[i] == 0) {
                w.set(i, 1, 0);
            }
        }

        results = new SimulationResults(piTotal / numFirms, piVec, averageSizeOverall, marketSizeOverall, quantityProduced, averagePrice, averagePriceEmissions, averageEmissions, simCS, simCarbon,
                simEmissions, nofirms, onefirm, twofirms, threefirms, fourfirms, fivefirms, sixfirms, w);
        tpes.shutdown();
    }

    /**
     * @return the results
     */
    public SimulationResults getResults() {
        return results;
    }
}
