package simulation.mpneSolver;

/*
 * centralControl.java
 *
 * Created on May 1, 2004, 9:07 AM
 */

import estimation.simulateW.simulateWResult;
import estimation.ssrulesLikelihood.investmentParametersContainer;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import simulation.mpneSimulation.simulationInterpolation;
import simulation.mpneSimulation.utility.simulationPrimitives;
import utility.JTextAreaAutoscroll;
import utility.demandCurve;

/**
 *
 * @author  Stephen P. Ryan
 */
public class centralControl {

    public static int NEW_MARKET = 1;
    public static int MATURE_MARKET = 2;
    stateSpace omega = null;
    JFrame f = null;
    JTextArea jt = null;
    public static boolean USE_GUI = simulationPrimitives.DISPLAY_GUI;

    demandCurve priceThesis;
    double[] estimatedCosts;
    investmentParametersContainer policyParameters;
    simulateWResult dynamicParametersResult;
    double estimatedEntryMu;
    double estimatedEntrySigma;

    /** Creates a new instance of centralControl */
    public centralControl(demandCurve priceThesis, double[] estimatedCosts, investmentParametersContainer policyParameters,
            simulateWResult dynamicParametersResult, double estimatedEntryMu, double estimatedEntrySigma) {

        this.dynamicParametersResult = dynamicParametersResult;
        this.priceThesis = priceThesis;
        this.estimatedCosts = estimatedCosts;
        this.policyParameters = policyParameters;
        this.estimatedEntryMu = estimatedEntryMu;
        this.estimatedEntrySigma = estimatedEntrySigma;

        if (USE_GUI) {
            jt = new JTextAreaAutoscroll();
            f = new JFrame("Simulation");
            f.setBounds(22, 22, 800, 600);
            f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            jt.setEditable(false);
            f.getContentPane().add(new JScrollPane(jt));
        }

    }

    public void solve() {
        if (USE_GUI) {
            f.setVisible(true);
        }

        /**
         * Social planner would only ever have one firm; so make sure that sticks
         * here if we are using the social planner.
         */
        if (!simulationPrimitives.SOCIAL_PLANNER) {
            omega = new stateSpace(simulationPrimitives.NUM_FIRMS, simulationPrimitives.SOCIAL_PLANNER, priceThesis, estimatedCosts, dynamicParametersResult, estimatedEntryMu, estimatedEntrySigma);
        } else {
            omega = new stateSpace(1, simulationPrimitives.SOCIAL_PLANNER, priceThesis, estimatedCosts, dynamicParametersResult, estimatedEntryMu, estimatedEntrySigma);
        }
        iteratedSolver isolve = new iteratedSolver(omega);
        isolve.execute();
        // isolve.dispose();

        if (USE_GUI) {
            f.dispose();
        }
    }

    public SimulationResults getSimulationResults(int startingConfiguration) {
        if (startingConfiguration == NEW_MARKET) {
            simulationPrimitives.CAPACITY[0] = 0;
            simulationPrimitives.CAPACITY[1] = 0;
        }
        if (startingConfiguration == MATURE_MARKET) {
            simulationPrimitives.CAPACITY[0] = 1500;
            simulationPrimitives.CAPACITY[1] = 750;
            System.out.println("------------- Mature Market Calculations -----------------");
        }
        simulationInterpolation counterfactualSimulation = new simulationInterpolation(jt, omega);

        return counterfactualSimulation.getResults();
    }
}
