package simulation.mpneSolver;

/*
 * iteratedSolver.java
 *
 * Created on May 1, 2004, 9:21 AM
 */
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import simulation.mpneSimulation.utility.simulationPrimitives;
import utility.pmUtility;

/**
 *
 * @author  Stephen P. Ryan
 */
public class iteratedSolver {

    boolean vgraph = true;
    static double normConvergenceThreshold = 1E-2;
    static double pNormConvergenceThreshold = 1E-3;
    static double prInvConvergenceThreshold = 1E-6;
    double dampenI = 0;
    double dampenIP = 0;
    double dampenP = 0;
    double dampenInvNoAdj = 0;
    java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
    int turnover = 1;
    int pcount = turnover - 1;
    double dampenLevel = simulationPrimitives.DAMPENER;
    boolean stopnow = false;
    // gui components
    JButton buttonShowInvestmentChange = new JButton("Show Inv");
    JButton graphButton2 = new JButton("Show Convergence");
    JCheckBox solveIndicator = new JCheckBox("Solve Inv", true);
    JCheckBox dampenIndicator = new JCheckBox("Dampen Val", false);
    JCheckBox dampenIndicatorI = new JCheckBox("Dampen Inv", false);
    JCheckBox dampenIndicatorP = new JCheckBox("Dampen PrEntry", true);
    JCheckBox dampenIndicatorIP = new JCheckBox("Dampen PrInv", true);
    JCheckBox dampenIndicatorInvNoAdj = new JCheckBox("Dampen InvNoAdj", true);
    JButton crushButton = new JButton("Crush");
    JProgressBar bar = new JProgressBar();
    JLabel status = new JLabel("Status bar");
    org.jfree.data.xy.XYSeries xy = new org.jfree.data.xy.XYSeries("Value");
    org.jfree.data.xy.XYSeries policy = new org.jfree.data.xy.XYSeries("Policy");
    org.jfree.data.xy.XYSeries conv = new org.jfree.data.xy.XYSeries("Value Norm");
    XYSeries policyConvergence = new XYSeries("Investment Norm");
    XYSeries investmentProbabilityConvergence = new XYSeries("Inv Prob Norm");
    // state space
    stateSpace s;
    // specification modifiers
    String period = "Early";
    String socialPlanner = "";
    String cf = "";
    String ceasar = "";
    ExecutorService tpes = Executors.newFixedThreadPool(simulationPrimitives.NUM_THREADS);
    /**
     * GUI elements
     */
    JFrame tableFrame = null;
    JDialog statusDialog = null;

    /** Creates a new instance of iteratedSolver */
    public iteratedSolver(stateSpace s) {
        this.s = s;
        if (simulationPrimitives.SOCIAL_PLANNER) {
            socialPlanner = "Social Planner";
        }
        if (simulationPrimitives.LATE_PERIOD) {
            period = "Late";
        }
        if (simulationPrimitives.COUNTERFACTUAL) {
            cf = "cf";
        }
        if (simulationPrimitives.COMPETITIVE) {
            ceasar = "comp";
        }
        tableFrame = new JFrame("Table " + s.getStateSpaceSize() + " " + period + " " + socialPlanner + " " + cf + " " + ceasar);
        tableFrame.setBounds(100, 100, 1000, 800);
        JTable table = s.getTable(this);
        JPanel tablePanel = new JPanel(true);
        tablePanel.setLayout(new BorderLayout());
        tablePanel.add(new JScrollPane(table), BorderLayout.CENTER);
        tableFrame.getContentPane().add(new JScrollPane(table), BorderLayout.CENTER);
        JButton showDialog = new JButton("Options");

        JPanel south = new JPanel();
        south.setLayout(new BorderLayout());
        south.add(showDialog, BorderLayout.EAST);
        south.add(status, BorderLayout.WEST);

        tableFrame.getContentPane().add(south, BorderLayout.SOUTH);

        statusDialog = new JDialog(tableFrame, "Options", false);
        showDialog.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (statusDialog.isVisible()) {
                    statusDialog.setVisible(false);
                } else {
                    statusDialog.setVisible(true);
                }
            }
        });

        dampenIP = dampenLevel;
        dampenInvNoAdj = dampenLevel;
        dampenP = dampenLevel;  // probability of entry
        JPanel rightPanel = new JPanel();

        rightPanel.add(dampenIndicator);
        rightPanel.add(dampenIndicatorI);
        rightPanel.add(dampenIndicatorP);
        rightPanel.add(dampenIndicatorIP);
        rightPanel.add(dampenIndicatorInvNoAdj);
        rightPanel.add(solveIndicator);
        rightPanel.add(buttonShowInvestmentChange);
        rightPanel.add(graphButton2);
        rightPanel.add(crushButton);

        crushButton.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                simulationPrimitives.DAMPENER = (1.0 - (1.0 - simulationPrimitives.DAMPENER) / 10.0);
                System.out.println("Dampener: " + simulationPrimitives.DAMPENER);
            }
        });

        final JButton finishButton = new JButton("You are beaten");
        finishButton.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                stopnow = true;
                finishButton.setText("Shutting Down");
                dampenIndicator.setEnabled(false);
                dampenIndicatorI.setEnabled(false);
                dampenIndicatorP.setEnabled(false);
                dampenIndicatorIP.setEnabled(false);
                dampenIndicatorInvNoAdj.setEnabled(false);
                solveIndicator.setEnabled(false);
            }
        });
        rightPanel.add(finishButton);
        final JTextField dampenField = new JTextField(Double.toString(dampenLevel));
        dampenField.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                dampenLevel = Double.parseDouble(dampenField.getText());
            }
        });

        JPanel dampenFieldPanel = new JPanel();
        dampenFieldPanel.add(new JLabel("Dampen: "));
        dampenFieldPanel.add(dampenField);
        rightPanel.add(dampenFieldPanel);

        SpinnerNumberModel model = new SpinnerNumberModel(turnover, 1, 25, 1);
        final JSpinner mySpin = new JSpinner(model);
        rightPanel.add(mySpin);
        if (s.getNumFirms() > 3) {
            rightPanel.add(bar);
        }
        // rightPanel.setLayout(new BoxLayout(rightPanel, BoxLayout.Y_AXIS));
        rightPanel.setLayout(new GridLayout(0, 1));
        statusDialog.getContentPane().add(rightPanel, BorderLayout.CENTER);
        statusDialog.setBounds(1100, 500, 400, 400);
        statusDialog.pack();
        statusDialog.setVisible(simulationPrimitives.DISPLAY_GUI);
        tableFrame.setVisible(simulationPrimitives.DISPLAY_GUI);
        tableFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        mySpin.addChangeListener(new ChangeListener() {

            public void stateChanged(ChangeEvent e) {
                turnover = ((Integer) mySpin.getValue()).intValue();
                pcount = turnover - 1;
            }
        });

        dampenIndicator.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (dampenIndicator.isSelected()) {
                    simulationPrimitives.DAMPENER = dampenLevel;
                } else {
                    simulationPrimitives.DAMPENER = 0;
                }
            }
        });

        dampenIndicatorI.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (dampenIndicatorI.isSelected()) {
                    dampenI = dampenLevel;
                } else {
                    dampenI = 0;
                }
            }
        });

        dampenIndicatorP.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (dampenIndicatorP.isSelected()) {
                    dampenP = dampenLevel;
                } else {
                    dampenP = 0;
                }
            }
        });

        dampenIndicatorIP.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (dampenIndicatorIP.isSelected()) {
                    dampenIP = dampenLevel;
                } else {
                    dampenIP = 0;
                }
            }
        });

        dampenIndicatorInvNoAdj.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (dampenIndicatorInvNoAdj.isSelected()) {
                    dampenInvNoAdj = dampenLevel;
                } else {
                    dampenInvNoAdj = 0;
                }
            }
        });

        org.jfree.data.xy.XYSeriesCollection sc = new org.jfree.data.xy.XYSeriesCollection();

        sc.addSeries(xy);
        sc.addSeries(policy);
        ChartPanel panel = new ChartPanel(ChartFactory.createXYLineChart("Policy Convergence", "State", "Change", sc, PlotOrientation.VERTICAL, false, false, false));
        final JDialog policyChangeGraph = new JDialog(tableFrame, "Policy Function change");
        policyChangeGraph.setBounds(300, 300, 400, 400);
        policyChangeGraph.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        policyChangeGraph.getContentPane().add(panel);

        org.jfree.data.xy.XYSeriesCollection convColl = new org.jfree.data.xy.XYSeriesCollection();

        convColl.addSeries(conv);
        convColl.addSeries(policyConvergence);
        convColl.addSeries(investmentProbabilityConvergence);
        ChartPanel panel2 = new ChartPanel(ChartFactory.createXYLineChart("Convergence", "Iteration", "Log Change", convColl, PlotOrientation.VERTICAL, true, true, true));
        final JDialog graph2 = new JDialog(tableFrame, "Convergence");
        graph2.setBounds(700, 300, 400, 400);
        graph2.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        graph2.getContentPane().add(panel2);

        // graph.setVisible(true);

        buttonShowInvestmentChange.addActionListener(new java.awt.event.ActionListener() {

            public void actionPerformed(java.awt.event.ActionEvent e) {
                if (policyChangeGraph.isVisible()) {
                    policyChangeGraph.setVisible(false);
                    buttonShowInvestmentChange.setText("Show Policy");
                } else {
                    policyChangeGraph.setVisible(true);
                    buttonShowInvestmentChange.setText("Hide Policy");
                }
            }
        });

        graphButton2.addActionListener(new java.awt.event.ActionListener() {

            public void actionPerformed(java.awt.event.ActionEvent e) {
                if (graph2.isVisible()) {
                    graph2.setVisible(false);
                    graphButton2.setText("Show Convergence");
                } else {
                    graph2.setVisible(true);
                    graphButton2.setText("Hide Convergence");
                }
            }
        });
    }

    void execute() {
        boolean converged = false;
        boolean first = true;
        Jama.Matrix v1 = new Jama.Matrix(s.getStateSpaceSize(), 1);
        Jama.Matrix v2 = new Jama.Matrix(s.getStateSpaceSize(), 1);
        Jama.Matrix p1 = new Jama.Matrix(s.getStateSpaceSize(), 1);
        Jama.Matrix p2 = new Jama.Matrix(s.getStateSpaceSize(), 1);
        Jama.Matrix probInv1 = new Jama.Matrix(s.getStateSpaceSize(), 1);
        Jama.Matrix probInv2 = new Jama.Matrix(s.getStateSpaceSize(), 1);

        // initialize coefficients and pre-specified weights of spline
        // need to expand it for the heterogeneous case to be 100% kosher
        coefficientValues coefficients = new coefficientValues(s, true);

        for (int i = 0; i < v1.getRowDimension(); i++) {
            xy.add(i, v2.get(i, 0) - v1.get(i, 0));
            policy.add(i, p2.get(i, 0) - p1.get(i, 0));
        }

        double lastNorm = 0;
        long t1, t2;

        t1 = System.currentTimeMillis();

        double previousNorm = 0;
        double previousPNorm = 0;
        double previousPrInv = 0;
        int count = 0;


        for (int t = 0; t < s.getNumStates(); t++) {
            if (pmUtility.sortMatrixAscending(s.getExpandedList().get(t).getMatrix()).minus(s.getExpandedList().get(t).getMatrix()).norm1() == 0) {
                // Print current expanded state
                System.out.print("Now solving ");
                pmUtility.prettyPrintVector(s.getExpandedList().get(t).getMatrix());

                // create lists of joint expanded states
                int[] temp = new int[s.getStateSpaceSize()];
                count = 0;
                for (int i = 0; i < s.getStateSpaceSize(); i++) {
                    if (pmUtility.sortMatrixAscending(s.getPoint(i).getMatrix().getMatrix(0, simulationPrimitives.NUM_FIRMS - 1, 1, 1)).minus(s.getExpandedList().get(t).getMatrix()).norm1() == 0) {
                        temp[count] = i;
                        count += 1;
                    }
                }
                int[] list = new int[count];
                for (int i = 0; i < count; i++) {
                    list[i] = temp[i];
                }

                int iterations = 0;
                converged = false;
                simulationPrimitives.DAMPENER = 0.99;

                simulationPrimitives.ABSORB = false;
                while (!converged) {
                    if (!converged) {
                        iterations++;
                        long time1 = System.currentTimeMillis();
                        bar.setMaximum(s.getStateSpaceSize() - 1);

                        // eventually increase dampener after going to a closer point
                        if (iterations == 200) {
                            simulationPrimitives.DAMPENER = 0.99;
                        }
                        if (iterations == 250) {
                            simulationPrimitives.ABSORB = true;
                        }
                        if(iterations==1000) {
                            simulationPrimitives.DAMPENER = 0.999;
                        }
                        if (iterations == 4000) {
                            simulationPrimitives.ABSORB = false;
                        }
                        // solve for the new iteration update
                        solveBellman(true, coefficients, list);


                        // update value and policy functions
                        for (int i : list) { //
                            stateSpacePoint point = s.getPoint(i);
                            point.updateState();
                            v2.set(i, 0, v1.get(i, 0));
                            v1.set(i, 0, point.getValue());
                            p2.set(i, 0, p1.get(i, 0));
                            p1.set(i, 0, point.getPolicy());
                            probInv2.set(i, 0, probInv1.get(i, 0));
                            probInv1.set(i, 0, point.getInvestProb());
                            s.updateTable(i);
                        }
                        // pmUtility.prettyPrintVector(v1);

                        // update coefficient with dampened values
                        coefficients.updateCoefficient(v1);


                        // check convergence and others
                        long time2 = System.currentTimeMillis();
                        Jama.Matrix change = v1.minus(v2);
                        Jama.Matrix pChange = p1.minus(p2);
                        double norm = change.norm2() / (v2.norm2()*((1.0 - simulationPrimitives.DAMPENER) * count));
                        double pNorm = pChange.norm2() / (p2.norm2()*((1.0 - simulationPrimitives.DAMPENER) * count));
                        double prInvNorm = (probInv2.minus(probInv1)).norm2() / (probInv2.norm2()*((1.0 - simulationPrimitives.DAMPENER) * count));

                        double maxChangeP = pChange.get(0,0);
                        for(int p=0;p<pChange.getRowDimension();p++) {
                            if(Math.abs(pChange.get(p, 0))>maxChangeP) {
                                maxChangeP = Math.abs(pChange.get(p,0));
                            }
                        }


                        if (first) {
                            first = false;
                        } else if (simulationPrimitives.DISPLAY_GUI) {
                            status.setText(norm + " (" + (100 * (norm - previousNorm - lastNorm) / lastNorm) + ") time: " + ((time2 - time1) / 1000.0));
                            conv.add(iterations, Math.log(1 + norm - previousNorm));
                            policyConvergence.add(iterations, Math.log(1 + pNorm - previousPNorm));
                            investmentProbabilityConvergence.add(iterations, Math.log(1 + prInvNorm - previousPrInv));
                            //System.out.println(norm+" "+pNorm+" "+prInvNorm);
                        }
                        if (norm - previousNorm < normConvergenceThreshold && pNorm - previousPNorm < pNormConvergenceThreshold && prInvNorm - previousPrInv < prInvConvergenceThreshold) {
                            converged = true;
                            previousNorm = norm;
                            previousPNorm = pNorm;
                            previousPrInv = prInvNorm;
                        }
//                        if(maxChangeP < 10) {
//                            stopnow = true;
//                        }
//                        if(norm<1E-5) {
//                            stopnow = true;
//                        }
                        if (stopnow) {
                            converged = true;
                            previousNorm = norm;
                            previousPNorm = pNorm;
                            previousPrInv = prInvNorm;
                        }
                        if (iterations == 2000) {
                            converged = true;
                            System.out.println("False convergence--stopping due to the iteration limit with norms " + (norm - previousNorm) + " " + (pNorm - previousPNorm) + " " + (prInvNorm - previousPrInv));
                            previousNorm = norm;
                            previousPNorm = pNorm;
                            previousPrInv = prInvNorm;
                        }
                        lastNorm = norm;
                    }
                }
                t2 = System.currentTimeMillis();
                System.out.println("Time to convergence " + (t2 - t1) / 1000 + " Number of iterations: " + iterations);

            }
        }

        tpes.shutdown();
    }

    private void solveBellman(boolean solvePolicy, coefficientValues c, int[] list) {
        // solve Bellman if interpolation
        boolean useMultipleCPU = true;
        if (useMultipleCPU) {
            ArrayList<Future<Integer>> futureList = new ArrayList<Future<Integer>>();
            for (int i : list) {
                futureList.add(tpes.submit(new SolveBellmanTask(s.getPoint(i), solvePolicy, c)));
            }
            try {
                for (int i = 0; i < futureList.size(); i++) {
                    futureList.get(i).get();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            for (int i = 0; i < s.getStateSpaceSize(); i++) {
                stateSpacePoint point = s.getPoint(i);
                point.solveValue(solvePolicy, c);
            }
        }
    }

    void dispose() {
        statusDialog.dispose();
        tableFrame.dispose();
    }
}
