package simulation.mpneSolver;

/*
 * quantitySolver.java
 *
 * Created on August 26, 2004, 3:31 PM
 * Major revision June 2005
 */
import utility.*;
import utility.pmUtility.*;
import java.util.*;
import mcmc.gibbsLTEGeneralized;
import optimization.Uncmin_f77;

/**
 *
 * @author  Stephen P. Ryan
 */
public class quantitySolver implements optimization.Uncmin_methods, mcmc.mcmcFunction {

    private Jama.Matrix state;
    private Jama.Matrix qVec;
    private int numParams;
    private int numFirms;
    private double marketQ;
    private double totalEmissions;
    private double carbonRevenue;
    private Jama.Matrix profits;
    private double price;
    private double priceEmissions;
    private boolean priceTaker = false;
    boolean verbose = false;
    demandCurve priceThesis;
    private double CAPACITY_COST_THRESHOLD;
    private double CAPACITY_COST;
    private double MARGINAL_COST;
    private double A;
    private double B;

    /** Creates a new instance of marginalCostSolver */
    public quantitySolver(Jama.Matrix statePassed, boolean priceTaker, demandCurve priceThesis, double[] estimatedCosts) {
        this.priceTaker = priceTaker;
        this.priceThesis = priceThesis;

        A = -priceThesis.getA() / priceThesis.getB(0);
        B = -1.0 / priceThesis.getB(0);
//        System.out.println("A: "+A);
//        System.out.println("B: "+B);

        CAPACITY_COST = estimatedCosts[1];
        CAPACITY_COST_THRESHOLD = Math.exp(estimatedCosts[2]) / (1.0 + Math.exp(estimatedCosts[2]));
        MARGINAL_COST = estimatedCosts[3];

        int numActive = 0;
        ArrayList<Integer> activePos = new ArrayList<Integer>();
        for (int i = 0; i < statePassed.getRowDimension(); i++) {
            if (statePassed.get(i, 0) > 0) {
                numActive++;
                activePos.add(i);
            }
        }
        state = new Jama.Matrix(numActive, 2);
        for (int i = 0; i < numActive; i++) {
            int indexNext = activePos.get(i);
            state.set(i, 0, statePassed.get(indexNext, 0));
        }

        numParams = state.getRowDimension();
        numFirms = state.getRowDimension();

        double[] guess = new double[numParams + 1];
        for (int i = 0; i < numFirms; i++) {
            guess[i + 1] = CAPACITY_COST_THRESHOLD * state.get(i, 0);
        }

        if (numActive > 0) {
            gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 1000, 0, guess);
            guess = lte.getLowestPoint();

            Uncmin_f77 minimizer = new Uncmin_f77(false);

            double[] xpls = new double[numParams + 1];
            double[] fpls = new double[2];
            double[] gpls = new double[numParams + 1];
            int[] itrmcd = new int[2];
            double[][] a = new double[numParams + 1][numParams + 1];
            double[] udiag = new double[numParams + 1];
            double[] typsiz = new double[numParams + 1];
            for (int i = 1; i <= numParams; i++) {
                typsiz[i] = 1;
            }
            double[] fscale = {0, 1E-8};
            int[] method = {0, 3};
            int[] iexp = {0, 0};
            int[] msg = {0, 1};
            int[] ndigit = {0, 15};
            int[] itnlim = {0, 150};
            int[] iagflg = {0, 0};
            int[] iahflg = {0, 0};
            double[] dlt = {0, 1};
            double[] gradtl = {0, 1E-6};
            double[] stepmx = {0, 1E8};
            double[] steptl = {0, 1E-8};

            // pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
            minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
            // pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
        }

        price = calcPrice(marketQ);
        if (Double.isInfinite(price)) {
            price = 0;
        }

        profits = new Jama.Matrix(statePassed.getRowDimension(), 1, 0);
        qVec = new Jama.Matrix(statePassed.getRowDimension(), 1, 0);

        for (int i = 0; i < numActive; i++) {
            // calculate profits for each firm here
            int indexNext = activePos.get(i);
            double pi = guess[i + 1] * (price - MARGINAL_COST);
            double utilPct = guess[i + 1] / state.get(i, 0);
            if (utilPct > CAPACITY_COST_THRESHOLD) {
                // capacityCost * (x[i] - nu * capacity[i]);
                pi -= 0.5 * CAPACITY_COST * Math.pow(guess[i + 1] - (CAPACITY_COST_THRESHOLD * state.get(i, 0)), 2);
            }
            qVec.set(indexNext, 0, guess[i + 1]);
            profits.set(indexNext, 0, pi);
        }
    }

    public double getConsumerSurplus(double quantity) {
        if (quantity == 0) {
            return 0;
        } else {
            return (Math.exp(A) * Math.pow(quantity, -B + 1.0) / (-B + 1.0)) - (quantity * calcPrice(quantity));
        }
    }

    public double calcPrice(double quantity) {
        return priceThesis.getPrice(quantity, 0);
    }

    public double calcPricePrime(double quantity) {
        return calcPrice(quantity) * (-B) / quantity;
    }

    public double f_to_minimize(double[] x) {
        marketQ = 0;
        //System.out.println(x.length);
        for (int i = 1; i <= numFirms; i++) {
            x[i] = Math.max(1E-16, x[i]);  // need to bound quantities to be positive and away from the infinite prices at q=0
            x[i] = Math.min(x[i], state.get(i - 1, 0));
            marketQ += x[i];
        }

        price = calcPrice(marketQ);
        double pricePrime = calcPricePrime(marketQ);

        Jama.Matrix fvec = new Jama.Matrix(x.length - 1, 1);
        for (int i = 1; i <= numFirms; i++) {
            double marginalBenefit = price;
            if (!priceTaker) {
                marginalBenefit += x[i] * pricePrime;
            }
            double marginalCost = MARGINAL_COST;
            double utilPct = x[i] / state.get(i - 1, 0);
            if (utilPct > CAPACITY_COST_THRESHOLD) {
                marginalCost += CAPACITY_COST * (x[i] - (CAPACITY_COST_THRESHOLD * state.get(i - 1, 0)));
            }
            fvec.set(i - 1, 0, marginalBenefit - marginalCost);
        }
        double value = ((fvec.transpose()).times(fvec)).get(0, 0);
        return value;
    }

    public void gradient(double[] x, double[] g) {
    }

    public void hessian(double[] x, double[][] h) {
    }

    public Jama.Matrix getProfits() {
        return profits;
    }

    public void setProfits(Jama.Matrix profits) {
        this.profits = profits;
    }

    public double getPrice() {
        return price;
    }

    public void setPrice(double price) {
        this.price = price;
    }

    public double getPriceEmissions() {
        return priceEmissions;
    }

    double getEmissions() {
        return totalEmissions;
    }

    public double getCarbonRevenue() {
        return carbonRevenue;
    }

    public double getMarketQ() {
        return marketQ;
    }

    public void setMarketQ(double marketQ) {
        this.marketQ = marketQ;
    }

    public Jama.Matrix getQVec() {
        return qVec;
    }

    public void setQVec(Jama.Matrix qVec) {
        this.qVec = qVec;
    }

    public double objectiveFunction(double[] x) {
        return -f_to_minimize(x);
    }

    public double pi(double[] x) {
        return 1;
    }
}
