package simulation.mpneSolver;

/*
 * stateSpace.java
 *
 * Created on May 1, 2004, 9:19 AM
 */
import estimation.simulateW.simulateWResult;
import java.util.ArrayList;
import java.util.Collections;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.swing.JTable;
import simulation.mpneSimulation.utility.simulationPrimitives;
import utility.demandCurve;

/**
 *
 * @author  Stephen P. Ryan
 */
public class stateSpace implements java.io.Serializable {

    private ArrayList<stateSpacePoint> list;
    private ArrayList<expandedSpacePoint> expandedList;
    private int numFirms;
    private int numType1;
    private int numType2;
    private stateSpaceTableModel model;
    private int numStates;
    private ArrayList<Jama.Matrix> actionSpace;
    private demandCurve priceThesis;
    private double[] productionCosts;
    private double ADJUSTMENT_MU;
    private double ADJUSTMENT_SIGMA;
    private double SCRAP_MU;
    private double SCRAP_SIGMA;
    private double ENTRY_MU;
    private double ENTRY_SIGMA;
    private double INVESTMENT_MARGINAL_COST;
    private double INVESTMENT_MARGINAL_COST2;

    /** Creates a new instance of stateSpace */
    public stateSpace(int numberFirms, boolean socialPlanner, demandCurve priceThesis, double[] productionCosts,
            simulateWResult dynamicParametersResult, double estimatedEntryMu, double estimatedEntrySigma) {
        this.priceThesis = priceThesis;
        this.productionCosts = productionCosts;

        ADJUSTMENT_MU = dynamicParametersResult.getiDist()[1];
        ADJUSTMENT_SIGMA = dynamicParametersResult.getiDist()[2];
        SCRAP_MU = dynamicParametersResult.getxDist()[1];
        SCRAP_SIGMA = dynamicParametersResult.getxDist()[2];
        ENTRY_MU = estimatedEntryMu;
        ENTRY_SIGMA = estimatedEntrySigma;

        INVESTMENT_MARGINAL_COST = dynamicParametersResult.getGuess()[7];
        INVESTMENT_MARGINAL_COST2 = dynamicParametersResult.getGuess()[8];

        numFirms = numberFirms;
        numType1 = simulationPrimitives.NUM_TYPE1;
        numType2 = simulationPrimitives.NUM_TYPE2;
        list = new ArrayList<stateSpacePoint>();
        expandedList = new ArrayList<expandedSpacePoint>();
        actionSpace  = actionSpace(numFirms-1);
        numStates = 0;
        // create combinations of types of firms (expanded state) - use symmetry when possible
        Jama.Matrix seed = new Jama.Matrix(numFirms, 1);
        this.addExpandedSpace();
        // populate each combination of types with a state space that has extra column for types
        // should parallelize this task
        ExecutorService tpes = Executors.newFixedThreadPool(1);
        ArrayList<Future<Integer>> futureList = new ArrayList<Future<Integer>>();
        for (int i = 0; i < expandedList.size(); i++) {
            Jama.Matrix expanded = expandedList.get(i).getMatrix();
            seed = new Jama.Matrix(numFirms, 2);
            seed.setMatrix(0, numFirms-1, 1,1, expanded);
            futureList.add(tpes.submit(new SetUpSpaceTask(this,seed,numFirms)));
            numStates += 1;
        }
        try {
            for (int i = 0; i < futureList.size(); i++) {
                futureList.get(i).get();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        Collections.sort(list);
        Collections.sort(expandedList);


        tpes.shutdown();
    }

    /* implements state space with expanded state space and reduced states when possible */
    public void addSeedingExpandedState(Jama.Matrix last, int level, int lastCap, ArrayList<stateSpacePoint> stateSeedingList, int maximumLevel) {
        if (level == maximumLevel) {
            if (lastCap == 0 || last.get(level-1,1) != last.get(level-2,1)) {
                Jama.Matrix vector = last.copy();
                vector.set(level - 1, 0, 0);
                if (vector.get(level-1, 1)==0) {
                    stateSeedingList.add(new stateSpacePoint(vector, this));
                }
                for (int cap = (int) simulationPrimitives.MIN_CAPACITY; cap <= simulationPrimitives.MAX_CAPACITY; cap += simulationPrimitives.CAPACITY_INC) {
                    vector = last.copy();
                    vector.set(level - 1, 0, cap);
                    stateSeedingList.add(new stateSpacePoint(vector, this));
                }
            } else {
                for (int cap = lastCap; cap <= simulationPrimitives.MAX_CAPACITY; cap += simulationPrimitives.CAPACITY_INC) {
                    Jama.Matrix vector = last.copy();
                    vector.set(level - 1, 0, cap);
                    stateSeedingList.add(new stateSpacePoint(vector, this));
                }
            }
        } else if  (level == 1) {
            Jama.Matrix vector = last.copy();
            vector.set(level - 1, 0, lastCap);
            if (vector.get(level-1, 1)==0) {
                addSeedingExpandedState(vector, level + 1, lastCap, stateSeedingList, maximumLevel);
            }
            for (int cap = (int) simulationPrimitives.MIN_CAPACITY; cap <= simulationPrimitives.MAX_CAPACITY; cap += simulationPrimitives.CAPACITY_INC) {
                vector = last.copy();
                vector.set(level - 1, 0, cap);
                addSeedingExpandedState(vector, level + 1, 0, stateSeedingList, maximumLevel);
            }
        } else {
            if (lastCap == 0 || last.get(level-1,1) != last.get(level-2,1)) {
                Jama.Matrix vector = last.copy();
                vector.set(level - 1, 0, lastCap);
                if (vector.get(level-1, 1)==0) {
                    addSeedingExpandedState(vector, level + 1, lastCap, stateSeedingList, maximumLevel);
                }
                for (int cap = (int) simulationPrimitives.MIN_CAPACITY; cap <= simulationPrimitives.MAX_CAPACITY; cap += simulationPrimitives.CAPACITY_INC) {
                    vector = last.copy();
                    vector.set(level - 1, 0, cap);
                    addSeedingExpandedState(vector, level + 1, cap, stateSeedingList, maximumLevel);
                }
            } else {
                for (int cap = lastCap; cap <= simulationPrimitives.MAX_CAPACITY; cap += simulationPrimitives.CAPACITY_INC) {
                    Jama.Matrix vector = last.copy();
                    vector.set(level - 1, 0, cap);
                    addSeedingExpandedState(vector, level + 1, cap, stateSeedingList, maximumLevel);
                }
            }
        }
    }

    // Creates expanded state space taking into account number of firms of each type
    public void addExpandedSpace() {
           // if I am an incumbent
            for (int n2 = 0; n2 <= numType2; n2++) {
            for (int n1 = 0; n1 <= numType1; n1++) {
                if (n1+n2 < numFirms) {
                  Jama.Matrix vector = new Jama.Matrix(numFirms,1,0);
                  if (n1 > 0) vector.setMatrix(numFirms-n2-n1,numFirms-n2-1,0,0,new Jama.Matrix(n1,1,1));
                  if (n2 > 0) vector.setMatrix(numFirms-n2,numFirms-1,0,0,new Jama.Matrix(n2,1,2));
                  //vector.print(0, 0);
                  expandedList.add(new expandedSpacePoint(vector));
                }
              }
            }
           // if I am type 1
            for (int n2 = 0; n2 <= numType2; n2++) {
            for (int n1 = 0; n1 <= numType1-1; n1++) {
                if (n1+n2 <= numFirms) {
                  Jama.Matrix vector = new Jama.Matrix(numFirms,1,0);
                  vector.set(0, 0, 1);
                  if (n1 > 0) vector.setMatrix(numFirms-n2-n1,numFirms-n2-1,0,0,new Jama.Matrix(n1,1,1));
                  if (n2 > 0) vector.setMatrix(numFirms-n2,numFirms-1,0,0,new Jama.Matrix(n2,1,2));
                  //vector.print(0, 0);
                  expandedList.add(new expandedSpacePoint(vector));
                }
              }
            }
           // if I am type 2
            for (int n2 = 0; n2 <= numType2-1; n2++) {
            for (int n1 = 0; n1 <= numType1; n1++) {
                if (n1+n2 <= numFirms) {
                  Jama.Matrix vector = new Jama.Matrix(numFirms,1,0);
                  vector.set(0, 0, 2);
                  if (n1 > 0) vector.setMatrix(numFirms-n2-n1,numFirms-n2-1,0,0,new Jama.Matrix(n1,1,1));
                  if (n2 > 0) vector.setMatrix(numFirms-n2,numFirms-1,0,0,new Jama.Matrix(n2,1,2));
                  //vector.print(0, 0);
                  expandedList.add(new expandedSpacePoint(vector));
                }
            }
            }

    }

    // This generates a list of others possible actions (not enter or exit, stay, and stay and invest)
    public static ArrayList<Jama.Matrix> actionSpace(int numberFirms) {
        Jama.Matrix seed = new Jama.Matrix(numberFirms, 1);
        ArrayList<Jama.Matrix> list = new ArrayList<Jama.Matrix>();
        actionSpace(seed, 1, list, numberFirms);
        return list;
    }
    
    private static void actionSpace(Jama.Matrix last, int level, ArrayList<Jama.Matrix> list, int numberFirms) {
        int lastCap = 0;
        if (level == numberFirms) {
            for (int cap = lastCap; cap <= 2; cap += 1) {
                Jama.Matrix vector = last.copy();
                vector.set(level - 1, 0, cap);
                //vector.print(0,0);
                list.add(vector);
            }
        } else {
            for (int cap = lastCap; cap <= 2; cap += 1) {
                Jama.Matrix vector = last.copy();
                vector.set(level - 1, 0, cap);
                actionSpace(vector, level + 1, list, numberFirms);
            }
        }
    }

    public void setupSpace(ArrayList<stateSpacePoint> plist, int numberFirms) {
        list = plist;
        numFirms = numberFirms;
        model = new stateSpaceTableModel(this);
    }

    public ArrayList<stateSpacePoint> getList() {
        return list;
    }

    public ArrayList<expandedSpacePoint> getExpandedList() {
        return expandedList;
    }

    public stateSpaceTableModel getTableModel() {
        return model;
    }

    public int getNumFirms() {
        return numFirms;
    }

    public int getNumStates() {
        return numStates;
    }

    public ArrayList<Jama.Matrix> getActionSpace() {
        return actionSpace;
    }

    public JTable getTable(iteratedSolver solver) {
        model = new stateSpaceTableModel(this);
        JTable table = new JTable(model);
        return table;
    }

    public void updateTable(int row) {
        model.fireTableRowsUpdated(row, row);
    }

    public int getStateSpaceSize() {
        return list.size();
    }

    public void printResults() {
        for (int i = 0; i < list.size(); i++) {
            stateSpacePoint point = list.get(i);
            System.out.println(point + "," + point.getValue() + "," + point.getPolicy() + "," + point.getEntryProbability() + "," + point.getExitProbability());
        }
    }

    public stateSpacePoint getPoint(Jama.Matrix s) {
        // have to sort s
        s = sortOnFirmId(s);
        stateSpacePoint point = null;
        try {
            point = list.get(Collections.binarySearch(list, new stateSpacePoint(s)));
        } catch (Exception e) {
            e.printStackTrace();
            System.out.print("********** ");
            System.out.println("State not found!");
            s.print(0,0);
            System.out.println("********** ");
        }
        return point;
    }

    public stateSpacePoint getPoint(Jama.Matrix s, int exchangeIndex) {
        Jama.Matrix ns = s.copy();
        ns.set(0, 0, s.get(exchangeIndex, 0));
        ns.set(exchangeIndex, 0, s.get(0, 0));
        ns.set(0, 1, s.get(exchangeIndex, 1));
        ns.set(exchangeIndex, 1, s.get(0, 1));
        //ns.print(0,0);
        return getPoint(ns);
    }

    public stateSpacePoint getPoint(int i) {
        return list.get(i);
    }

    public int getPointIndex(stateSpacePoint state) {
        return list.indexOf(state);
    }

    public int getExpandedIndex(Jama.Matrix expanded) {
        expandedSpacePoint expandedState = expandedList.get(Collections.binarySearch(expandedList, new expandedSpacePoint(expanded)));
        return expandedList.indexOf(expandedState);
    }

    void updateTable() {
        model.fireTableDataChanged();
    }

    private static Jama.Matrix sortOn(Jama.Matrix x) {
        // sort firms depending on their id (used in shortcut)
        int size = x.getRowDimension();  // number of firms
        Jama.Matrix temp = new Jama.Matrix(size, 1);
        temp.set(0, 0, x.get(0, 0)); // firm state
        ArrayList<Double> list = new ArrayList<Double>(size-1);
        for (int i = 1; i < size; i++) {
           list.add(new Double(x.get(i, 0)));
        }
        Collections.sort(list);
        for (int i = 1; i < size; i++) {
           double value = (list.get(i-1)).doubleValue();
           temp.set(i, 0, value);
        }
        return temp;
    }

    private static Jama.Matrix sortOnFirmId(Jama.Matrix x) {
        // sort firms depending on their id (used in shortcut)
        int size = x.getRowDimension();  // number of firms
        Jama.Matrix temp = new Jama.Matrix(size, 2);
        temp.set(0, 0, x.get(0, 0)); // firm state
        temp.set(0, 1, x.get(0, 1)); // firm id
        int count = 0; int lastCount = 0;
        // select firms with different ids and sort them
        while (count <= size) {
            int intCount = 0;
            ArrayList<Double> list = new ArrayList<Double>(count);
            for (int i = 1; i < size; i++) {
                if (x.get(i,1) == count) { // different type of firms
                    list.add(new Double(x.get(i, 0)));
                    intCount += 1;
            }
            }
            Collections.sort(list);
            for (int i = 0; i < intCount; i++) {
                double value = (list.get(i)).doubleValue();
                temp.set(lastCount+1+i, 0, value);
                temp.set(lastCount+1+i, 1, count);
            }
            lastCount += intCount;
            count += 1;
        }
        return temp;
    }

    /**
     * @return the priceThesis
     */
    public demandCurve getPriceThesis() {
        return priceThesis;
    }

    /**
     * @return the productionCosts
     */
    public double[] getProductionCosts() {
        return productionCosts;
    }

    /**
     * @return the ADJUSTMENT_MU
     */
    public double getADJUSTMENT_MU() {
        return ADJUSTMENT_MU;
    }

    /**
     * @return the ADJUSTMENT_SIGMA
     */
    public double getADJUSTMENT_SIGMA() {
        return ADJUSTMENT_SIGMA;
    }

    /**
     * @return the SCRAP_MU
     */
    public double getSCRAP_MU() {
        return SCRAP_MU;
    }

    /**
     * @return the SCRAP_SIGMA
     */
    public double getSCRAP_SIGMA() {
        return SCRAP_SIGMA;
    }

    /**
     * @return the ENTRY_MU
     */
    public double getENTRY_MU() {
        return ENTRY_MU;
    }

    /**
     * @return the ENTRY_SIGMA
     */
    public double getENTRY_SIGMA() {
        return ENTRY_SIGMA;
    }

    /**
     * @return the INVESTMENT_MARGINAL_COST
     */
    public double getINVESTMENT_MARGINAL_COST() {
        return INVESTMENT_MARGINAL_COST;
    }

    /**
     * @return the INVESTMENT_MARGINAL_COST2
     */
    public double getINVESTMENT_MARGINAL_COST2() {
        return INVESTMENT_MARGINAL_COST2;
    }


}
