package simulation.mpneSolver;

/*
 * stateSpacePoint.java
 *
 * Created on May 1, 2004, 9:12 AM
 */
import JSci.maths.statistics.NormalDistribution;
import Jama.Matrix;
import java.io.Serializable;
import java.util.ArrayList;
import simulation.mpneSimulation.utility.simulationPrimitives;
import utility.pmUtility;

/**
 *
 * @author  Stephen P. Ryan and Mar Reguant
 *
 * Note: all common functions at the beginning
 * Major methods to compute Bellman and solve for investment at the END in two blocks
 * 1. Methods for discrete model
 * 2. Methods for continuous model
 */
public class stateSpacePoint implements Comparable<stateSpacePoint>, Serializable {

    private Jama.Matrix state;
    private Jama.Matrix firmId;
    private stateSpace space;
    private double stateValue;
    private double stateInvestment;
    private Jama.Matrix profits;
    private int numFirms;
    private double exitProb = 0;
    private double entryProb = 0;
    private double price;
    private double priceEmissions;
    private double emissions;
    private Boolean shortcut = null;
    private Integer shortcutIndex = null;
    private double consumerWelfare = 0;
    private double carbonRevenue = 0;
    private double marketQ = 0;
    private double investProb = 0;
    private static NormalDistribution standardNormal = new NormalDistribution(0, 1);
    private double rV = 0;
    private double rProbabilityExit = 0;
    private double rProbabilityEntry = 0;
    private double rProbabilityInvestment = 0;
    private double rInvestment = 0;
    private double changeV = 0;
    static boolean DEBUG = false;
    static double DEBUG_STATE_1 = 0;
    static double DEBUG_STATE_2 = 3000;
    private boolean ABSORBED = false;
    private Jama.Matrix modifiedCoefficient;
    static double[] nodes = {3.436159118, 2.532731674, 1.756683649, 1.036610829, 0.3429013272};
    static double[] weights = {0.7640432855 * Math.pow(10, -5), 0.1343645746 * Math.pow(10, -2), 0.03387439445, 0.2401386110, 0.6108626337};
    private double valueDoingNothing;
    private double valueOptimalInvestment;
    private double entryValueCutoff;
    private double expectedMaxOptions;

    /** Creates a new instance of stateSpacePoint */
    public stateSpacePoint(Jama.Matrix seed) {
        state = seed.copy();
        firmId = seed.getMatrix(0, numFirms - 1, 1, 1);
    }

    public void updateState() {
        updateValueOnly();
        updatePolicyOnly();
    }

    public void updateValueOnly() {
        changeV = stateValue - rV;
        // rV = (1.0 - primitives.DAMPENER) * stateValue + primitives.DAMPENER * rV;
        rV = (1.0 - simulationPrimitives.DAMPENER_VALUE) * stateValue + simulationPrimitives.DAMPENER_VALUE * rV;
    }

    public void updatePolicyOnly() {
        rProbabilityExit = (1.0 - simulationPrimitives.DAMPENER) * exitProb + simulationPrimitives.DAMPENER * rProbabilityExit;
        rProbabilityEntry = (1.0 - simulationPrimitives.DAMPENER) * entryProb + simulationPrimitives.DAMPENER * rProbabilityEntry;
        rProbabilityInvestment = (1.0 - simulationPrimitives.DAMPENER) * investProb + simulationPrimitives.DAMPENER * rProbabilityInvestment;
        rInvestment = (1.0 - simulationPrimitives.DAMPENER) * stateInvestment + simulationPrimitives.DAMPENER * rInvestment; //primitives.DAMPENER
        if (stateInvestment < 10) { // reset to zero if very low investment
            rInvestment = 0;
        }
        if (stateInvestment < simulationPrimitives.MIN_CAPACITY & state.get(0, 0) == 0) { // set investment to minimum capacity
            rInvestment = simulationPrimitives.MIN_CAPACITY;
        }
    }

    public stateSpacePoint(Jama.Matrix seed, stateSpace spaceRef) {
        space = spaceRef;
        numFirms = seed.getRowDimension();
        state = seed.copy();
        firmId = seed.getMatrix(0, numFirms - 1, 1, 1);
        if (state.get(0, 0) > 0) {
            entryProb = 1;
        } else {
            entryProb = 0;
        }
        // initialize quantities and profits here
        if (simulationPrimitives.SOCIAL_PLANNER || simulationPrimitives.COMPETITIVE) {
            //state.print(0,0);
            quantitySolver solver = new quantitySolver(state, true, space.getPriceThesis(), space.getProductionCosts());
            profits = solver.getProfits();
            price = solver.getPrice();
            priceEmissions = solver.getPriceEmissions();
            emissions = solver.getEmissions();
            marketQ = solver.getMarketQ();
            stateInvestment = 0;
            consumerWelfare = solver.getConsumerSurplus(marketQ);
            if (simulationPrimitives.SOCIAL_PLANNER) {
                stateValue = (consumerWelfare + profits.get(0, 0)) * 10;
            } else {
                stateValue = 0; // profits.get(0,0)*10;
            }
            rV = stateValue;
        } else {
            quantitySolver solver = new quantitySolver(state, false, space.getPriceThesis(), space.getProductionCosts());
            profits = solver.getProfits();
            price = solver.getPrice();
            priceEmissions = solver.getPriceEmissions();
            emissions = solver.getEmissions();
            stateValue = profits.get(0, 0) * 10;
            rV = stateValue;
            stateInvestment = 0;
            marketQ = solver.getMarketQ();
            consumerWelfare = solver.getConsumerSurplus(marketQ);
            carbonRevenue = solver.getCarbonRevenue();
        }
        updateState();
    }

    public double getPrice() {
        return price;
    }

    public double getPriceEmissions() {
        return priceEmissions;
    }

    public double getEmissions() {
        return emissions;
    }

    public double getExitProbability() {
        // exitProb = 0;
        return rProbabilityExit;
    }

    public double getEntryProbability() {
        return rProbabilityEntry;
    }

    public double getProfit() {
        return profits.get(0, 0);
    }

    public double getValue() {
        return rV;
    }

    public double getPolicy() {
        return rInvestment;
    }

    @Override
    public String toString() {
        return pmUtility.stringPrettyPrintVector(state) + "," + rV + "," + rInvestment + "," + rProbabilityInvestment + "," + rProbabilityEntry + "," + rProbabilityExit; // +" : "+utility.pmUtility.stringPrettyPrintVector(quantities)+" : "+utility.pmUtility.stringPrettyPrintVector(profits);
    }

    public Jama.Matrix getMatrix() {
        return state;
    }

    public int compareTo(stateSpacePoint s) {
        for (int i = 0; i < s.getMatrix().getRowDimension(); i++) {
            if (s.getMatrix().get(i, 1) < state.get(i, 1)) { // modified: used to be >
                return 1;
            }
            if (s.getMatrix().get(i, 1) > state.get(i, 1)) { // modified: used to be <
                return -1;
            }
        }
        for (int i = 0; i < s.getMatrix().getRowDimension(); i++) {
            if (s.getMatrix().get(i, 0) < state.get(i, 0)) { // modified: used to be >
                return 1;
            }
            if (s.getMatrix().get(i, 0) > state.get(i, 0)) { // modified: used to be <
                return -1;
            }
        }
        return 0;
    }

    public Boolean getShortcut() {
        return shortcut;
    }

    public Integer getShortcutIndex() {
        return shortcutIndex;
    }

    public void setShortcut(boolean shortcut) {
        this.shortcut = shortcut;
    }

    public void setShortcutIndex(int shortcutIndex) {
        this.shortcutIndex = shortcutIndex;
    }

    public double getConsumerWelfare() {
        return consumerWelfare;
    }

    public double getCarbonRevenue() {
        return carbonRevenue;
    }

    public void setConsumerWelfare(double consumerWelfare) {
        this.consumerWelfare = consumerWelfare;
    }

    public double getMarketQ() {
        return marketQ;
    }

    public void setMarketQ(double marketQ) {
        this.marketQ = marketQ;
    }

    public double getInvestProb() {
        return rProbabilityInvestment;
    }

    public double getExitCutoff() {
        // cutoff is the state value
        return stateValue;
    }

    void setEntryProbabilityShortcut(double entryProbability) {
        entryProb = entryProbability;
    }

    void setExitProbabilityShortcut(double exitProbability) {
        exitProb = exitProbability;
    }

    void setInvestmentProbability(double investmentProbability) {
        investProb = investmentProbability;
    }

    void setPolicyShortcut(double policy) {
        stateInvestment = policy;
    }

    void setValueShortcut(double value) {
        stateValue = value;
    }

    public double getChangeV() {
        return changeV;
    }

    public boolean getABSORBED() {
        return ABSORBED;
    }

    /**
     * 
     * METHODS FOR VALUE FUNCTION AND POLICY COMPUTATION
     *
     */
    void solveValue(boolean solvePolicy, coefficientValues c) {
        // solve value using interpolation

        // set strategies of firms //
        Jama.Matrix policyVector = new Jama.Matrix(numFirms, 2);
        Jama.Matrix entryVector = new Jama.Matrix(numFirms, 1);
        Jama.Matrix exitVector = new Jama.Matrix(numFirms, 1);
        policyVector.set(0, 1, 1);
        entryVector.set(0, 0, 1);  // have to assume that the entry is going to happen to find the optimal investment
        exitVector.set(0, 0, 0);  // converse situation with exiting
        for (int i = 1; i < numFirms; i++) {
            stateSpacePoint point = space.getPoint(state, i);
            if (DEBUG) {
                if (state.get(0, 0) == DEBUG_STATE_1 && state.get(1, 0) == DEBUG_STATE_2) {
                    System.out.println("* filling policy vectors " + point + " x: " + point.getPolicy() + " pr: " + point.getInvestProb());
                }
            }
            policyVector.set(i, 0, point.getPolicy());
            policyVector.set(i, 1, point.getInvestProb());
            entryVector.set(i, 0, point.getEntryProbability());
            exitVector.set(i, 0, point.getExitProbability());
        } // what others do

        // summarize the expected value function conditional on others actions //
        Jama.Matrix modifiedCoeff = new Jama.Matrix(simulationPrimitives.CAPACITY_POINTS + 2, 1);
        if (numFirms == 1) {
            int offset = space.getExpandedIndex(firmId);
            modifiedCoeff = c.getCoefficients(offset).getMatrix(0, simulationPrimitives.CAPACITY_POINTS + 1, 0, 0);
        } else {
            modifiedCoeff = modifiedCoefficient(this, policyVector, entryVector, exitVector, c);
        }

        // compute optimal investment given modified coefficients //
        double optimalInvestmentIgnoringFixedCosts = stateInvestment;
        if (solvePolicy) {
            optimalInvestmentIgnoringFixedCosts = solveOptimalInvestment(modifiedCoeff);
        }
        stateInvestment = optimalInvestmentIgnoringFixedCosts;

        // compute expected value at optimal investment and no investment //
        double valuePositiveAdjustment = computeBellman(optimalInvestmentIgnoringFixedCosts, modifiedCoeff);
        double valueNoInvestment = 0;
        if (state.get(0, 0) == 0 && !simulationPrimitives.SOCIAL_PLANNER) {
            valueNoInvestment = 0;
        } else {
            valueNoInvestment = computeBellman(0, modifiedCoeff);
        }

        valueDoingNothing = valueNoInvestment;
        valueOptimalInvestment = valuePositiveAdjustment;

        // compute probability of investment //
        /**
         * So this needs to be integrated against the probability that the
         * draw is less than the value of exiting at this state.
         */
        if (Double.isNaN(valuePositiveAdjustment)) {
            System.out.println("Detecting that value of positive adjustment is NaN in stateSpacePoint.java");
            System.exit(0);
        }

        /**
         * New timing means that I need to calculate W = Emax(V+ - gamma_1, V0) (turning off divestment here)
         * 
         * Then compute Emax ( phi, W)
         * 
         * W is just a number, so just iterated application of conditional means
         */
        double V_diff = valuePositiveAdjustment - valueNoInvestment;
        // new rules:
        // Emax(V+ - gamma1, V0)
        NormalDistribution gammaDist = new NormalDistribution(space.getADJUSTMENT_MU(), space.getADJUSTMENT_SIGMA() * space.getADJUSTMENT_SIGMA());
        double probabilityGammaUnderCutoff = gammaDist.cumulative(V_diff);
        double V_diff_normalized = (V_diff - space.getADJUSTMENT_MU()) / space.getADJUSTMENT_SIGMA();
        double emax = valueNoInvestment * (1.0 - probabilityGammaUnderCutoff);
        
        double expectedGammaUnderCutoff = space.getADJUSTMENT_MU();
        expectedGammaUnderCutoff -= space.getADJUSTMENT_SIGMA() * (standardNormal.probability(V_diff_normalized) / standardNormal.cumulative(V_diff_normalized));
        if(probabilityGammaUnderCutoff>1E-10) {
            emax += probabilityGammaUnderCutoff*(valuePositiveAdjustment-expectedGammaUnderCutoff);
        }
        expectedMaxOptions = emax;

        /*
         * Now, compute the emax(phi, W)
         * 
         * Cutoff is W.  Compute expected phi | phi > W.
         */
        double W_diff_normalized = (emax - space.getSCRAP_MU())/space.getSCRAP_SIGMA();
        double egamma = space.getSCRAP_MU();
        egamma += space.getSCRAP_SIGMA()*(standardNormal.probability(W_diff_normalized)/standardNormal.cumulative(W_diff_normalized));
        NormalDistribution phiDist = new NormalDistribution(space.getSCRAP_MU(), space.getSCRAP_SIGMA()*space.getSCRAP_SIGMA());
        double probabilityPhiAboveCutoff = 1.0-phiDist.cumulative(emax);
        if(probabilityPhiAboveCutoff<1E-10) {
            egamma = 0;
        }
        double emaxv = probabilityPhiAboveCutoff*egamma + (1.0-probabilityPhiAboveCutoff)*emax;

        if (simulationPrimitives.SOCIAL_PLANNER) {
            stateValue += consumerWelfare;
            for (int i = 1; i < numFirms; i++) {
                stateSpacePoint swap = space.getPoint(state, i);
                stateValue += swap.getProfit();
            }
        }

        // compute entry and exit probabilities and set stateValue //
        if (state.get(0, 0) == 0) {
            /**
             * Calculation for New Entrants
             */
            exitProb = 0;
            investProb = 1;

            if (optimalInvestmentIgnoringFixedCosts > 1) { // && investProb > 1e-15
                /**
                 * Adjust for fact that entrant has draws on both investment
                 * costs and entry costs.
                 */
                double sumVariance = Math.pow(space.getENTRY_SIGMA(), 2) + Math.pow(space.getADJUSTMENT_SIGMA(), 2);
                double entryAlpha = (valueOptimalInvestment - (space.getADJUSTMENT_MU() + space.getENTRY_MU())) / Math.sqrt(sumVariance);
                entryProb = standardNormal.cumulative(entryAlpha);

                entryValueCutoff = valueOptimalInvestment; // i think; although I am not seeing how this is not zero
                // System.out.println("incumbent capacity: " + state.get(1, 0) + " probEntry: " + entryProb + " cutoff: " + stateValue);
            } else {
                entryProb = 0;
            }
            if (investProb < 0.0001 && entryProb < 0.0001 && simulationPrimitives.ABSORB && modifiedCoeff == modifiedCoefficient) {
                ABSORBED = true;
            }
            /**
             * Update state value for potential entrants.
             */
            stateValue = valueOptimalInvestment;
        } else {
            /**
             * Calculation for Incumbents
             */
            entryProb = 1;

            exitProb = probabilityPhiAboveCutoff;
            investProb = (1.0-exitProb)*(probabilityGammaUnderCutoff);
            stateValue = emaxv;
        }

        // update coefficients
        modifiedCoefficient = modifiedCoeff;

    }

    private double computeBellman(double investment, Matrix c) {

        double value = 0;

        // evaluate expected value at final capacity using spline //
        value = Interpolate.getValue(state.get(0, 0) + investment, c, 1, simulationPrimitives.CAPACITY_POINTS,
                simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
        value *= simulationPrimitives.DISCOUNT_FACTOR;

        // add current profit and substract investment costs //
        value += getProfit();
        if (investment > 0) {
            value -= investment * space.getINVESTMENT_MARGINAL_COST(); // simulationPrimitives.INVESTMENT_MARGINAL_COST;
            value -= investment * investment * space.getINVESTMENT_MARGINAL_COST2(); // simulationPrimitives.INVESTMENT_MARGINAL_COST2;
        }

        if (Double.isNaN(value)) {
            System.out.println("---------------");
            System.out.println("State causing problems: "+this);
            System.out.println("investment: " + investment + " profit: " + getProfit() + " value: " + value + " " + (Interpolate.getValue(state.get(0, 0) + investment, c, 1, simulationPrimitives.CAPACITY_POINTS,
                    simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC)));
        }
        return value;

    }

    private double solveOptimalInvestment(Matrix c) {

        double optimalInvestment = 0;
        if (state.get(0, 0) == simulationPrimitives.MAX_CAPACITY) {
            return 0;
        }

        /* NOTE: gets new spline coefficients taking into account costs of investment
        alternative way: add marginal cost of investment to getOptimum method */

        // compute value at investment candidates //
        Jama.Matrix valueInvestment = new Jama.Matrix(simulationPrimitives.CAPACITY_POINTS, 1);
        double inc = 0;
        double lowerb = 0;
        double minInvest = 0;
        if (state.get(0, 0) > 0) {
            inc = (simulationPrimitives.MAX_CAPACITY - state.get(0, 0)) / (simulationPrimitives.CAPACITY_POINTS - 1);
        } else { // if new entrant
            inc = (simulationPrimitives.MAX_CAPACITY - simulationPrimitives.MIN_CAPACITY) / (simulationPrimitives.CAPACITY_POINTS - 1);
            lowerb = simulationPrimitives.MIN_CAPACITY;
            minInvest = simulationPrimitives.MIN_CAPACITY;
        }
        //System.out.print("For state "+state.get(0, 0)+" ");
        for (int i = 0; i < simulationPrimitives.CAPACITY_POINTS; i++) { // create candidate investments and evaluate value
            double invest = minInvest + inc * i;
            //System.out.print(invest+" ");
            double value = computeBellman(invest, c);
            valueInvestment.set(i, 0, value);
        }
        //System.out.println();

        // fit spline and get optimum //
        Jama.Matrix cInvest = Interpolate.getCoeff(valueInvestment, 0, 0, 1, simulationPrimitives.CAPACITY_POINTS, inc);
        optimalInvestment = Interpolate.getOptimum(cInvest, 1, simulationPrimitives.CAPACITY_POINTS, lowerb, inc);

        //System.out.println(lowerb+" "+inc+" "+optimalInvestment);
        //valueInvestment.print(0,0);

        // minmax it to be within established range //
        return Math.min(Math.max(lowerb, optimalInvestment), simulationPrimitives.MAX_CAPACITY);


    }

    public Jama.Matrix modifiedCoefficient(stateSpacePoint start, Jama.Matrix policy, Jama.Matrix entry, Jama.Matrix exit, coefficientValues c) {
        /** this function computes the coefficients that represent
         * the single-dimensional spline of tomorrows EV conditional on what others do
         * it is similar to the transition matrix computation */
        int N = simulationPrimitives.CAPACITY_POINTS;
        int D = simulationPrimitives.NUM_FIRMS;
        Jama.Matrix modifiedCoeff = new Jama.Matrix(N + 2, 1);
        double[][] tempC = new double[N + 2][1];

        // loop over different actions that others can do (i.e. enter/exit/stay/invest)
        double sumPrS = 0;
        ArrayList actionList = space.getActionSpace();
        Jama.Matrix probability = new Jama.Matrix(D, 3, 0);
        for (int n = 1; n < D; n++) {
            // exit or not enter
            if (start.getMatrix().get(n, 0) == 0) {
                probability.set(n, 0, (1 - entry.get(n, 0))); //System.out.print(prS);
            } else {
                probability.set(n, 0, exit.get(n, 0));
            }
            // stay
            if (start.getMatrix().get(n, 0) != 0) {
                probability.set(n, 1, (1 - exit.get(n, 0)) * (1 - policy.get(n, 1)));
            }
            // stay and invest or enter (assumes if enter then invest)
            if (start.getMatrix().get(n, 0) == 0) {
                probability.set(n, 2, entry.get(n, 0)); //System.out.print(prS);
            } else {
                probability.set(n, 2, (1 - exit.get(n, 0)) * policy.get(n, 1));
            }
        }
        for (int i = 0; i < actionList.size(); i++) { // actions loop
            // 1. Compute probability of action happening and active firms
            double prS = 1;
            Jama.Matrix currentAction = (Jama.Matrix) actionList.get(i);
            for (int n = 1; n < D; n++) {
                prS = prS * probability.get(n, (int) currentAction.get(n - 1, 0));
            }
            // 2. Include in expectation if positive probability of action (higher than 0.001)
            if (prS > 0.001) {
                sumPrS += prS;
                int dActive = 0; // number of OTHER FIRMS active
                Jama.Matrix result = new Jama.Matrix(D, 2);
                result.set(0, 0, start.getMatrix().get(0, 0) + policy.get(0, 0));
                result.set(0, 1, start.getMatrix().get(0, 1));
                for (int n = 1; n < D; n++) {
                    if (currentAction.get(n - 1, 0) == 0) { // exit or not enter
                        result.set(n, 0, 0);
                        result.set(n, 1, 0); // set as potential entrant for tomorrow
                    } else if (currentAction.get(n - 1, 0) == 1) { // stay
                        result.set(n, 0, start.getMatrix().get(n, 0));
                        result.set(n, 1, start.getMatrix().get(n, 1)); // stay with same id
                        dActive += 1;
                    } else { // stay and invest or enter (assumes if enter then invest)
                        result.set(n, 0, start.getMatrix().get(n, 0) + policy.get(n, 0));
                        result.set(n, 1, start.getMatrix().get(n, 1)); // stay with same id
                        dActive += 1;
                    }
                }
                if (prS > .999 && modifiedCoefficient != null && simulationPrimitives.ABSORB && result.minus(start.getMatrix()).norm2() == 0) {
                    //System.out.println("Avoiding computation! :)");
                    return modifiedCoefficient;
                } else {
                    // setup offset and pick relevant coefficients
                    Jama.Matrix coef = c.getCoefficients(space.getExpandedIndex(result.getMatrix(0, D - 1, 1, 1)));
                    for (int j = 0; j < N + 2; j++) { // for each coefficient
                        double newC = 0;
                        // determine relevant coefficient range to reduce computation time along each dimension (other firms active)
                        double[][] basis = new double[dActive][4];
                        int[][] index = new int[dActive][4];
                        int count = 0;
                        for (int d = 1; d < D; d++) {
                            if (result.get(d, 0) > 0) {
                                int ind = (int) java.lang.Math.floor((result.get(d, 0) - simulationPrimitives.MIN_CAPACITY) / simulationPrimitives.CAPACITY_INC);
                                for (int k = java.lang.Math.max(ind, 0); k < java.lang.Math.min(ind + 4, N + 2); k++) { // loop only relevant states
                                    basis[count][k - ind] = Interpolate.cubicFunction(result.get(d, 0), k, N, simulationPrimitives.MIN_CAPACITY, simulationPrimitives.CAPACITY_INC);
                                    index[count][k - ind] = k;
                                }
                                count += 1;
                            }
                        }
                        // compute coefficient
                        for (int k = 0; k < java.lang.Math.pow(4, dActive); k++) {  // loop only over relevant coefficients - 4 each dimension
                            double update = 1;
                            int ind = 0;
                            count = 0;
                            for (int d = 1; d < D; d++) {
                                if (result.get(d, 0) > 0) {
                                    int t = (int) (java.lang.Math.floor(k / java.lang.Math.pow(4, count))
                                            - java.lang.Math.floor(k / java.lang.Math.pow(4, count + 1)) * 4);
                                    update = update * basis[count][t];
                                    ind += index[count][t] * java.lang.Math.pow(N + 2, dActive - count - 1);
                                    count += 1;
                                }
                            }
                            ind = (int) (ind + j * java.lang.Math.pow(N + 2, dActive));
                            newC += update * coef.get(ind, dActive); // pick relevant coefficient based on other active firms
                        }
                        // weight by probability of actions combination
                        tempC[j][0] += prS * newC;
                    } // each coefficient
                }
            } // if positive probability
        } // each action
        //System.out.println(sumPrS);

        modifiedCoeff.setMatrix(0, N + 1, 0, 0, new Jama.Matrix(tempC).times(1 / sumPrS));
        return modifiedCoeff;
    }

    public double getValueOptimalInvestment() {
        return valueOptimalInvestment;
    }

    public double getValueDoingNothing() {
        return valueDoingNothing;
    }

    public double getEntryValueCutoff() {
        return entryValueCutoff;
    }

    /**
     * @return the expectedMaxOptions
     */
    public double getExpectedMaxOptions() {
        return expectedMaxOptions;
    }
}
