package simulation.mpneSolver;

/*
 * stateSpaceTableModel.java
 *
 * Created on May 4, 2004, 8:29 AM
 */
import java.io.Serializable;
import java.text.NumberFormat;
import simulation.mpneSimulation.utility.simulationPrimitives;
import utility.pmUtility;

/**
 *
 * @author  Stephen P. Ryan
 */
public class stateSpaceTableModel extends javax.swing.table.AbstractTableModel implements Serializable {

    private stateSpace space;
    private NumberFormat nf = NumberFormat.getInstance();

    /** Creates a new instance of stateSpaceTableModel */
    public stateSpaceTableModel(stateSpace spaceRef) {
        space = spaceRef;
    }

    public int getColumnCount() {
        return 14;
    }

    public int getRowCount() {
        return space.getStateSpaceSize();
    }

    @Override
    public Class<?> getColumnClass(int columnIndex) {
        return getValueAt(0, columnIndex).getClass();
    }

    @Override
    public String getColumnName(int col) {
        String[] names = {"i", "Point", "Type", "Value", "ProbInv", "Exit", "Entry", "Profit", "Price", "ABSORBd", "ChangeV", "CS", "MarketQ", "NoAdjInv"};
        return names[col];
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        stateSpacePoint point = space.getPoint(rowIndex);
        if (columnIndex == 0) {
            return nf.format(rowIndex);
        }
        if (columnIndex == 1) {
            return pmUtility.stringPrettyPrintVector(point.getMatrix(), 0);
        }
        int adjustment = 0;
        if (columnIndex + adjustment == 2) {
            return pmUtility.stringPrettyPrintVector(point.getMatrix().getMatrix(0, simulationPrimitives.NUM_FIRMS - 1, 1, 1), 0);
        }
        if (columnIndex + adjustment == 3) {
            return nf.format(point.getValue());
        }
        if (columnIndex + adjustment == 4) {
            // return nf.format(point.getPolicy());
            return nf.format(point.getInvestProb());
        }
        if (columnIndex + adjustment == 5) {
            return nf.format(point.getExitProbability());
        }
        if (columnIndex + adjustment == 6) {
            return nf.format(point.getEntryProbability());
        }
        if (columnIndex + adjustment == 7) {
            return nf.format(point.getProfit());
        }
        if (columnIndex + adjustment == 8) {
            return nf.format(point.getPrice());
        }
        if (columnIndex + adjustment == 9) {
            if (point.getABSORBED()) {
                return new Boolean(point.getABSORBED());
            } else {
                return false;
            }
        }
        if (columnIndex + adjustment == 10) {
            return nf.format(point.getChangeV());
        }
        if (columnIndex + adjustment == 11) {
            return nf.format(point.getConsumerWelfare());
        }
        if (columnIndex + adjustment == 12) {
            return nf.format(point.getMarketQ());
        }
        if (columnIndex + adjustment == 13) {
            return nf.format(point.getPolicy());
        }
        return "Null";
    }
}
