/*
 * bsplineBasisGenerator.java
 *
 * Created on May 20, 2004, 8:37 AM
 */
package utility;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.Serializable;

/**
 *
 * @author  Stephen P. Ryan
 */
public class bsplineBasisGenerator implements Serializable {

    private Jama.Matrix basis;
    private Jama.Matrix x;
    public int NUM_INTERIOR_KNOTS = 2;  // NOTE: this has been 2 throughout; experiment to see what happens when this gets bigger
    private double[] knotVector = new double[NUM_INTERIOR_KNOTS + 8];
    private static double END_KNOT_SPACING = 0.1;
    private static double END_KNOT_GAP = 1; // 1E-5;

    // Define interior knots as those not including the two knots outside the data range, and the two knots at the limits
    // of the data range.
    public bsplineBasisGenerator(double low, double high) {
        // just want to create a uniform b-spline over this interval
        // for putting into my regressions
        // only have to initialize the knot vector and then use getBasis(double x) below to get the required stuff
        double range = high - low;
        double knotSpacing = range / (NUM_INTERIOR_KNOTS + 1);
        for (int i = 4; i < knotVector.length - 2; i++) {
            knotVector[i] = low + knotSpacing * (i - 3);
        }
        for (int i = 0; i < 4; i++) {
            knotVector[i] = low - END_KNOT_GAP + END_KNOT_SPACING * i;
            knotVector[knotVector.length - i - 1] = high + END_KNOT_GAP - END_KNOT_SPACING * i;
        }
    }

    public bsplineBasisGenerator(double low, double high, int numInteriorKnots) {
        NUM_INTERIOR_KNOTS = numInteriorKnots;
        knotVector = new double[NUM_INTERIOR_KNOTS+8];
        // just want to create a uniform b-spline over this interval
        // for putting into my regressions
        // only have to initialize the knot vector and then use getBasis(double x) below to get the required stuff
        double range = high - low;
        double knotSpacing = range / (NUM_INTERIOR_KNOTS + 1);
        for (int i = 4; i < knotVector.length - 2; i++) {
            knotVector[i] = low + knotSpacing * (i - 3);
        }
        for (int i = 0; i < 4; i++) {
            knotVector[i] = low - END_KNOT_GAP + END_KNOT_SPACING * i;
            knotVector[knotVector.length - i - 1] = high + END_KNOT_GAP - END_KNOT_SPACING * i;
        }
    }

    public Jama.Matrix getBasisMatrix(Jama.Matrix x) {
        basis = new Jama.Matrix(x.getRowDimension(), NUM_INTERIOR_KNOTS + 4);
        for (int i = 0; i < x.getRowDimension(); i++) {
            for (int j = 0; j < NUM_INTERIOR_KNOTS + 4; j++) {
                basis.set(i, j, B(j, 4, x.get(i, 0), knotVector));
            }
        }
        return basis;
    }

    public Jama.Matrix getBasis(double x) {
        Jama.Matrix result = new Jama.Matrix(1, NUM_INTERIOR_KNOTS + 4);
        for (int j = 0; j < NUM_INTERIOR_KNOTS + 4; j++) {
            result.set(0, j, B(j, 4, x, knotVector));
        }
        return result;
    }

    private double B(int ell, int m, double x, double[] knotVector) {
        // System.out.println("L: "+ell+" m: "+m+" x: "+x);
        double value = 0;
        if (m == 1) {
            if (x >= knotVector[ell] && x < knotVector[ell + 1]) {
                value = 1;
            }
        } else {
            double piece1 = (x - knotVector[ell]) / (knotVector[ell + m - 1] - knotVector[ell]);
            if (knotVector[ell + m - 1] - knotVector[ell] == 0) {
                piece1 = 0;
            }
            piece1 *= B(ell, m - 1, x, knotVector);
            double piece2 = (knotVector[ell + m] - x) / (knotVector[ell + m] - knotVector[ell + 1]);
            if (knotVector[ell + m] - knotVector[ell + 1] == 0) {
                piece2 = 0;
            }
            piece2 *= B(ell + 1, m - 1, x, knotVector);
            value = piece1 + piece2;
        }
        return value;
    }

    public Jama.Matrix getBasis() {
        return basis;
    }

    public Jama.Matrix getFits(Jama.Matrix y) {
        return pmUtility.getFitsAndErrors(y, getBasis(), pmUtility.OLS(getBasis(), y, false));
    }

    public void writeResults(String filename, Jama.Matrix y) {
        Jama.Matrix fits = getFits(y);

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(filename));
            out.write("x,y,fit,error\n");
            for (int i = 0; i < fits.getRowDimension(); i++) {
                out.write(x.get(i, 0) + "," + fits.get(i, 0) + "," + fits.get(i, 1) + "," + fits.get(i, 2) + "\n");
            }
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
