/*
 * priceThesis.java
 *
 * Created on July 2, 2004, 5:56 PM
 */
package utility;

import JSci.maths.statistics.NormalDistribution;
import java.util.Random;

/**
 *
 * @author  Stephen P. Ryan
 */
public class demandCurve {

    // these are the LIML estimates using ivreg2 logq (logp = wage96 gas96 elec96 coal96) regionDummy*, liml
    private double a; //  = 20.38852;
    private double b; //  = -2.960379;
    private double[] State = {-0.3419287, .2974796, -.5765546, .1727799, 1.048445,
        -.1286832, .3664693, -.4060629, -.3646592, -.6234806, -.5296743, -.2935221, -.5737035,
        -.3068601, -.4714587, .2949728, -.0206987, -.1160749, -.7546033, .2830412, -.9172346,
        -.4300832, 2.028863, .2426433, -.2217379};
    private double shockVariance=0;
    private NormalDistribution stdNormal = new NormalDistribution();
    private boolean USE_SHOCK = false;
    private Random rng = new Random(919);

    public demandCurve() {
    }

    public double getPrice(double quantity, int state) {
        double stateAdjustment = 0;
        if (state > 0) {
            stateAdjustment = State[state - 1];
        }
        double logPrice = (Math.log(quantity) - a - stateAdjustment) / b;
        /**
         * This approach for sure isn't going to work since it assigns different
         * prices to the same firms at the same time.
         */
        if(USE_SHOCK) {
            logPrice += stdNormal.inverse(rng.nextDouble());
        }
        if(state==0) {
        // System.out.println("a: " + a + " b:" + b + " State: " + state + " Quantity: " + quantity + " Price: " + Math.exp(logPrice));
        // System.out.println("Returning: "+Math.exp(logPrice));
        }
        return Math.exp(logPrice);
    }

    // This is actually required when calculating the BR functions in the quantity-setting game
    public double getA() {
        return a;
    }

    public double getB(int state) {
        return b;
    }

    public double getF(int state) {
        if (state > 0) {
            return State[state - 1];
        }
        return 0;
    }

    public void setConstant(double demandConstant) {
        a = demandConstant;
    }

    public void setFixedEffects(double[] stateIntercepts) {
        State = stateIntercepts;
    }

    public void setSlope(double slope) {
        b = slope;
    }

    /**
     * @return the shockVariance
     */
    public double getShockVariance() {
        return shockVariance;
    }

    /**
     * @param aShockVariance the shockVariance to set
     */
    public void setShockVariance(double aShockVariance) {
        shockVariance = aShockVariance;
        stdNormal = new NormalDistribution(0, shockVariance);
    }
}
